/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2002-2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: LockManager.java,v 1.2 2003/06/09 06:28:38 tanderson Exp $
 */
package	org.exolab.core.database.recman;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.text.DateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;


/**
 * Manages database locks for {@link PageManagedDatabase} instances
 *
 * @version     $Revision: 1.2 $ $Date: 2003/06/09 06:28:38 $
 * @author      <a href="mailto:tima@intalio.com">Tim Anderson</a>
 * @see         PageManagedDatabase
 */
class LockManager implements Runnable {

    /**
     * The set of lock names and their corresponding files
     */
    private HashMap _locks = new HashMap();

    /**
     * The interval between lock refreshes
     */
    private long _interval = DEFAULT_INTERVAL;

    /**
     * If true, disables lock expiry
     */
    private boolean _disableExpiry = false;

    /**
     * The singleton instance
     */
    private static LockManager _instance = null;

    /**
     * The logger
     */
    private static final Log _log = LogFactory.getLog(LockManager.class);

    /**
     * The default interval between lock refreshes (2 minutes)
     */
    private static final long DEFAULT_INTERVAL = 1000 * 2 * 60;

    /**
     * System property to change the interval between refreshes (in seconds)
     */
    private static final String INTERVAL =
        "org.exolab.core.database.recman.lockRefreshInterval";

    /**
     * System property name to disable expiry
     */
    private static final String DISABLE_EXPIRY = 
        "org.exolab.core.database.recman.disableExpiry";

    /** 
     * Create a new lock manager
     */
    private LockManager() {
        Integer interval = Integer.getInteger(INTERVAL);
        if (interval != null && (interval.intValue() >= 30)) {
            // only change the interval if the property >= 30 seconds
            _interval = interval.intValue() * 1000;
        }
        _log.debug("Lock refresh interval=" + (_interval / 1000));

        if (System.getProperty(DISABLE_EXPIRY) != null) {
            _disableExpiry = true;
        }

        if (!_disableExpiry) {
            Thread thread = new Thread(this, "PMD-LockManager");
            thread.setDaemon(true);
            thread.setPriority(Thread.MIN_PRIORITY);
            thread.start();
        } else {
            _log.debug("Lock expiry is disabled");
        }
    }

    /**
     * Return the singleton instance
     */
    public static synchronized LockManager getInstance() {
        if (_instance == null) {
            _instance = new LockManager();
        }
        return _instance;
    }

    /**
     * Acquire a new lock
     */
    public synchronized void acquire(String name) throws IOException {
        String lockName = getLockName(name);
        File lock = new File(lockName);
        if (lock.exists()) {
            if (!_disableExpiry) {
                long time = read(lock);
                if (expired(time)) {
                    _log.debug("Removing expired lock file: " +
                                       lockName);
                    lock.delete();
                } else {
                    DateFormat formatter = DateFormat.getDateTimeInstance(
                        DateFormat.SHORT, DateFormat.SHORT);

                    throw new IOException(
                        "Cannot acquire lock=" + lockName +
                        " - a lock exists that will expire at " +
                        formatter.format(new Date(time)));
                }
            } else {
                throw new IOException("Cannot acquire lock=" + lockName +
                                      " - a lock already exists");
            }
        }
                
        if (lock.createNewFile()) {
            lock.deleteOnExit();
        } else {
            throw new IOException("Cannot acquire lock for name=" + name);
        }
        _log.debug("Acquiring lock=" + lockName);
        write(lock);
        _locks.put(name, lock);
    }

    /**
     * Release a lock
     */
    public synchronized void release(String name) throws IOException {
        String lockName = getLockName(name);
        File lock = (File) _locks.remove(name);
        if (lock == null) {
            throw new IOException("Lock=" + lockName + " does not exist");
        }
        _log.debug("Releasing lock=" + lockName);
        lock.delete();
    }

    public void run() {
        while (true) {
            try {
                Thread.currentThread().sleep(_interval);
            } catch (InterruptedException ignore) {
            }
            refreshLocks();
        }
    }

    private boolean expired(long time) {
        long now = System.currentTimeMillis();
        return (now > time);
    }

    private synchronized void refreshLocks() {
        Iterator iter = _locks.entrySet().iterator();
        while (iter.hasNext()) {
            Map.Entry entry = (Map.Entry) iter.next();
            String name = (String) entry.getKey();
            File lock = (File) entry.getValue();
            refreshLock(name, lock);
        }
    }

    private void refreshLock(String name, File lock) {
        String lockName = getLockName(name);
        _log.debug("Refreshing lock=" + lockName);
        try {
            write(lock);
        } catch (IOException exception) {
            _log.error("Failed to refresh lock=" + lockName, exception);
        }
    }

    private void write(File lock) throws IOException {
        ObjectOutputStream stream = new ObjectOutputStream(
            new FileOutputStream(lock));        
        long expiry = System.currentTimeMillis() + (_interval * 2);
        stream.writeLong(expiry);
        stream.close();
    }

    private long read(File lock) throws IOException {
        ObjectInputStream stream = new ObjectInputStream(
            new FileInputStream(lock));
        long result = stream.readLong();
        stream.close();
        return result;
    }

    private String getLockName(String name) {
        return name + ".lock"; 
    }        
       	
} //-- LockManager
