/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 1999 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: PMDHandle.java,v 1.3 2001/06/12 07:11:35 jima Exp $
 *
 * Date         Author  Changes
 * 2/24/2000    jima    Created
 */
package org.exolab.core.database.recman;


import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import org.exolab.core.foundation.PersistentCapableIfc;
import org.exolab.core.foundation.FailedToCreateHandleException;
import org.exolab.core.foundation.SessionIfc;
import org.exolab.core.foundation.HandleIfc;
import org.exolab.core.foundation.PersistentObject;

/**
 * This handle references an object in the PageManaged database. Resolving
 * this object implies fetching by id.
 *
 * @version     $Revision: 1.3 $ $Date: 2001/06/12 07:11:35 $
 * @author      <a href="mailto:jima@exoffice.com">Jim Alateras</a>
 **/
public class PMDHandle
    extends PersistentObject
    implements HandleIfc, Externalizable
{
    /**
     * Default constructor. No real use except for serialization
     */
    public PMDHandle()
    {
    }

    /**
     * Construct an instance of this class given the specified persistent
     * object. An object consists of the object identifier
     *
     * @param       object      persistent object
     */
    public PMDHandle(PersistentCapableIfc object)
        throws FailedToCreateHandleException
    {
        if (object.getObjectId().getId() > 0)
        {
            cachedObject_ = object;
            id_ = object.getObjectId().getId();
        }
        else
        {
            throw new FailedToCreateHandleException(
                "The object must be persistent before a creating handle");
        }
    }

    /**
     * Resolve this handle and return a reference to the object or 
     * null if it cannot be resolved.
     *
     * @return      PersistentCapableIfc        resolved object
     */
    public Object resolve()
    {
        if (cachedObject_ == null)
        {
            try
            {
                SessionIfc session = PMDSessionManager.instance().getSession();
                cachedObject_ = session.retrieveObject(id_);
            }
            catch (Exception exception)
            {
                //log an error
            }
        }

        return cachedObject_;
    }
    
    /**
     * Clear the cahced object so that a call to resolve will actually refetch
     * the object from the datanase
     */
    public void clearCache()
    {
        cachedObject_ = null;           
    }

    /**
     * Return true if the object id's are the same and false otherwise
     *
     * @param           object          object to compare against
     * @return          boolean
     */
    public boolean equals(Object object)
    {
        boolean result = false;

        if ((object instanceof PMDHandle) &&
            (((PMDHandle)object).id_ == id_))
        {
            result = true;
        }

        return result;
    }

    // implementation of Externalizable.writeExternal
    public void writeExternal(ObjectOutput stream)
        throws IOException
    {
        stream.writeLong(serialVerionUID);
        stream.writeLong(id_);
        super.writeExternal(stream);
    }

    // implementation of Externalizable.writeExternal
    public void readExternal(ObjectInput stream)
        throws IOException, ClassNotFoundException
    {
		long version = stream.readLong();
		if (version == serialVerionUID) {	
            id_ = stream.readLong();
            super.readExternal(stream); 
        } else {
            throw new IOException("PMDHandle with version " +
                version + " is not supported.");
        }
    }


    transient private PersistentCapableIfc cachedObject_ = null;
    private long id_ = -1;
    static final long serialVerionUID = 1;
}

