/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 1999 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: PMDHashMap.java,v 1.4 2001/06/12 07:11:35 jima Exp $
 *
 * Date         Author  Changes
 * 2/23/2000    jima    Created
 */
package org.exolab.core.database.recman;


import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;
import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import org.exolab.core.foundation.HashMapIfc;
import org.exolab.core.foundation.PersistentCapableIfc;
import org.exolab.core.foundation.PersistentObject;

/**
 * This is an implementation of the HashMapIfc specific to the page managed
 * database where references to objects are managed as PMDHandle instances
 * rather than the objects themselves. This is necessary to optimise the
 * concurrency and performance of the database.
 * <p>
 * In addition this class deals with PersistentObjects rather than just
 * Object instance, since it assumes you are using this clas in a persistence
 * context. Therefore some method, that take an Object may throw a runtime
 * exception if the objects are not of this type.
 * <p>
 * Finally, an instance of this class is created within the context of a 
 * session and within a valid transaction....which I don't have a plan for
 * <p>
 * Instrument for better serialization by overriding the writeObject and 
 * readObject methods in a later version
 *
 * @version     $Revision: 1.4 $ $Date: 2001/06/12 07:11:35 $
 * @author      <a href="mailto:jima@exoffice.com">Jim Alateras</a>
 **/
public class PMDHashMap
    extends PersistentObject
    implements HashMapIfc, Externalizable
{
    /**
     * Return the size of the map
     *
     * @return      int
     */
    public int size()
    {
        return map_.size();
    }

    /**
     * Test for an empty hashmap
     *
     * @return      boolean     true if empty; false otherwise
     */
    public boolean isEmpty()
    {
        return map_.isEmpty();
    }

    /**
     * Return the enumeration of all keys in the map
     *
     * @return      Enumeration
     */ 
    public synchronized Enumeration keys()
    {
        return map_.keys();
    }

    /**
     * Return an enumeration of all the elements in the map
     *
     * @return      Enumeration
     */
    public synchronized Enumeration elements()
    {
        return map_.elements();
    }

    /** 
     * Test to see whether the map contains the specified value
     *
     * @param       element     element to test against
     * @return      boolean     true if key is mapped; false otherwise
     */
    public synchronized boolean containsValue(PersistentCapableIfc element)
    {
        return map_.containsValue(element);
    }
    
    /**
     * Return true if the specified key maps to an element in the map
     *
     * @param       key         key to test against
     * @return      boolean     true if it maps; false otherwise
     */
    public synchronized boolean containsKey(Object key)
    {
        return map_.containsKey(key);
    }

    /**
     * Return the object mapped to this key.
     *
     * @param       key         key to lookup
     * @return      PersistentCapableIfc
     */
    public synchronized PersistentCapableIfc get(Object key)
    {
        return (PersistentCapableIfc)map_.get(key);
    }

    /**
     * Add the key and associated element to the map and return a reference
     * to the object. In this implementation of the HashMapIfc the element
     * must also be of type PersistentCapableIfc
     *
     * @param       key         key to map
     * @param       element     corresponding element
     * @return      PersistentCapableIfc
     */
    public synchronized PersistentCapableIfc put(Object key, 
        PersistentCapableIfc element)
    {
        return (PersistentCapableIfc)map_.put(key, element);
    }

    /**
     * Remove the object mapped to the specified key from the map and return
     * it to the client
     *
     * @param       key         key to remove
     * @param       PersistentCapableIfc
     */
    public synchronized PersistentCapableIfc remove(Object key)
    {
        return (PersistentCapableIfc)map_.remove(key);
    }

    /**
     * Clear the map
     */
    public synchronized void clear()
    {
        map_.clear();
    }

    /**
     * Return a refernece to the enclosed map. To be used it will have
     * be cast to the Hashtable
     *
     * @return      Object
     */
    public Object getMapObject()
    {
        return map_;
    }

    // implementation of Externalizable.writeExternal
    public void writeExternal(ObjectOutput stream)
        throws IOException
    {
        stream.writeLong(serialVerionUID);
        stream.writeObject(map_);
        super.writeExternal(stream);
    }

    // implementation of Externalizable.writeExternal
    public void readExternal(ObjectInput stream)
        throws IOException, ClassNotFoundException
    {
		long version = stream.readLong();
		if (version == serialVerionUID) {	
            map_ = (Hashtable)stream.readObject();
            super.readExternal(stream); 
        } else {
            throw new IOException("PMDHashMap with version " +
                version + " is not supported.");
        }
    }

    static final long serialVerionUID = 1;
    private Hashtable map_ = new Hashtable();
}


