/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 1999 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: PMDRootTable.java,v 1.6 2003/06/09 06:28:39 tanderson Exp $
 *
 * Date         Author  Changes
 * 2/24/2000    jima    Created
 */
package org.exolab.core.database.recman;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.util.Hashtable;
import java.util.Enumeration;
import org.exolab.core.foundation.PersistentObject;
import org.exolab.core.foundation.PersistentCapableIfc;

/**
 * This object uses the hashtable to implement the root table. A root table
 * stores all the named roots for the database. It implements the 
 * PersistentCapableIfc inerfacce through PersistentObject
 * this object implies fetching by id.
 *
 * @version     $Revision: 1.6 $ $Date: 2003/06/09 06:28:39 $
 * @author      <a href="mailto:jima@exoffice.com">Jim Alateras</a>
 **/
class PMDRootTable
    extends PersistentObject
    implements Cloneable, Externalizable
{
    /**
     * Construct a default size root table
     */
    public PMDRootTable()
    {
        table_ = new Hashtable();
    }

    /**
     * Create an instance of this class using the following
     * table.
     *
     * @param       table
     */
    PMDRootTable(Hashtable table)
    {
        table_ = table;
    }

    /**
     * Clone this object and its base objects. Make sure to
     * catch CloneNotSupportedException when calling the base
     * class
     */
    public Object clone()
    {
        PMDRootTable clone = null;

        try {
            clone = (PMDRootTable)super.clone();
            clone.table_ = (Hashtable)clone.table_.clone();
        } catch (CloneNotSupportedException exception) {
            clone = null;
        }

        return clone;
    }

    /**
     *  Return the names of all roots in the table
     *
     * @return          Enumeration         enumeration of String objects
     */
    public Enumeration names()
    {
        return table_.keys();
    }

    /**
     * Return the PersistentCapableIfc object specified by the name or 
     * null is one is not defined
     *
     * @param           name                name of object to return
     * @return          Object
     */
    public Object get(String name)
    {
        return table_.get(name);
    }

    /**
     * Place the object with the specified name into the table
     *
     * @param       name                    name of object
     * @param       object                  object to place
     */
    public void put(String name, Object object)
    {
        table_.put(name, object);
    }
     
    /**
     * Remove the object with the specified name from the table
     *
     * @param       name                name of object to remove
     */
    public void remove(String name)
    {
        table_.remove(name);
    }
     
    /**
     * Test if the following name exists in the table
     *                                  
     * @param       name            name of object
     * @return      boolean         true if it exists
     */
    public boolean containsName(String name)
    {
        return table_.containsKey(name);
    }

    /** 
     * Print the contents of the table
     *
     * @return      String
     */
    public String toString()
    {
        StringBuffer buf = new StringBuffer("RMDRootTable [");

        buf.append(" oid = ");
        buf.append(getId());
        buf.append(" version = ");
        buf.append(getVersion());
        buf.append(" ");
        
        Enumeration keys = table_.keys();
        while (keys.hasMoreElements())
        {
            String key = (String)keys.nextElement(); 
            buf.append(key);
            buf.append(":");
            buf.append(table_.get(key).getClass().getName());
            buf.append(" ");
        }
        buf.append("]");

        return buf.toString();
    }

    // implementation of Externalizable.writeExternal
    public void writeExternal(ObjectOutput stream)
        throws IOException
    {
        stream.writeLong(serialVerionUID);
        stream.writeObject(table_);
        super.writeExternal(stream);
    }

    // implementation of Externalizable.writeExternal
    public void readExternal(ObjectInput stream)
        throws IOException, ClassNotFoundException
    {
		long version = stream.readLong();
		if (version == serialVerionUID) {	
            table_ = (Hashtable)stream.readObject();
            super.readExternal(stream); 
        } else {
            throw new IOException("PMDRootTable with version " +
                version + " is not supported.");
        }
    }

    static final long serialVerionUID = 1;
    private Hashtable table_ = null;
}


