/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 1999 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: PMDSession.java,v 1.4 2003/06/09 06:28:39 tanderson Exp $
 *
 * Date         Author  Changes
 * 1/14/2000    jima    Created
 */

package org.exolab.core.database.recman;

import java.util.Enumeration;
import java.util.Vector;
import java.util.Hashtable;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.exolab.core.foundation.CollectionManagerIfc;
import org.exolab.core.foundation.SessionManagerIfc;
import org.exolab.core.foundation.SessionIfc;
import org.exolab.core.foundation.DatabaseIfc;
import org.exolab.core.foundation.TransactionStatus;
import org.exolab.core.foundation.TransactionIfc;
import org.exolab.core.foundation.ObjectNameNotFoundException;
import org.exolab.core.foundation.ObjectNameExistsException;
import org.exolab.core.foundation.FailedToAcquireLockException;
import org.exolab.core.foundation.FailedToCreateLockException;
import org.exolab.core.foundation.FailedToCreateHandleException;
import org.exolab.core.foundation.DatabaseIOException;
import org.exolab.core.foundation.TransactionInProgressException;
import org.exolab.core.foundation.TransactionNotInProgressException;
import org.exolab.core.foundation.TransactionException;
import org.exolab.core.foundation.DatabaseIOException;
import org.exolab.core.foundation.TransactionContext;
import org.exolab.core.foundation.PersistentCapableIfc;
import org.exolab.core.foundation.HandleIfc;

/**
 * This is an implementation of the SessionIfc interface for the page managed 
 * database. A session is a single-threaded entity that is only  accessible 
 * through the PMDSessionManager (this is where the constraint is enforced).
 * <p>
 * A Session is a mediating class to the pagem managed database and delegates
 * the request to the appropriate component.
 * <p>
 * There is a lack of synchronization in this class and hence lacks thread 
 * safety since it is expected that only one thread will be accessing an
 * instance of this class at any one time.
 *
 * @version     $Revision: 1.4 $ $Date: 2003/06/09 06:28:39 $
 * @author      <a href="mailto:jima@exoffice.com">Jim Alateras</a>
 * @see         PMDSessionManager
 **/
public class PMDSession
    implements SessionIfc, TransactionIfc
{

    /**
     * The logger
     */
    private static final Log _log = LogFactory.getLog(PMDSession.class);

    /**
     * Construct an instance of this class with the specified database. The 
     * manager denotes the creating enitity
     *
     * @param       manager         session manager creating this entity
     * @param       database        database that this session is related too.
     */
    PMDSession(SessionManagerIfc manager, PageManagedDatabase database)
    {
        sessionMgr_ = manager;
        cache_ = new PMDTransactionCache(database);
        txStatus_ = TransactionStatus.CLOSED;
    }

    // implementation of SessionIfc.getRoots
    public Enumeration getRootNames()
    {
        return cache_.getRootNames();
    }

    // implementation of SessionIfc.bind
    public void bind(String name, PersistentCapableIfc object)
        throws ObjectNameExistsException, DatabaseIOException,
            TransactionNotInProgressException
    {
        if (txStatus_.isOpened())
        {
            this.acquireRootLock();
            if (!cache_.containsRoot(name))
            {
                cache_.createRoot(name, object);
            }
            else
            {
                // the root already exists; throw an exception
                throw new ObjectNameExistsException("bind failed since " + 
                    name + " already exists.");
            }
        }
        else
        {
            throw new TransactionNotInProgressException(
                "Must be in a transaction");
        }
    }

    // implementation of SessionIfc.unbind
    public void unbind(String name)
        throws ObjectNameNotFoundException, DatabaseIOException,
            TransactionNotInProgressException
    {
        if (txStatus_.isOpened())
        {
            this.acquireRootLock();
            if (cache_.containsRoot(name))
            {
                cache_.deleteRoot(name);
            }
            else
            {
                // the root does not exist throw an exception
                throw new ObjectNameNotFoundException("unbind failed since " + 
                    name + " does not exists.");
            }
        }
        else
        {
            throw new TransactionNotInProgressException(
                "Must be in a transaction");
        }
    }

    // implementation of SessionIfc.lookup
    public PersistentCapableIfc lookup(String name)
    {
        PersistentCapableIfc result = null;

        try
        {
            result = cache_.lookup(name);
        }
        catch (Exception exception)
        {
            _log.error(exception);
        }

        return result;
    }

    // implementation of SessionIfc.isBound
    public boolean isBound(String name)
    {
        boolean result = false;

        if (this.lookup(name) != null)
        {
            result = true;
        }

        return result;
    }

    // implementation of SessionIfc.acquireLock
    public void acquireLock(PersistentCapableIfc object, int timeout)
        throws FailedToAcquireLockException, FailedToCreateLockException,
            TransactionNotInProgressException
    {
        if (txStatus_.isOpened())
        {
            cache_.acquireObjectLock(object, timeout);
        }
        else
        {
            throw new TransactionNotInProgressException(
                "Must be in a transaction");
        }
    }

    // implementation of TransactionIfc.begin
    public void begin()
        throws TransactionInProgressException
    {
        if (!txStatus_.isOpened())
        {
            // mark the transaction as opened and clear the cache
            cache_.begin();
            txStatus_ = TransactionStatus.OPENED;
            rollbackOnly_ = false;
        }
        else
        {
            _log.error("begin failed. Status of transaction is " +
                txStatus_);
            throw new TransactionInProgressException("status of the tx is " 
                + txStatus_);
        }
    }

    // implementation of TransactionIfc.commit
    public void commit()
        throws TransactionNotInProgressException, TransactionException, 
            DatabaseIOException
    {
        if (txStatus_.isOpened())
        {
            // commit the cache. This will throw a TransactionException is 
            // it fails to commit all the cache. It will always transition the
            // transaction state to committed. If the rollbackOnly flag is set
            // then force a rollback and throw a TransactionException.
            if (rollbackOnly_ == true)
            {
                this.rollback();
                throw new TransactionException(
                    "Transaction was forced to rollback in commit");
            }
            else
            {
                cache_.commit();
                txStatus_ = TransactionStatus.COMMITTED;
            }
        }
        else
        {
            _log.error("commit failed. The status of the tx is " +
                txStatus_);
            throw new TransactionNotInProgressException(
                "commit failed, tx status " + txStatus_);
        }
    }

    // implementation of TransactionIfc.abort
    public void abort()
        throws TransactionNotInProgressException, DatabaseIOException
    {
        if (txStatus_.isOpened())
        {
            // if the transaction is opened, abort the cache and transition
            // the transaction status to ABORTED
            txStatus_ = TransactionStatus.ABORTED;
            cache_.abort();
        }
        else
        {
            _log.error("abort failed. The status of the tx is " +
                txStatus_);
            throw new TransactionNotInProgressException(
                "abort failed, tx status " + txStatus_);
        }
    }

    // implementation of TransactionIfc.abort()
    public void rollback()
        throws TransactionNotInProgressException, DatabaseIOException
    {
        this.abort();
    }

    // implementation of TransactionIfc.setRollbackOnly()
    public void setRollbackOnly(boolean status)
        throws TransactionNotInProgressException
    {
        if (txStatus_.isOpened())
        {
            rollbackOnly_ = status;
        }
        else
        {
            throw new TransactionNotInProgressException(
                "Can't call setRollbackOnly if there is no tx in progress");
        }
    }

    // implementation of TransactionIfc.isOpened()
    public boolean isOpened()
    {
         return txStatus_.isOpened();
    }

    // implementation of TransactionIfc.getTransactionContext
    public TransactionContext getTransactionContext()
    {
        return cache_;
    }
    
    // implementation of SessionIfc.getCurrentTransaction
    public TransactionIfc getCurrentTransaction()
    {
        return this;
    }

    // implementation of SessionIfc.getDatabase()
    public DatabaseIfc getDatabase()
    {
        return cache_.getDatabase();
    }
    
    // implementation of SessionIfc.getSessionManager
    public SessionManagerIfc getSessionManager()
    {
        return sessionMgr_;
    }

    // implementation of SessionIfc.setSessionManager
    public void setSessionManager(SessionManagerIfc manager)
    {
        sessionMgr_ = manager;
    }

    // implementation of SessionIfc.createObject
    public void createObject(PersistentCapableIfc object)
        throws TransactionNotInProgressException, DatabaseIOException
    {
        if (txStatus_.isOpened())
        {
            cache_.createObject(object);
        }
        else
        {
            throw new TransactionNotInProgressException(
                "Can't call createObject if there is no tx in progress");
        }
    }

    
    // implementation of SessionIfc.updateObject
    public void updateObject(PersistentCapableIfc object)
        throws TransactionNotInProgressException, DatabaseIOException
    {
        if (txStatus_.isOpened())
        {
            cache_.updateObject(object);
        }
        else
        {
            throw new TransactionNotInProgressException(
                "Can't call updateObject if there is no tx in progress");
        }
    }
    
    // implementation of SessionIfc.deleteObject
    public void deleteObject(PersistentCapableIfc object)
        throws TransactionNotInProgressException, DatabaseIOException
    {
        if (txStatus_.isOpened())
        {
            cache_.deleteObject(object);
        }
        else
        {
            throw new TransactionNotInProgressException(
                "Can't call deleteObject if there is no tx in progress");
        }
    }

    // implementation of SessionIfc.retrieveObject
    public PersistentCapableIfc retrieveObject(long id)
        throws TransactionNotInProgressException, DatabaseIOException
    {
        PersistentCapableIfc object = null;

        if (txStatus_.isOpened())
        {
            object = cache_.retrieveObject(id);
        }
        else
        {
            throw new TransactionNotInProgressException(
                "Can't call retrieveObject if there is no tx in progress");
        }

        return object;
    }
    
    /**
     * Create a handle for the specified PersistentCapableIfc object. If it 
     * cannot create the handle then throw the FailedToCreateHandleException.
     * If transaction is not in progress then TransactionNotInProgressException 
     * is raised.
     * <p>
     * A handle can be used to reference a persistent entity and can be carried
     * across transactions...although there is some more code that needs to be
     * done in order to make this a reality. (i.e clear the cached object between
     * transactions).
     *
     * @param       object          PersistentCapableIfc
     * @param       HandleIfc
     * @exception   FailedToCreateHandleException
     * @exception   TransactionNotInProgressException
     */
    public HandleIfc createHandle(PersistentCapableIfc object)
        throws FailedToCreateHandleException, TransactionNotInProgressException
    {
        HandleIfc handle = null;

        if (txStatus_.isOpened())
        {
            handle = new PMDHandle(object);
        }
        else
        {
            throw new TransactionNotInProgressException(
                "Can't call createHandle if there is no tx in progress");
        }

        return handle;
    }

    /**
     * Return a reference to the collection manager, which is used to allocate
     * different collection types includinng queues and vectors.
     *
     * @return      CollectionManagerIfc
     */
    public CollectionManagerIfc getCollectionManager()
    {
        return PMDCollectionManager.instance();
    }

    // implementation of SessionIfc.getSessionId
    public long getSessionId()
    {
        return sessionId_;
    }

    // implementation of SessionIfc.isValid
    public boolean isValid()
    {
        // a valid sessionId implies a vlaid session
        return (sessionId_ > 0);
    }

    /**
     * Set the session's identity. This method is scoped at the package level
     *
     * @param       long            session identifier
     */
    void setSessionId(long id)
    {
        sessionId_ = id;
    }

    /**
     * Acquire a lock on the root object. This delegates to the cache. Any
     * error in acquiring the lcok will be translated into a 
     * DatabaseIOException.
     *
     * @exception      DatabaseIOException
     */
    void acquireRootLock()
        throws DatabaseIOException
    {
        try
        {
            cache_.acquireRootLock();
        }
        catch (Exception exception)
        {
            _log.error(exception);
            throw new DatabaseIOException("Error in acquireRootLock " + 
                exception.toString());
        }
    }
    
    private transient long sessionId_ = -1;
    private transient boolean rollbackOnly_ = false;
    private transient SessionManagerIfc sessionMgr_ = null;
    private transient TransactionStatus txStatus_ = TransactionStatus.CLOSED;
    private transient PMDTransactionCache cache_ = null;
}

