/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 1999 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: PMDSessionManager.java,v 1.2 2003/06/09 06:28:39 tanderson Exp $
 *
 * Date         Author  Changes
 * 3/07/2000    jima    Created
 */

package org.exolab.core.database.recman;

import java.util.Enumeration;
import org.exolab.core.util.FastThreadLocal;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.exolab.core.foundation.SessionManagerIfc;
import org.exolab.core.foundation.SessionIfc;
import org.exolab.core.foundation.DatabaseIfc;
import org.exolab.core.foundation.FailedToDestroySessionException;
import org.exolab.core.foundation.FailedToCreateSessionException;

/**
 * Implements the SessionManagerIfc for the page managed database and is 
 * responsible for managing the lifecycle all sessions. A session-per-thread
 * restriction is enforced, meaning that a session belongs to one and only one
 * session and that threads cannot share sessions.
 * <p>
 * This class conforms to the singleton design pattern and is accessible 
 * through PMDSessionManager.instance().
 
 * for a number of interfaces and i
 *
 * @version     $Revision: 1.2 $ $Date: 2003/06/09 06:28:39 $
 * @author      <a href="mailto:jima@exoffice.com">Jim Alateras</a>
 * @see         org.exolab.core.foundation.SessionManagerIfc
 * @see         org,exolab.core.foundation.SessionIfc
 **/
public class PMDSessionManager
    implements SessionManagerIfc
{
    /**
     *  The logger
     */
    private static final Log _log = LogFactory.getLog(PMDSessionManager.class);

    /**
     * The constructor is scoped to the class level to ensure that only one 
     * instance of this class is ever created.
     */
    private PMDSessionManager()
    {
    }

    /**
     * This static method must be called to initialise the session manager 
     * with a database. The SessionManager can only manage one database per
     * virtual machine.
     * <p>
     * Although not programmatically enforced this method must be called
     * before accessing any other class or instance method
     *
     * @param       database
     */
    public static void init(PageManagedDatabase database)
    {
        database__ = database;
    }

    /**
     * This method returns the sxingleton instance of the PMDSessionManager. If
     * such an instance does not yet exist then create it before returning.
     *
     * @return      PMDSessionManager       singleton instance
     */
    public static PMDSessionManager instance()
    {
        if (instance__ == null)
        {
            instance__ = new PMDSessionManager();
        }

        return instance__;
    }

    // implementation of SessionManagerIfc.getSession()
    public SessionIfc getSession()
        throws FailedToCreateSessionException
    {
        SessionIfc session = (SessionIfc)sessions__.get();
        if (session == null)
        {
            // try and create the session
            try
            {
                session = new PMDSession(this, database__);
                sessions__.set(session);
            }
            catch (Exception exception)
            {
                _log.error(exception);
                throw new FailedToCreateSessionException(
                    "PMDSessionManager.getSession : " + exception);
            }
        }

        return session;
    }

    // implementation of SessionManagerIfc.destroySession
    public void destroySession()
        throws FailedToDestroySessionException
    {
        SessionIfc session = (SessionIfc)sessions__.get();
        if (session != null)
        {
            sessions__.set(null);
        }
        else
        {
            // the session does not exist.
            _log.warn("No session is registered for this thread " +
                Thread.currentThread().getName());
            throw new FailedToDestroySessionException("No session active for " +
                Thread.currentThread().getName());
        }
    }

    // manages the sessions per thread
    private transient static PageManagedDatabase database__ = null;
    private transient static FastThreadLocal sessions__ = new FastThreadLocal();
    private transient static PMDSessionManager instance__ = null;
}

