/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 1999 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: PMDTransactionCache.java,v 1.9 2003/06/09 06:28:39 tanderson Exp $
 *
 * Date         Author  Changes
 * 3/07/2000    jima    Created
 */

package org.exolab.core.database.recman;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;
import java.io.IOException;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.exolab.core.foundation.Lock;
import org.exolab.core.foundation.FailedToAcquireLockException;
import org.exolab.core.foundation.FailedToCreateLockException;
import org.exolab.core.foundation.DatabaseIfc;
import org.exolab.core.foundation.ObjectNameExistsException;
import org.exolab.core.foundation.DatabaseIOException;
import org.exolab.core.foundation.ObjectNameNotFoundException;
import org.exolab.core.foundation.PersistentCapableIfc;
import org.exolab.core.foundation.TransactionContext;
import org.exolab.core.foundation.ObjectId;


/**
 * A transaction cache is a layer that lies between the database and the 
 * interacting session (a session is an endpoint) that manages the persistent
 * objects (i.e. PersistentCapableIfc). The PMDTransactionCache has package
 * level scoping so its lifecycle must be managed by other package classes.
 * <p>
 * The cache maintaisn a delta between the session client and the database. It
 * is like a transaction log of changes that must be applied to the database 
 * when a transaction is committed. Similarly, when a transaction is aborted or
 * rolled back then the database must be rolledback.
 * <p>
 * A cache is maintained for roots and all other objects that are accessible 
 * through the roots.
 * <p>
 * Locking is not managed at this level, it is managed by the Session level
 * 
 * @version     $Revision: 1.9 $ $Date: 2003/06/09 06:28:39 $
 * @author      <a href="mailto:jima@exoffice.com">Jim Alateras</a>
 * @see         org.exolab.core.foundation.TransactionContext
 **/
class PMDTransactionCache
    extends TransactionContext
{

    /**
     * The logger
     */
    private static final Log _log = LogFactory.getLog(PMDTransactionCache.class);

    /**
     * Create an instance of the cache given the specified database..
     *
     * @param           database            create a cache for this database
     */
    PMDTransactionCache(PageManagedDatabase database)
    {
        database_ = database;
    }

    /**
     * Return an intance of the database or null if none is assigned
     *
     * @return          DatabaseIfc 
     */
    DatabaseIfc getDatabase()
    {
        return database_;
    }

    /**
     * Return true if the database contains the specified root object or 
     * false otherwise. The root may have been created in this session or 
     * may have been deleted in this session so a systematic check must be
     * made..
     * <p>
     * It will throw a DatabaseIOException if there is a problem working with
     * the database
     *
     * @param           name                query to see if this name is bound
     * @return          boolean
     * @exception       DatabaseIOException
     */
    boolean containsRoot(String name)
        throws DatabaseIOException
    {
        try
        {
            return (lookup(name) != null) ? true : false;
        }
        catch (Exception exception)
        {
            throw new DatabaseIOException("containsRoot failed " + exception);
        }
    }

    /**
     * Return all the root names in the database
     *
     * @return      Enumeration         enumeration of String
     */
    Enumeration getRootNames()
    {
        Enumeration names = null;

        try
        {
            names = database_.getRootNames();
        }
        catch (Exception exception)
        {
            _log.error("getRootNames failed with  " + exception);
        }

        return names;
    }

    /**
     * Create a root object given the name and object. If the root name already
     * exists then throw ObjectNameExistsException. If it cannot create the 
     * root then throw a DatabaseIOException.
     * <p>
     * NOTE: If the object has not yet been persistent in the database it 
     * will be made persistent before it is bound...am I imposing too much
     * policy here. Should I throw an exception
     *
     * @parsm           name                name binding
     * @param           object              object to bind
     * @exception       ObjectNameExistsException
     * @exception       DatabaseIOException
     */
    void createRoot(String name, PersistentCapableIfc object)
        throws ObjectNameExistsException, DatabaseIOException
    {
        // acquire an object on the root of al objects. This is a real
        // hotspot...as are all collections
        try
        {
            acquireRootLock();
            PMDRootTable table = getNamedRootTable();

            // check to see if the object is bound and if it is throw an
            // exception
            if (table.get(name) != null)
            {
                throw new ObjectNameExistsException(name + 
                    " is already bound to an object in the database");
            }
            else
            {
                // okay now we should check that the object has been initially
                // persistet by the database, If it hasn't then insert it into
                // the database, update the object id and then do the bind.
                if (object.getObjectId().getId() == 0)
                {
                    this.createObject(object);
                }

                // bind the name to the object and create a trans log
                // to update the named root
                table.put(name, object.getObjectId());
                modifiedRootTable_ = true;
                objectTranLog_.put(table.getObjectId(), 
                    new TranLogEntry(TranLogEntry.UPDATED, table));
            }
        }
        catch(Exception exception)
        {
            throw new DatabaseIOException(
                "Can't acquire lock to root object");
        }
    }

    /**
     * Delete the root object specified by name. If the root object does not
     * exist then throw the ObjectNameNotFoundException, Additionally, if there
     * is a database error then throw DatabaseIOException
     *
     * @param           name                root to delete
     * @exception       ObjectNameNotFoundException
     * @exception       DatabaseIOException
     */
    void deleteRoot(String name)
        throws ObjectNameNotFoundException, DatabaseIOException
    {
        // acquire an object on the root of al objects. This is a real
        // hotspot...as are all collections
        try
        {
            acquireRootLock();
            PMDRootTable table = getNamedRootTable();

            // check to see if the object is bound and if it is throw an
            // exception
            if (table.get(name) == null)
            {
                throw new ObjectNameNotFoundException(name + 
                    " is not bound to an object in the database");
            }
            else
            {
                // unbind the name to the object and create a trans log
                // to update the named root
                table.remove(name);
                modifiedRootTable_ = true;
                objectTranLog_.put(table.getObjectId(), 
                    new TranLogEntry(TranLogEntry.UPDATED, table));
            }
        }
        catch (Exception exception)
        {
            throw new DatabaseIOException("Exception in deleteRoot " + 
                exception);
        }
    }

    /**
     * Lookup the specified root name and return the associated object
     * Do not currently cache the object, although at some later time
     * we may decide to do so. If there are any errors then throw the 
     * DatabaseIOException
     *
     * @param       name            name of object to lookup
     * @return      PersistentCapableIfc
     * @exception   DatabaseIOException
     */
    PersistentCapableIfc lookup(String name)
        throws DatabaseIOException
    {
        PersistentCapableIfc result = null;
        try
        {
            ObjectId oid = (ObjectId)getNamedRootTable().get(name);
            if (oid != null)
            {
                // check to see if the object has already been cached locally
                // if it has then return it from the cache, otherwise read it 
                // in from the database
                result = (PersistentCapableIfc)locallyAccessedObjects_.get(oid);
                if (result == null)
                {
                    result = (PersistentCapableIfc)database_.readObject(
                        oid.getId());
                    if (result != null)
                    {
                        // add it to the cache for future access
                        locallyAccessedObjects_.put(oid, result);
                    }
                    else
                    {
                        // this should never happen and if it does there is an
                        // inconsistency in the database
                        throw new DatabaseIOException(
                            "Could not find the root object " + name);
                    }
                }
            }
        }
        catch (Exception exception)
        {
            throw new DatabaseIOException("lookup failed " + exception);
        }

        return result;
    }


    /**
     * Create the specified persistent object in the database. and place it 
     * in the cache. If there is a problem creating the object then throw 
     * a DatabaseIOException.
     * <p>
     * If all is successful, it initialises the object id and version and 
     * inserts a transaction log entry
     *
     * @param           object              object to create
     * @exception       DatabaseIOException
     */
    void createObject(PersistentCapableIfc object)
        throws DatabaseIOException
    {
        if (object != null)
        {
            try
            {
                // must insert prior to acquiring since the acquire lock
                // requires the id to be set
                database_.insert(object);
                acquireObjectLock(object, 10);
    
                // create a transaction log entry and add it to the 
                // set of locally accessed objects
                objectTranLog_.put(object.getObjectId(), new TranLogEntry(
                    TranLogEntry.CREATED, object));
                locallyAccessedObjects_.put(object.getObjectId(), object);
            }
            catch (FailedToAcquireLockException exception)
            {
                throw new DatabaseIOException("Failed to acquire lock on "+
                    object.getObjectId());
            }
            catch (FailedToCreateLockException exception)
            {
                throw new DatabaseIOException("Failed to create lock on "+
                    object.getObjectId());
            }
        }
        else
        {
            throw new DatabaseIOException("Specified object is null");
        }
    }

    /**
     * Delete the specified object from the database. If the object does not
     * exist then throw the DatabaseIOException.
     * <p>
     * If it is successful then add transaction log entry.
     *
     * @param           object          object to delete
     * @exception       DatabaseIOException
     */
    void deleteObject(PersistentCapableIfc object)
        throws DatabaseIOException
    {
       if ((object != null) &&
           (object.getObjectId().getId() > 0))
       {
           // valid object with object identity
           if (locallyAccessedObjects_.get(object.getObjectId()) != null)
           {
               // the object was created in this session. Remove from the 
               // local cache and update the tran log entry for that object
               locallyAccessedObjects_.remove(object.getObjectId());
               locallyDeletedObjects_.put(object.getObjectId(), object);
               objectTranLog_.put(object.getObjectId(), new TranLogEntry(
                    TranLogEntry.DELETED, object)); 
           }
           else if (database_.readObject(object.getObjectId().getId()) != null)
           {
               // the object exists in the database need to mark it for 
               // deletion
               try
               {
                   acquireObjectLock(object, 10);
                   locallyDeletedObjects_.put(object.getObjectId(), object);
                   objectTranLog_.put(object.getObjectId(), new TranLogEntry(
                       TranLogEntry.DELETED, object));
               }
               catch (Exception exception)
               {
                   throw new DatabaseIOException("deleteObject failed on " +
                        object.getObjectId() + " - failed to acquire lock");
               }
           }
           else
           {
               throw new DatabaseIOException("Failed to delete object " +
                    object.getObjectId());
           }
       }
       else
       {
           throw new DatabaseIOException(
               "Failed to delete null or malformed object");
       }
    }
     
    /**
     * Update the specified persistent object. If the object doesnot exist or
     * if the object has been previously marked for deletion then throw the 
     * DatabaseIOException
     *
     * @param       object              persistent object to delete
     * @exception   DatabaseIOException
     */
    void updateObject(PersistentCapableIfc object)
        throws DatabaseIOException
    {
        if (object != null)
        {
            // check to see if the object has already been deleted. 
            if (locallyDeletedObjects_.get(object.getObjectId()) != null)
            {
                // If it has then throw an exception indicating that it cannot
                // be updated.
                throw new  DatabaseIOException("Cannot update a deleted object "
                    + object.getObjectId());
            }
            else if (locallyAccessedObjects_.containsKey(object.getObjectId()))
            {
                // the object was created in this transaction so update the 
                // trnasaction log. 
                locallyAccessedObjects_.put(object.getObjectId(), object);
                objectTranLog_.put(object.getObjectId(), new TranLogEntry(
                    TranLogEntry.UPDATED, object));
            }
            else
            {
                // the object may have not previously been touched in this
                // transaction. acquire a lock on this object snd then
                // check to ensure that it still exists and has not been
                // modified by any other transaction
                try
                {
                    acquireObjectLock(object, 10);
                    objectTranLog_.put(object.getObjectId(), new TranLogEntry(
                        TranLogEntry.UPDATED, object));
                }
                catch (Exception exception)
                {
                    throw new DatabaseIOException("updateObject failed on " +
                        object.getObjectId() + " " + exception);
                }
            }
        }
        else
        {
           throw new DatabaseIOException(
               "Cannot specify null object in updateObject");
        }
    }

    /**
     * Retrieve the persistent object associated with the id. If the object
     * does not exist then throw a DatabaseIOException, It first checks that
     * its identity is valid and then to ensures that it either hasn't been
     * deleted or that it isn't cached. If it has been deleted then throw an
     * DatabasEIOException. If ti has been cached then return the cached copy.
     * If it is neither then retrieve it from the database
     *
     * @param       id          identity of object to retrieve
     * @return      PersistentCapableIfc
     * @exception   DatabaseIOException
     */
    public PersistentCapableIfc retrieveObject(long id)
        throws DatabaseIOException
    {
        PersistentCapableIfc object = null;

        if (id > 0L)
        {
            ObjectId oid = new ObjectId(id);

            // check to see if the object has already been deleted. 
            if (locallyDeletedObjects_.get(oid) != null)
            {
                // If it has then throw an exception indicating that it cannot
                // be updated.
                throw new  DatabaseIOException("Cannot retrieve a deleted object "
                    + oid);
            }
            else if (locallyAccessedObjects_.containsKey(oid))
            {
                // the object was created in this transaction so retrieve it 
                // from the cache
                object = (PersistentCapableIfc)locallyAccessedObjects_.get(oid);
            }
            else
            {
                // otherwise get it from the database and cache it locally
                object = (PersistentCapableIfc)database_.readObject(id);
                if (object != null)
                {
                    locallyAccessedObjects_.put(oid, object);
                }
            }
        }

        return object;
    }

    
    /**
     * Ther begin operation is very simply it simly calls clear
     */
    void begin()
    {
        clear();
    }

    /**
     * The cache needs to be committed to the store. Iterate through all the 
     * log entries in the root and object transaction log andftake appropriate 
     * actions.
     * <p>
     * If there is a problem committing the cache then this is a fatal error. 
     * At the moment simply log a fatal message and throw a DatabaseIOException
     *
     * @exception       DatabaseIOException
     */
    void commit()
        throws DatabaseIOException
    {
        try
        {
            Enumeration tranlog = null;

            // enumerate through the object transaction log
            tranlog = objectTranLog_.keys();
            while (tranlog.hasMoreElements())
            {
                ObjectId oid = null;
                try
                {
                    oid = (ObjectId)tranlog.nextElement();
                    TranLogEntry entry = (TranLogEntry)objectTranLog_.get(oid);
                    if ((entry.type_ == TranLogEntry.UPDATED) ||
                        (entry.type_ == TranLogEntry.CREATED))
                    {
                        // update the entry in the database
                        // the root object has been created. Ensure that
                        // the version of the object has not changed
                        // and then update it
                        PersistentCapableIfc original_obj = 
                            (PersistentCapableIfc)database_.readObject(
                                entry.object_.getObjectId().getId());
                        if (original_obj.getObjectVersion().equals(
                            entry.object_.getObjectVersion()))
                        {
                            // yes we can update it since it hasn't been
                            // touched since we had access to it. Don't
                            // forget to increment the version num
                            long version = entry.object_.getObjectVersion(
                                ).getVersion();
                            entry.object_.getObjectVersion().setVersion(
                                ++version);
                            database_.update(entry.object_);
                        }
                        else
                        {
                            // there has been a commit by a previous
                            // transaction
                            throw new DatabaseIOException("Failed to commit " +
                                entry.object_ + " since version num. differs" +
                                original_obj.getObjectVersion() + " with " +
                                entry.object_.getObjectVersion());
                        }
                    }
                    else if (entry.type_ == TranLogEntry.DELETED)
                    {
                        // remove the entry from the database
                        database_.delete(entry.object_);
                    }
                    else
                    {
                        // log an error
                        _log.fatal("The TranLogEntry type " +
                            entry.type_ + " is not supported");
                    }
                }
                catch (Exception exception)
                {
                    // catch and log the error. We have an database integrity 
                    // problems here don't know hoe to handle them at the
                    // moment log a fatal message
                    throw new DatabaseIOException("Failed to commit " + 
                        exception);
                }
            }

            // if the named root table has been modified then replace it
            // with the local copy.
            if (modifiedRootTable_)
            {
                try
                {
                    database_.setNamedRootTable(rootTable_);
                }
                catch (IOException exception)
                {
                    _log.fatal("FAILED TO UPDATE NAMED ROOT TABLE");
                    throw new DatabaseIOException(exception.toString());
                }
            }
        }
        finally
        {
            // clear the cache regardless
            releaseLocalLocks();
            clear();
        }
    }
     
    /**
     * The cache needs to be aborted. Iterate through all the log entries in the
     * root and object transaction log and take appropriate actions.
     * <p>
     * If there is a problem committing the cache then this is a fatal error. 
     * At the moment simply log a fatal message and throw a DatabaseIOException
     *
     * @exception       DatabaseIOException
     */
    void  abort()
        throws DatabaseIOException
    {
        try
        {
            Enumeration tranlog = null;

            // enumerate through the object transaction log
            tranlog = objectTranLog_.keys();
            while (tranlog.hasMoreElements())
            {
                ObjectId oid = null;
                try
                {
                    oid = (ObjectId)tranlog.nextElement();
                    TranLogEntry entry = (TranLogEntry)objectTranLog_.get(oid);
        
                    if (entry.type_ == TranLogEntry.CREATED)
                    {
                        // remove the object from the database
                        database_.delete(entry.object_);

                        // also check to see if it was a root object
                        // and if so update the remove it from the
                        // namedRootTable_
                    }
                    else if (entry.type_ == TranLogEntry.DELETED)
                    {
                        // do nothing, deleted entry does not need to
                        // be committed
                    }
                    else if (entry.type_ == TranLogEntry.UPDATED)
                    {
                        // do nothing updated entry does not need to 
                        // be committed
                    }
                    else
                    {
                        // log an error
                        _log.fatal("The TranLogEntry type " +
                            entry.type_ + " is not supported");
                    }
                }
                catch (Exception exception)
                {
                    // catch and log the error. We have an database integrity 
                    // problems here don't know hoe to handle them at the
                    // moment log a fatal message
                    _log.fatal("COMMIT FALIED on OID " + oid.getId()
                        + " with exception " + exception);
                    throw new DatabaseIOException("Failed to commit " + 
                        exception);
                }
            }
        }
        finally
        {
            // clear the cache regardless
            releaseLocalLocks();
            clear();
        }
    }
     

    /**
     * Clear the cache. This will clear both the root and object caches
     * unconditionally. It does not check locks or whether created objects
     * need to be deleted. It only provides the mechanism.
     */
    void clear()
    {
        objectTranLog_.clear();
        locallyAccessedObjects_.clear();
        locallyDeletedObjects_.clear();
        sessionLocks_.clear();
        rootTable_ = null;
        modifiedRootTable_ = false;
    }

    /**
     * Acquire a lock for the specified persistent object. If the lock
     * cannot be acquired within the specified tim (in seconds) then 
     * throw the FailedToAcquireLockException
     *
     * @param           object          object to acquire lock on
     * @param           int             timeout in seconds
     * @return          Lock            the acquire lock
     * @exception       FailedToAcquireLockException
     */
    Lock acquireObjectLock(PersistentCapableIfc object, int timeout)
        throws FailedToAcquireLockException, FailedToCreateLockException
    {
        Lock obj_lock = null;

        if (object != null)
        {
            // see if there is a lock currently allocated on this
            // object, if not then attempt to create one.
            obj_lock = getLock(object);
            if (obj_lock == null)
            {
                // creating a lock can still throw failed to acquire lock
                // exception.
                obj_lock = createLock(object);
            }

            // at this point a lock exists. now try to acquire it, but before
            // doing so check to see that this transaction does not already own
            // a lock on it.
            try
            {
                if (!obj_lock.hasLock(this))
                {
                    obj_lock.acquire(this, timeout);
    
                    // add the lock to the list of lock locks if it hasn't yet
                    // acquired
                    sessionLocks_.addElement(obj_lock);
                }
            }
            catch (Exception exception)
            {
                // rethrow the exception to the top level as a 
                // FailedToAcquireLockException exception
                throw new FailedToAcquireLockException(
                    "Failed to acquire lock on " + object.toString() + 
                    " because " + exception);
            }

            // we have the lock now. We should update the object so that 
            // the version number reflects the last version number in the
            // database.
            try {
                PersistentCapableIfc original_obj = 
                    (PersistentCapableIfc)database_.readObject(
                        object.getObjectId().getId());
                if (original_obj != null) {
                    object.getObjectVersion().setVersion(
                        original_obj.getObjectVersion().getVersion());
                }
            } catch (DatabaseIOException exception) {
                throw new FailedToAcquireLockException(
                    "Failed to acquire lock on " + object.toString() + 
                    " because " + exception);
            }
        }
        else
        {
            _log.error("Trying to acquireLock on null object");
            throw new FailedToAcquireLockException(
                "Null object specfied acquireLock");
        }

        return obj_lock;
    }

    /**
     * Acquire a lock on the root object. This is a very expensive operation
     * since it locks out all other threads. If it cannot acquire the lock in
     * the specified time then the FailedToAcquireLockException is raised.
     * <p>
     * The FailedToCreateLockException is raised if a lock for this object 
     * could not be created
     *
     * @return          Lock
     * @exception      FailedToAcquireLockException
     * @exception      FailedToCreateLockException
     */
    Lock acquireRootLock()
        throws FailedToAcquireLockException, FailedToCreateLockException
    {
        return acquireObjectLock(getNamedRootTable(), 5000);
    }
    
    /**
     * Return a cached instance of the named root table. If you need to obtain
     * an exclusive copy of this table then you must call acquireRootLock first
     * <p>
     * Always fetch this for the time being
     *
     * @return          PMDRootTable
     */
    synchronized PMDRootTable getNamedRootTable()
    {
        if (rootTable_ == null)
        {
            rootTable_ = (PMDRootTable)database_.getNamedRootTable().clone();
        }

        return rootTable_;
    }

    /**
     * Return the lock associated with the specified PersistentCapableIfc 
     * object. If none exists then return null. This is a synchronized method
     * since the lock cache manages all locks for all sessions
     *
     * @param       object          object to query on
     * @return      Lock            object lock or null if one does not exist
     */
    protected Lock getLock(PersistentCapableIfc object)
    {
        Lock result = null;

        synchronized(locks__)
        {
            if (locks__.containsKey(object.getObjectId()))
            {
                result = (Lock)locks__.get(object.getObjectId());
            }
        }

        return result;
    }

    /**
     * Create a lock on the specified object. It first checks to ensure that a
     * lock really doesn't exist before creating one. If it does exist then it
     * will throw FailedToCreateLockException
     *
     * @param       object              create a lock on this object
     * @return      Lcok                created lock
     * @exception   FailedToCreateLockException
     */
    protected Lock createLock(PersistentCapableIfc object)
        throws FailedToCreateLockException
    {
        Lock result = null;

        synchronized(locks__)
        {
            if (locks__.containsKey(object.getObjectId()))
            {
                // a lock already exists throw an exception
                throw new FailedToCreateLockException(
                    "Lock already exists on object id " + 
                    object.getObjectId());
            }
            else
            {
                Lock lock = new Lock(object);
                locks__.put(object.getObjectId(), lock);

                result = lock;
            }
        }

        return result;
    }

    /**
     * Delete the lock associated with the specified transaction if there are
     * no other transactions waiting on it.
     *
     * @param           object          object, whose lock we want to query
     */
    protected void deleteLock(PersistentCapableIfc object)
    {
        synchronized(locks__)
        {
            this.deleteLock((Lock)locks__.get(object.getObjectId()));
        }
    }

    /**
     * Delete the specified lock if there are no other transactions waiting for
     * it.
     *
     * @param           lock            lock to delete
     */
    protected void deleteLock(Lock lock)
    {
        if ((lock != null) &&
            (!lock.anyTxWaiting()))
        {
            locks__.remove(lock.getObject().getObjectId());
        }
    }

    /**
     * Go through the locks held by this transaction and release them so that
     * waiting transactions can continue their work
     */
    protected void releaseLocalLocks()
    {
        Enumeration locks = sessionLocks_.elements();
        while (locks.hasMoreElements())
        {
            Lock lock = (Lock)locks.nextElement();
            lock.release(this);
            this.deleteLock(lock);
        }

        // clear all the local session locks.
        sessionLocks_.clear();
    }

    /**
     * The protected static inner class is used to define an object transaction 
     * log entry. These are entries relared to the database objects
     */
    protected static class TranLogEntry
    {
        TranLogEntry(int type, PersistentCapableIfc object)
        {
            type_ = type;
            object_ = object;
        }

        long type_;
        PersistentCapableIfc object_;


        final public static int CREATED = 1;
        final public static int DELETED = 2;
        final public static int UPDATED = 3;
    }
     
     
     
    private transient PageManagedDatabase database_ = null;
    private transient Hashtable objectTranLog_ = new Hashtable();
    private transient Hashtable locallyAccessedObjects_ = new Hashtable();
    private transient Hashtable locallyDeletedObjects_ = new Hashtable();
    private transient Vector sessionLocks_ = new Vector();
    private transient PMDRootTable rootTable_ = null;
    private transient boolean modifiedRootTable_ = false;

    // The lock cache maintains a list of locks for all sessions. All locks are
    // exclusive and currently there is no distinction between a write or a
    // read lock (something to consider for improved concurrency). Would use 
    // weak references if they weren't so expensive
    static private transient Hashtable locks__ = new Hashtable();

}

