/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 1999 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: PMDTreeMap.java,v 1.3 2001/06/12 07:11:35 jima Exp $
 *
 * Date         Author  Changes
 * 04/04/2000   jima    Created
 */
package org.exolab.core.database.recman;

import java.util.Comparator;
import java.util.TreeMap;
import java.util.Enumeration;
import java.util.Collection;
import java.util.Set;
import java.util.SortedMap;
import java.io.Serializable;
import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.util.Hashtable;
import java.util.Enumeration;
import org.exolab.core.foundation.PersistentObject;
import org.exolab.core.foundation.TreeMapIfc;

/**
 * This is an implementation of the TreeMapIfc for the PMD datastore. The 
 * implementation currently uses java.util.TreeMap but this will be changed
 * in a subsequent release to provide a more concurrent capable container and
 * take advantages of this database's features.
 *
 * @version     $Revision: 1.3 $ $Date: 2001/06/12 07:11:35 $
 * @author      <a href="mailto:jima@exoffice.com">Jim Alateras</a>
 * @see         org.exolab.core.foundation.TreeMapIfc
 * @see         org.exolab.core.foundation.PersistentObject
 **/
public class PMDTreeMap
    extends PersistentObject
    implements TreeMapIfc, Externalizable
{
    /**
     * Default constructor simply instantiates an instance of a TreeMap
     */
    public PMDTreeMap()
    {
        map_ = new TreeMap();
    }
    
    /**
     * Construct an instance of a TreeMap with the specified comparator. The
     * comparator is used to sort the elements in the map. If a null comparator
     * is specified then natural order pervails.
     *
     * @param       comparator          used for ordering
     */
    public PMDTreeMap(Comparator comparator)
    {
        map_ = new TreeMap(comparator);
    }

    /**
     * Returns the number of key-value mappings in this map.
     *
     * @return                  number of mappings in this map.
     */
    public int size()
    {
        return map_.size();
    }

    /**
     * Test whether a particular key is mapped to an object in the container. 
     * If the key cannot be compared with the jeys in the map the ClassCast-
     * Exception is thrown. If the key is null and the map uses natural 
     * ordering or its comparator does not tolerate null raise the NullPointer-
     * Exception
     *
     * @param           key         key whose value to find
     * @return          boolean     true if the key is mapped
     * @throws          ClassCastException
     * @throws          NullPointerException
     */
    public boolean containsKey(Object key)
    {
        return map_.containsKey(key);
    }

    /**
     * Returns true if this map maps one or more keys to the specified value.  
     * More formally, returns true if and only if this map contains at least
     * one mapping to a value <tt>v</tt> such that <tt>(value==null ? 
     * v==null : value.equals(v))</tt>.  This operation will probably require
     * time linear in the Map size for most implementations of Map.
     *
     * @param       value               test for this value
     * @return      boolean             true if object is present
     */
    public boolean containsValue(Object value)
    {
        return map_.containsValue(value);
    }

    /**
     * Returns the value associated with the specified key. If the key cannot
     * be compared with the keys then the ClassCastException. If the key is 
     * null and this map uses natural ordering or its comparator does not tol-
     * erate null keys throw the NullPointerException
     *
     * @param           key         key whose value to find
     * @return          boolean     true if the key is mapped
     * @throws          ClassCastException
     * @throws          NullPointerException
     */
    public Object get(Object key)
    {
        return map_.get(key);
    }

    /**
     * Returns the comparator used to order this map or null if this map uses
     * natural ordering
     *
     * @return      Comparator
     */
    public Comparator comparator()
    {
        return map_.comparator();
    }

    /**
     * Returns the first (lowest) key currently in this sorted map.It throws 
     * NoSuchElementException if the element does not exist.
     *
     * @return      Object
     * @throws      NoSuchElementException 
     */
    public Object firstKey()
    {
        return map_.firstKey();
    }

    /**
     * Returns the last (highest) key currently in this sorted map.It throws 
     * NoSuchElementException if the element does not exist.
     *
     * @return      Object
     * @throws      NoSuchElementException 
     */
    public Object lastKey()
    {
        return map_.lastKey();
    }

    /**
     * Associates the specified value with the specified key in this map.
     * If the map previously contained a mapping for this key, the old
     * value is replaced.
     * <p>
     * If the key cannot be compared with other keys in the map it will
     * throw the ClassCastException. Similarly, if the key is null and the 
     * map uses natural ordering or its comparator cannot tolerate null keys
     * throw the NullPointerException
     *
     * @param       key         key object
     * @param       value       value object
     * @throws      ClassCastException 
     * @throws      NullPointerException 
     */
    public Object put(Object key, Object value)
    {
        return map_.put(key, value);
    }

    /**
     * Removes the mapping for this key from this TreeMap if present.If the 
     * key cannot be compared to other values in the map then thorw the Class-
     * castExcpetion. Similarly, ifthe key is null and the key uses natural 
     * order or its comparator does not tolerate null keys the NullPointer-
     * Exception is raised.
     *
     * @param       key         key whose value to remove
     * @return      Object      remove value
     * @throws      ClassCastEzception
     * @throws      NullPointerException
     */
    public Object remove(Object key)
    {
        return map_.remove(key);
    }

    /**
     * Removes all mappings from this TreeMap.
     */
    public void clear()
    {
        map_.clear();
    }

    /**
     * Returns a Set view of the keys contained in this map.  The set's
     * iterator will return the keys in ascending order.  The map is backed by
     * this <tt>TreeMap</tt> instance, so changes to this map are reflected in
     * the Set, and vice-versa.  The Set supports element removal,
     *
     * @return      Set         view of the keys contained in this TreeMap.
     */
    public Set keySet()
    {
        return map_.keySet();
    }

    /**
     * Test is the map is empty
     *
     * @return      boolean     true if the map is empty and false otherwise
     */
	public boolean isEmpty()
    {
        return map_.isEmpty();
    }

	/**
     * Return an enumeration of all the elements in the map
     *
     * @return      Enumeration
     */
    public synchronized Collection elements()
    {
        return map_.values();
    }

	/**
	 * Returns a view of the portion of this map whose keys are greater than 
	 * or equal to fromKey. 
	 *
	 * @param id The key to compare to.
	 * @return a view of the portion of this map whose keys are greater 
	 *	than or equal to fromKey.
	 *
	 */
	public SortedMap tailMap(Object key)
	{
		return map_.tailMap(key);
	}


	/**
	 * Returns the internal TreeMap
	 *
	 * @return The internal TreeMap used.
	 *
	 */
	public SortedMap sortedMap()
	{
		return map_;
	}
	

    // implementation of Externalizable.writeExternal
    public void writeExternal(ObjectOutput stream)
        throws IOException
    {
        stream.writeLong(serialVerionUID);
        stream.writeObject(map_);
        super.writeExternal(stream);
    }

    // implementation of Externalizable.writeExternal
    public void readExternal(ObjectInput stream)
        throws IOException, ClassNotFoundException
    {
		long version = stream.readLong();
		if (version == serialVerionUID) {	
            map_ = (TreeMap)stream.readObject();
            super.readExternal(stream); 
        } else {
            throw new IOException("PMDTreeMap with version " +
                version + " is not supported.");
        }
    }

    static final long serialVerionUID = 1;
    private TreeMap map_ = null;
}


