/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 1999 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: PageManagedDatabase.java,v 1.11 2003/06/09 06:28:39 tanderson Exp $
 *
 * Date         Author  Changes
 * 1/13/2000    jima    Created
 */
package org.exolab.core.database.recman;

import java.io.IOException;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.ObjectInputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.ObjectOutputStream;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import com.cdegroot.db.recman.RecordManager;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.exolab.core.foundation.DatabaseIfc;
import org.exolab.core.foundation.DatabaseIOException;
import org.exolab.core.foundation.FailedToOpenDatabaseException;
import org.exolab.core.foundation.FailedToCreateDatabaseException;
import org.exolab.core.foundation.ObjectNameExistsException;
import org.exolab.core.foundation.ObjectNameNotFoundException;
import org.exolab.core.foundation.NoAvailableRootException;
import org.exolab.core.foundation.PersistentCapableIfc;
import org.exolab.core.foundation.PersistentObject;
import org.exolab.core.foundation.LruObjectCache;


/**
 * This is a wrapper for PageManager supplied by Cees de Groot. This will be
 * used to support Serailized data storage in a page-based file.
 *
 * @version     $version$
 * @author      <a href="mailto:jima@exoffice.org">Jim Alateras</a>
 * @see         Database
 **/
public class PageManagedDatabase 
    implements DatabaseIfc
{
    private static final Log _log = LogFactory.getLog(PageManagedDatabase.class);

    /**
     * The constructor for a paged managed database takes a string to
     * identify the name of the database. 
     *
     * @param       name            name of the database
     */
    public PageManagedDatabase(String name)
        throws FailedToCreateDatabaseException
    {
        try
        {
            // make sure we can acquire the lock before we attempt to 
            // use the database. If we cannot acquire the lock then 
            // it either means that someone already owns the database or
            // a process that previously owned the lock has abnormally 
            // temrinated and did not release the lock. Force the lock to
            // be released on VM termination.
            // NOTE: That this lock can also be released when the database
            // is closed.
            LockManager.getInstance().acquire(name);
            name_ = name;
            db_ = new RecordManager(name);
            //db_.disableTransactions();
            opened_ = true;

            // fault in the named root table
            getNamedRootTable();
        }
        catch(IOException exception)
        {
            throw new FailedToCreateDatabaseException("Failed to create database "
                + name + " b/c" + exception);
        }
    }

    // implementation of DatabaseIfc.open
    public void open()
        throws FailedToOpenDatabaseException
    {
        if (!opened_)
        {
            throw new FailedToOpenDatabaseException("Failed to open db " +
                name_);
        }
    }

    // implementation of DatabaseIfc.close
    public void close()
    {
        try
        {
            db_.close();
            opened_ = false;
            LockManager.getInstance().release(name_);
        }
        catch (IOException exception)
        {
            _log.error("Failed to close db", exception);
        }
    }

    // implementation of DatabaseIfc.isOpen
    public boolean isOpen()
    {
        return opened_;
    }

    // implementation of DatabaseIfc.getRootNames
    public Enumeration getRootNames()
    {
        int root_count = db_.getRootCount();
        Vector roots = new Vector();

        synchronized(namedRootTable_)
        {
            Enumeration names = namedRootTable_.names();
            while (names.hasMoreElements())
            {
                String name = (String)names.nextElement();
                roots.addElement(name);
            }
        }

        return roots.elements();
    }

    // implementation of DatabaseIfc.bind
    public void bind(String name, PersistentCapableIfc object)
        throws ObjectNameExistsException, DatabaseIOException
    {
        this.setRootObject(name, object);
    }

    // implementation of DatabaseIfc.unbind
    public void unbind(String name)
        throws ObjectNameNotFoundException, DatabaseIOException
    {
        this.removeRootObject(name);
    }

    // implementation of DatabaseIfc.lookup
    public PersistentCapableIfc lookup(String name)
    {
        return (PersistentCapableIfc)this.getRootObject(name);
    }

    /**
     * Lookup the object specified by name. If the object does not exists
     * then return null.
     *
     * @param       name            lookup name
     * @return      PersistentCapableIfc          
     */
    public Object getRootObject(String name)
    {
        Object instance = null;

        if (namedRootTable_.containsName(name))
        {
            instance = (Object)namedRootTable_.get(name);
        }

        return instance;
    }

    /**
     * Store the specified object with the passed in name. This
     * is stored as a root objeect. Root objects are special database
     * entry points
     * <p>
     * If the name is already bound to another object then throw the
     * ObjectNameExistsException exception. If there is any other error
     * raise the  DatabaseIOException exception
     *
     * @param           name            name of the object
     * @param           object          object to bind
     * @exception       ObjectNameExistsException
     * @exception       DatabaseIOException
     */  
    public void setRootObject(String name, Object object)
        throws ObjectNameExistsException, DatabaseIOException
    {
        try
        {
            synchronized(namedRootTable_)
            {
                if (getRootObject(name) != null)
                {
                    throw new ObjectNameExistsException("The root named " +
                        name + " already exists");
                }
                else
                {
                    _log.debug("PMD RootObject " + object);
                    namedRootTable_.put(name, object);
                    updateNamedRootTable();
                }
            }
        }
        catch(IOException exception)
        {
            throw new DatabaseIOException("Error in bind for name " +
                name + " " + exception);
        }
    }

    /**
     * Remove the binding between the specified name and the object.If 
     * the nroot name does not exist then raise the ObjectNameNotFound=
     * Exception.
     * <p>
     * If there is any other problem removing the binding then throw the
     * DatabaseIOException exception
     *
     * @param           name            name of the object
     * @exception       ObjectNameNotFoundException
     * @exception       DatabaseIOException
     */
    public void removeRootObject(String name)
        throws ObjectNameNotFoundException, DatabaseIOException
    {
        try
        {
            synchronized(namedRootTable_)
            {
                if (getRootObject(name) == null)
                {
                    throw new ObjectNameNotFoundException("Failed to unbind " +
                        name);
    
                }
                else
                {
                    namedRootTable_.remove(name);
                    updateNamedRootTable();
                }
            }
        }
        catch(IOException exception)
        {
            throw new DatabaseIOException("Error in unbind for name " +
                name + " " + exception);
        }
    }

    /**
     * Check if the specified object is already bound in the database
     *
     * @param       name            lookup name
     * @return      boolean         true if name is bound top object
     */
    public boolean isBound(String name)
    {
        return (lookup(name) != null);
    }
    

    /**
     * Delete the object with the specified id from the database. If it fails
     * to delete the object then throw the DatabaseIOException exception
     *
     * @param       id              id of object to destroy
     * @exception   DatabaseIOException
     */
    public void deleteObject(long id)
        throws DatabaseIOException
    {
        try
        {
            db_.delete(id);

            // don't forget to remove it from the cache if it is defined there
            if (objectCache_ != null)
            {
                objectCache_.remove(new Long(id));
            }
        }
        catch (IOException exception)
        {
            throw new DatabaseIOException("Failed to delete object with id " +
                id + ". Reason : " + exception.toString());
        }
    }

    /**
     * This method will pre-allocate an obect of a particular size. It will 
     * initially fill the block in will nulls. Thisi sone way to overcome some of
     * the deficiencies of the JDBM, until a better block merging algorithm is
     * available. 
     * <p>
     * If it fails to delete the object then throw the DatabaseIOException 
     * exception.
     *
     * @param       int             sizeof the object
     * @return      long            identity of the object
     * @exception   DatabasEIOException
     */
    public long allocateObject(int size)
        throws DatabaseIOException
    {
        return this.insertBytes(new byte[size]);
    }
     
    // Implementation of DatabaseIfc.insert
    public void insert(PersistentCapableIfc object)
        throws DatabaseIOException
    {
        long id = this.insertObject(object);
        object.getObjectId().setId(id);
        this.updateObject(id, object);
    }

    // Implementation of DatabaseIfc.delete
    public void delete(PersistentCapableIfc object)
        throws DatabaseIOException
    {
        this.deleteObject(object.getObjectId().getId());
    }


    // Implementation of DatabaseIfc.update
    public void update(PersistentCapableIfc object)
        throws DatabaseIOException
    {
        this.updateObject(object.getObjectId().getId(), object);
    }

    /**
     * Insert the boolean entity into the database and return the id of 
     * the stored record.
     *
     * @param        value           boolean value to store
     * @return       long            id of entity
     * @exception    DatabaseIOException
     */
    public long insertBoolean(boolean value)
        throws DatabaseIOException
    {
        return writeBoolean(UNASSIGNED, value);
    }

    /**
     * Insert the byte entity into the database and return the id of 
     * the stored record.
     *
     * @param        value           byte value to store
     * @return       long            id of entity
     * @exception    DatabaseIOException
     */
    public long insertByte(byte value)
        throws DatabaseIOException
    {
        return writeByte(UNASSIGNED, value);
    }

    /**
     * Insert the short entity into the database and return the id of 
     * the stored record.
     *
     * @param       value           short value to store
     * @return      long            id of entity
     * @exception   DatabaseIOException
     */
    public long insertShort(short value)
        throws DatabaseIOException
    {
        long result = 0L;

        try
        {
            result = writeShort(UNASSIGNED, value);
        }
        catch(Exception exception)
        {
            throw new DatabaseIOException("insertShort failed - " +
                exception.toString());
        }

        return result;
    }

    /**
     * Insert the char entity into the database and return the id of 
     * the stored record.
     *
     * @param        value           char value to store
     * @return       long            id of entity
     * @exception    DatabaseIOException
     */
    public long insertChar(char value)
        throws DatabaseIOException
    {
        return writeChar(UNASSIGNED, value);
    }

    /**
     * Insert the int entity into the database and return the id of 
     * the stored record.
     *
     * @param        value           int value to store
     * @return       long            id of entity
     * @exception    DatabaseIOException
     */
    public long insertInt(int value)
        throws DatabaseIOException
    {
        return writeInt(UNASSIGNED, value);
    }

    /**
     * Insert the long entity into the database and return the id of 
     * the stored record.
     *
     * @param        value           long value to store
     * @return       long            id of entity
     * @exception    DatabaseIOException
     */
    public long insertLong(long value)
        throws DatabaseIOException
    {
        return writeLong(UNASSIGNED, value);
    }

    /**
     * Insert the float entity into the database and return the id of 
     * the stored record.
     *
     * @param        value           float value to store
     * @return       long            id of entity
     * @exception    DatabaseIOException
     */
    public long insertFloat(float value)
        throws DatabaseIOException
    {
        return writeFloat(UNASSIGNED, value);
    }

    /**
     * Insert the double entity into the database and return the id of 
     * the stored record.
     *
     * @param        value           double value to store
     * @return       long            id of entity
     * @exception    DatabaseIOException
     */
    public long insertDouble(double value)
        throws DatabaseIOException
    {
        return writeDouble(UNASSIGNED, value);
    }

    /**
     * Insert the String entity into the database and return the id of 
     * the stored record.
     *
     * @param        value           bytes to store
     * @return       long            id of entity
     * @exception    DatabaseIOException
     */
    public long insertBytes(byte[] value)
        throws DatabaseIOException
    {
        return writeBytes(UNASSIGNED, value);
    }

    /**
     * Insert the String entity into the database and return the id of 
     * the stored record.
     *
     * @param        value           String value to store
     * @return       long            id of entity
     * @exception    DatabaseIOException
     */
    public long insertString(String value)
        throws DatabaseIOException
    {
        return writeString(UNASSIGNED, value);
    }

    /**
     * Insert the object entity into the database and return the id of 
     * the stored record.
     *
     * @param        value           object value to store
     * @return       long            id of entity
     * @exception    DatabaseIOException
     */
    public long insertObject(Object value)
        throws DatabaseIOException
    {
        return writeObject(UNASSIGNED, value);
    }

    /**
     * Insert the boolean entity into the database and return the id of 
     * the stored record.
     *
     * @param        id              record id of entity to update
     * @param        value           boolean value to store
     * @return       long            id of entity
     * @exception    DatabaseIOException
     */
    public long updateBoolean(long id, boolean value)
        throws DatabaseIOException
    {
        return writeBoolean(id, value);
    }

    /**
     * Insert the byte entity into the database and return the id of 
     * the stored record.
     *
     * @param       id              record id of entity to update
     * @param       value           byte value to store
     * @return      long            id of entity
     * @exception   DatabaseIOException
     */
    public long updateByte(long id, byte value)
        throws DatabaseIOException
    {
        long result = 0L;

        try
        {
            result =  writeByte(id, value);
        }
        catch (Exception exception)
        {
            throw new DatabaseIOException("updateByte failed - " +
                exception.toString());
        }

        return result;
    }

    /**
     * Insert the short entity into the database and return the id of 
     * the stored record.
     *
     * @param        id              record id of entity to update
     * @param        value           short value to store
     * @return       long            id of entity
     * @exception    DatabaseIOException
     */
    public long updateShort(long id, short value)
        throws DatabaseIOException
    {
        return writeShort(id, value);
    }

    /**
     * Insert the char entity into the database and return the id of 
     * the stored record.
     *
     * @param        id              record id of entity to update
     * @param        value           char value to store
     * @return       long            id of entity
     * @exception    DatabaseIOException
     */
    public long updateChar(long id, char value)
        throws DatabaseIOException
    {
        return writeChar(id, value);
    }

    /**
     * Insert the int entity into the database and return the id of 
     * the stored record.
     *
     * @param        id              record id of entity to update
     * @param        value           int value to store
     * @return       long            id of entity
     * @exception    DatabaseIOException
     */
    public long updateInt(long id, int value)
        throws DatabaseIOException
    {
        return writeInt(id, value);
    }

    /**
     * Insert the long entity into the database and return the id of 
     * the stored record.
     *
     * @param        id              record id of entity to update
     * @param        value           long value to store
     * @return       long            id of entity
     * @exception    DatabaseIOException
     */
    public long updateLong(long id, long value)
        throws DatabaseIOException
    {
        return writeLong(id, value);
    }

    /**
     * Insert the float entity into the database and return the id of 
     * the stored record.
     *
     * @param        id              record id of entity to update
     * @param        value           float value to store
     * @return       long            id of entity
     * @exception    DatabaseIOException
     */
    public long updateFloat(long id, float value)
        throws DatabaseIOException
    {
        return writeFloat(id, value);
    }

    /**
     * Insert the double entity into the database and return the id of 
     * the stored record.
     *
     * @param        id              record id of entity to update
     * @param        value           double value to store
     * @return       long            id of entity
     * @exception    DatabaseIOException
     */
    public long updateDouble(long id, double value)
        throws DatabaseIOException
    {
        return writeDouble(id, value);
    }

    /**
     * Insert the String entity into the database and return the id of 
     * the stored record.
     *
     * @param        id              record id of entity to update
     * @param        value           bytes to store
     * @return       long            id of entity
     * @exception    DatabaseIOException
     */
    public long updateBytes(long id, byte[] value)
        throws DatabaseIOException
    {
        return writeBytes(id, value);
    }

    /**
     * Insert the String entity into the database and return the id of 
     * the stored record.
     *
     * @param        id              record id of entity to update
     * @param        value           String value to store
     * @return       long            id of entity
     * @exception    DatabaseIOException
     */
    public long updateString(long id, String value)
        throws DatabaseIOException
    {
        return writeString(id, value);
    }

    /**
     * Insert the object entity into the database and return the id of 
     * the stored record.
     *
     * @param        id              record id of entity to update
     * @param        value           object value to store
     * @return       long            id of entity
     * @exception    DatabaseIOException
     */
    public long updateObject(long id, Object value)
        throws DatabaseIOException
    {
        return writeObject(id, value);
    }

    /**
     * read the boolean from the database given the specified record id. If it 
     * fails then it throws a DatabaseIOException.
     *
     * @param        id              record identifier
     * @return       boolean         returned value
     * @exception    DatabaseIOException
     */
    public boolean readBoolean(long id)
        throws DatabaseIOException
    {
        boolean value = false;

        try
        {
            byte[] in_buf = db_.fetch(id);
            ByteArrayInputStream byte_stream = new ByteArrayInputStream(in_buf);
            DataInputStream in_stream = new DataInputStream(byte_stream);
            value = in_stream.readBoolean();
        }
        catch(IOException exception)
        {
            throw new DatabaseIOException("readBoolean failed for id " + 
                id + " with " + exception);                
        }
    
        return value;
    }

    /**
     * Read the byte from the database given the specified record id. If it 
     * fails then it throws a DatabaseIOException.
     *
     * @param        id              record identifier
     * @return       byte            returned value
     * @exception    DatabaseIOException
     */
    public byte readByte(long id)
        throws DatabaseIOException
    {
        byte value = 0;

        try
        {
            byte[] in_buf = db_.fetch(id);
            ByteArrayInputStream byte_stream = new ByteArrayInputStream(in_buf);
            DataInputStream in_stream = new DataInputStream(byte_stream);
            value = in_stream.readByte();
        }
        catch(IOException exception)
        {
            throw new DatabaseIOException("readByte failed for id " + 
                id + " with " + exception);                
        }
    
        return value;
    }

    /**
     * Read the short from the database given the specified record id. If it 
     * fails then it throws a DatabaseIOException.
     *
     * @param        id              record identifier
     * @return       short           returned value
     * @exception    DatabaseIOException
     */
    public short readShort(long id)
        throws DatabaseIOException
    {
        short value = 0;

        try
        {
            byte[] in_buf = db_.fetch(id);
            ByteArrayInputStream byte_stream = new ByteArrayInputStream(in_buf);
            DataInputStream in_stream = new DataInputStream(byte_stream);
            value = in_stream.readShort();
        }
        catch(IOException exception)
        {
            throw new DatabaseIOException("readShort failed for id " + 
                id + " with " + exception);                
        }
    
        return value;
    }

    /**
     * Read the char from the database given the specified record id. If it 
     * fails then it throws a DatabaseIOException.
     *
     * @param        id              record identifier
     * @return       char            returned value
     * @exception    DatabaseIOException
     */
    public char readChar(long id)
        throws DatabaseIOException
    {
        char value = '\0';

        try
        {
            byte[] in_buf = db_.fetch(id);
            ByteArrayInputStream byte_stream = new ByteArrayInputStream(in_buf);
            DataInputStream in_stream = new DataInputStream(byte_stream);
            value = in_stream.readChar();
        }
        catch(IOException exception)
        {
            throw new DatabaseIOException("readChar failed for id " + 
                id + " with " + exception);                
        }
    
        return value;
    }

    /**
     * Read the int from the database given the specified record id. If it 
     * fails then it throws a DatabaseIOException.
     *
     * @param        id              record identifier
     * @return       int             returned value
     * @exception    DatabaseIOException
     */
    public int readInt(long id)
        throws DatabaseIOException
    {
        int value = 0;

        try
        {
            byte[] in_buf = db_.fetch(id);
            ByteArrayInputStream byte_stream = new ByteArrayInputStream(in_buf);
            DataInputStream in_stream = new DataInputStream(byte_stream);
            value = in_stream.readInt();
        }
        catch(IOException exception)
        {
            _log.error(exception);
            throw new DatabaseIOException("readInt failed for id " + 
                id + " with " + exception);                
        }
    
        return value;
    }


    /**
     * Read the long from the database given the specified record id. If it 
     * fails then it throws a DatabaseIOException.
     *
     * @param        id              record identifier
     * @return       long            returned value
     * @exception    DatabaseIOException
     */
    public long readLong(long id)
        throws DatabaseIOException
    {
        long value = 0L;

        try
        {
            byte[] in_buf = db_.fetch(id);
            ByteArrayInputStream byte_stream = new ByteArrayInputStream(in_buf);
            DataInputStream in_stream = new DataInputStream(byte_stream);
            value = in_stream.readLong();
        }
        catch(IOException exception)
        {
            throw new DatabaseIOException("readLong failed for id " + 
                id + " with " + exception);                
        }
    
        return value;
    }


    /**
     * Read the float from the database given the specified record id. If it 
     * fails then it throws a DatabaseIOException.
     *
     * @param        id              record identifier
     * @return       float           returned value
     * @exception    DatabaseIOException
     */
    public float readFloat(long id)
        throws DatabaseIOException
    {
        float value = (float)0.0;

        try
        {
            byte[] in_buf = db_.fetch(id);
            ByteArrayInputStream byte_stream = new ByteArrayInputStream(in_buf);
            DataInputStream in_stream = new DataInputStream(byte_stream);
            value = in_stream.readFloat();
        }
        catch(IOException exception)
        {
            throw new DatabaseIOException("readFloat failed for id " + 
                id + " with " + exception);                
        }
    
        return value;
    }

    /**
     * Read the double from the database given the specified record id. If it 
     * fails then it throws a DatabaseIOException.
     *
     * @param        id              record identifier
     * @return       double          returned value
     * @exception    DatabaseIOException
     */
    public double readDouble(long id)
        throws DatabaseIOException
    {
        double value = 0.0;

        try
        {
            byte[] in_buf = db_.fetch(id);
            ByteArrayInputStream byte_stream = new ByteArrayInputStream(in_buf);
            DataInputStream in_stream = new DataInputStream(byte_stream);
            value = in_stream.readDouble();
        }
        catch(IOException exception)
        {
            throw new DatabaseIOException("readDouble failed for id " + 
                id + " with " + exception);                
        }
    
        return value;
    }

    /**
     * Read the byte[] from the database given the specified record id. If it 
     * fails then it throws a DatabaseIOException.
     *
     * @param        id              record identifier
     * @return       byte[]          returned value
     * @exception    DatabaseIOException
     */
    public byte[] readBytes(long id)
        throws DatabaseIOException
    {
        byte[] value = null;

        try
        {
            value = db_.fetch(id);
        }
        catch(IOException exception)
        {
            throw new DatabaseIOException("readByte failed for id " + 
                id + " with " + exception);                
        }
    
        return value;
    }

    /**
     * Read the string from the database given the specified record id. If it 
     * fails then it throws a DatabaseIOException.
     *
     * @param        id              record identifier
     * @return       String          returned value
     * @exception    DatabaseIOException
     */
    public String readString(long id)
        throws DatabaseIOException
    {
        return (String)this.readObject(id);
    }

    /**
     * Read the object from the database given the specified record id. If it 
     * fails then it throws a DatabaseIOException.
     *
     * @param        id              record identifier
     * @return       Object          returned value
     * @exception    DatabaseIOException
     */
    public Object readObject(long id)
        throws DatabaseIOException
    {
        Object value = null;

        try
        {
            // first try and fetch the object from the cache, if one 
            // is already defined
            if (objectCache_ != null)
            {
                value = objectCache_.get(new Long(id));
            }

            // if the value is not in the cache then retrieve it from the
            // database. Subsequently, insert it in the cache
            if (value == null)
            {
                byte[] in_buf = db_.fetch(id);
                ByteArrayInputStream byte_stream = new ByteArrayInputStream(in_buf);
                ObjectInputStream in_stream = new ObjectInputStream(byte_stream);
                value = in_stream.readObject();

                if ((objectCache_ != null) &&
                    (value != null))
                {
                    objectCache_.insert(new Long(id), value);
                }
            }
        }
        catch(Exception exception)
        {
            throw new DatabaseIOException("readObject failed for id " + 
                id + " with " + exception);                
        }
    
        return value;
    }

    /**
     * Set the cache size for the database. If the cases is already configured 
     * then it is cleared and reset to the new size. Setting the sizeo to 0 
     * will remove the cache. If 'count' is less than zero it will simply get 
     * ignored and a warning message will be issued
     *
     * @param       count           size of cache
     */
    public void setObjectCacheSize(int size)
    {
        if (size == 0)
        {
            objectCache_ = null;
        }
        else if (size < 0)
        {
            _log.warn("Can't specify a cache size less than 0");
        }
        else
        {
            objectCache_ = new LruObjectCache(size);
        }
    }
    
    /**
     * Write the boolean to the database and return the id of the written 
     * entity. If the value if id is UNASSIGNED then entity does not currently
     * exist in the database, otherwise the entity exists so we need to update
     * it. If it fails to write then throw a DatabaseIOException 
     *
     * @param        id              record id if avaialable
     * @param        value           boolean value to store
     * @return       long            id of entity
     * @exception    DatabaseIOException
     */
    protected long writeBoolean(long id, boolean value)
        throws DatabaseIOException
    {
        try
        {
            ByteArrayOutputStream out_buf = new ByteArrayOutputStream();
            DataOutputStream out_stream = new DataOutputStream(out_buf);
            out_stream.writeBoolean(value);

            if (id == UNASSIGNED)
            {
                id = db_.insert(out_buf.toByteArray());
            }
            else
            {
                db_.update(id, out_buf.toByteArray());
            }
        }
        catch(IOException exception)
        {
            throw new DatabaseIOException("writeBoolean failed " + 
                exception.toString());                
        }
    
        return id;
    }

    /**
     * Write the byte to the database and return the id of the written 
     * entity. If the value if id is UNASSIGNED then entity does not currently
     * exist in the database, otherwise the entity exists so we need to update
     * it. If it fails to write then throw a DatabaseIOException 
     *
     * @param        id              record id if avaialable
     * @param        value           byte value to store
     * @return       long            id of entity
     * @exception    DatabaseIOException
     */
    protected long writeByte(long id, byte value)
        throws DatabaseIOException
    {
        try
        {
            ByteArrayOutputStream out_buf = new ByteArrayOutputStream();
            DataOutputStream out_stream = new DataOutputStream(out_buf);
            out_stream.writeByte(value);

            if (id == UNASSIGNED)
            {
                id = db_.insert(out_buf.toByteArray());
            }
            else
            {
                db_.update(id, out_buf.toByteArray());
            }
        }
        catch(IOException exception)
        {
            throw new DatabaseIOException("writeByte failed " + 
                exception.toString());                
        }

        return id;
    }

    /**
     * Write the short to the database and return the id of the written 
     * entity. If the value if id is UNASSIGNED then entity does not currently
     * exist in the database, otherwise the entity exists so we need to update
     * it. If it fails to write then throw a DatabaseIOException 
     *
     * @param        id              record id if avaialable
     * @param        value           short value to store
     * @return       long            id of entity
     * @exception    DatabaseIOException
     */
    protected long writeShort(long id, short value)
        throws DatabaseIOException
    {
        try
        {
            ByteArrayOutputStream out_buf = new ByteArrayOutputStream();
            DataOutputStream out_stream = new DataOutputStream(out_buf);
            out_stream.writeShort(value);

            if (id == UNASSIGNED)
            {
                id = db_.insert(out_buf.toByteArray());
            }
            else
            {
                db_.update(id, out_buf.toByteArray());
            }
        }
        catch(IOException exception)
        {
            throw new DatabaseIOException("writeShort failed " + 
                exception.toString());                
        }
        
        return id;
    }

    /**
     * Write the char to the database and return the id of the written 
     * entity. If the value if id is UNASSIGNED then entity does not currently
     * exist in the database, otherwise the entity exists so we need to update
     * it. If it fails to write then throw a DatabaseIOException 
     *
     * @param        id              record id if avaialable
     * @param        value           char value to store
     * @return       long            id of entity
     * @exception    DatabaseIOException
     */
    protected long writeChar(long id, char value)
        throws DatabaseIOException
    {
        try
        {
            ByteArrayOutputStream out_buf = new ByteArrayOutputStream();
            DataOutputStream out_stream = new DataOutputStream(out_buf);
            out_stream.writeChar(value);

            if (id == UNASSIGNED)
            {
                id = db_.insert(out_buf.toByteArray());
            }
            else
            {
                db_.update(id, out_buf.toByteArray());
            }
        }
        catch(IOException exception)
        {
            throw new DatabaseIOException("writeChar failed " + 
                exception.toString());                
        }
    
        return id;
    }

    /**
     * Write the int to the database and return the id of the written 
     * entity. If the value if id is UNASSIGNED then entity does not currently
     * exist in the database, otherwise the entity exists so we need to update
     * it. If it fails to write then throw a DatabaseIOException 
     *
     * @param        id              long id if avaialable
     * @param        value           int value to store
     * @return       long            id of entity
     * @exception    DatabaseIOException
     */
    protected long writeInt(long id, int value)
        throws DatabaseIOException
    {
        try
        {
            ByteArrayOutputStream out_buf = new ByteArrayOutputStream();
            DataOutputStream out_stream = new DataOutputStream(out_buf);
            out_stream.writeInt(value);

            if (id == UNASSIGNED)
            {
                id = db_.insert(out_buf.toByteArray());
            }
            else
            {
                db_.update(id, out_buf.toByteArray());
            }
        }
        catch(IOException exception)
        {
            throw new DatabaseIOException("writeInt failed " + 
                exception.toString());                
        }

        return id;
    }

    /**
     * Write the long to the database and return the id of the written 
     * entity. If the value if id is UNASSIGNED then entity does not currently
     * exist in the database, otherwise the entity exists so we need to update
     * it. If it fails to write then throw a DatabaseIOException 
     *
     * @param        id              record id if avaialable
     * @param        value           long value to store
     * @return       long            id of entity
     * @exception    DatabaseIOException
     */
    protected long writeLong(long id, long value)
        throws DatabaseIOException
    {
        try
        {
            ByteArrayOutputStream out_buf = new ByteArrayOutputStream();
            DataOutputStream out_stream = new DataOutputStream(out_buf);
            out_stream.writeLong(value);

            if (id == UNASSIGNED)
            {
                id = db_.insert(out_buf.toByteArray());
            }
            else
            {
                db_.update(id, out_buf.toByteArray());
            }
        }
        catch(IOException exception)
        {
            throw new DatabaseIOException("writeLong failed " + 
                exception.toString());                
        }
    
        return id;
    }

    /**
     * Write the float to the database and return the id of the written 
     * entity. If the value if id is UNASSIGNED then entity does not currently
     * exist in the database, otherwise the entity exists so we need to update
     * it. If it fails to write then throw a DatabaseIOException 
     *
     * @param        id              record id if avaialable
     * @param        value           float value to store
     * @return       long            id of entity
     * @exception    DatabaseIOException
     */
    protected long writeFloat(long id, float value)
        throws DatabaseIOException
    {
        try
        {
            ByteArrayOutputStream out_buf = new ByteArrayOutputStream();
            DataOutputStream out_stream = new DataOutputStream(out_buf);
            out_stream.writeFloat(value);

            if (id == UNASSIGNED)
            {
                id = db_.insert(out_buf.toByteArray());
            }
            else
            {
                db_.update(id, out_buf.toByteArray());
            }
        }
        catch(IOException exception)
        {
            throw new DatabaseIOException("writeFloat failed " + 
                exception.toString());                
        }
    
        return id;
    }

    /**
     * Write the double to the database and return the id of the written 
     * entity. If the value if id is UNASSIGNED then entity does not currently
     * exist in the database, otherwise the entity exists so we need to update
     * it. If it fails to write then throw a DatabaseIOException 
     *
     * @param        id              record id if avaialable
     * @param        value           double value to store
     * @return       long            id of entity
     * @exception    DatabaseIOException
     */
    protected long writeDouble(long id, double value)
        throws DatabaseIOException
    {
        try
        {
            ByteArrayOutputStream out_buf = new ByteArrayOutputStream();
            DataOutputStream out_stream = new DataOutputStream(out_buf);
            out_stream.writeDouble(value);

            if (id == UNASSIGNED)
            {
                id = db_.insert(out_buf.toByteArray());
            }
            else
            {
                db_.update(id, out_buf.toByteArray());
            }
        }
        catch(Exception exception)
        {
            throw new DatabaseIOException("writeDouble failed " + 
                exception.toString());                
        }
    
        return id;
    }

    /**
     * Write the bytes[] to the database and return the id of the written 
     * entity. If the value if id is UNASSIGNED then entity does not currently
     * exist in the database, otherwise the entity exists so we need to update
     * it. If it fails to write then throw a DatabaseIOException 
     *
     * @param        id              record id if avaialable
     * @param        value           bytes[] value to store
     * @return       long            id of entity
     * @exception    DatabaseIOException
     */
    protected long writeBytes(long id, byte[] value)
        throws DatabaseIOException
    {
        try
        {
            if (id == UNASSIGNED)
            {
                id = db_.insert(value);
            }
            else
            {
                db_.update(id, value);
            }
        }
        catch(Exception exception)
        {
            throw new DatabaseIOException("writeBytes failed " + 
                exception.toString());                
        }

        return id;
    }

    /**
     * Write the string to the database and return the id of the written 
     * entity. If the value if id is UNASSIGNED then entity does not currently
     * exist in the database, otherwise the entity exists so we need to update
     * it. If it fails to write then throw a DatabaseIOException 
     *
     * @param        id              record id if avaialable
     * @param        value           string value to store
     * @return       long            id of entity
     * @exception    DatabaseIOException
     */
    protected long writeString(long id, String value)
        throws DatabaseIOException
    {
        return this.writeObject(id, value);
    }

    /**
    * Write the object to the database and return the id of the written 
    * entity. If the value if id is UNASSIGNED then entity does not currently
    * exist in the database, otherwise the entity exists so we need to update
    * it. If it fails to write then throw a DatabaseIOException 
    * <p>
    * This also does some additional work if the object is of type Persistent-
    * ObjectIfc. It checks the minimumObjectSize to see whether it should 
    * create an object which is greater than the current streamed-object size.
    * This is only valid in the 'insert' mode and can drastically effect the 
    * size of the database.
    *
    * @param        id              record id if avaialable
    * @param        value           object value to store
    * @return       long            id of entity
    * @exception    DatabaseIOException
    */
    protected long writeObject(long id, Object value)
        throws DatabaseIOException
    {
        try
        {
            ByteArrayOutputStream out_buf = new ByteArrayOutputStream(1024);
            ObjectOutputStream out_stream = new ObjectOutputStream(out_buf);
            out_stream.writeObject(value);

            if (id == UNASSIGNED)
            {
                if (value instanceof PersistentCapableIfc)
                {
                    int min_size = 
                        ((PersistentCapableIfc)value).getMinimumObjectSize();
                    if (min_size > out_buf.size())
                    {
                        id = db_.insert(new byte[min_size]);
                        db_.update(id, out_buf.toByteArray());
                    }
                    else
                    {
                        id = db_.insert(out_buf.toByteArray());
                    }
                }
                else
                {
                    id = db_.insert(out_buf.toByteArray());
                }
            }
            else
            {
                db_.update(id, out_buf.toByteArray());
            }

            // if the cache is enabled then stick the object in the 
            // cache for future reference
            if (objectCache_ != null)
            {
                objectCache_.insert(new Long(id), value);
            }
        }
        catch(IOException exception)
        {
            throw new DatabaseIOException("writeObject failed " + 
                exception.toString());                
        }
    
        return id;
    }

    /**
     * Return the next available root slot. If there are no free root nodes 
     * then throw NoAvailableRootException. If there is a generic error then
     * throw the DatabaseIOException exception
     *
     * @exception   NoAvailableRootException
     * @exception   OException   
     */
    protected int getFirstFreeRootSlot()
        throws NoAvailableRootException, IOException
    {
        int slot;
        int start = lastAllocatedRoot_ + 1;

        // check whether we have reached the end of the list
        int root_count = db_.getRootCount();
        if (start >= root_count)
        {
            start = 0;
        }

        while (start != lastAllocatedRoot_)
        {
            if (db_.getRoot(start) == 0L)
            {
                // found it so exit.
                break;
            }
            else
            {
                if (++start >= root_count)
                {
                    start = 0;
                }
            }
        }

        if (start == lastAllocatedRoot_)
        {
            throw new NoAvailableRootException("All root slots are in use.");
        }
        else
        {
            lastAllocatedRoot_ = start;
        }

        return start;
    }

    /**
     * Set the named root table. It assumes that it already exists. Should
     * only be called to replace the complete root table. It throws either 
     * DatabaseIOException or a IOException if there are any problems updating
     * the root table
     *
     * @param       PMDRootTable            new root table
     * @exception   DatabaseIOException
     * @exception   IOException
     */
    protected void setNamedRootTable(PMDRootTable table)
        throws IOException, DatabaseIOException
    {
        namedRootTable_ = table;
        updateNamedRootTable();
    }
     
    /**
     * Return the named root table. If it doesn't exist then create it. This
     * holds all the named root nodes
     * 
     * @return      Hashtable
     */
    protected PMDRootTable getNamedRootTable()
    {
        try
        {
            long rec_id = db_.getRoot(ROOT_NAMES_INDEX);
            if (rec_id == 0L)
            {
                namedRootTable_ = new PMDRootTable();
                rec_id = this.insertObject(namedRootTable_);
                // this is currently the only interface that the page
                // manager provides to get the id.
                namedRootTable_.setId(rec_id);
                this.updateObject(rec_id, namedRootTable_);
                db_.setRoot(ROOT_NAMES_INDEX, rec_id);
            }
            else
            {
                if (namedRootTable_ == null)
                {
                    namedRootTable_ = (PMDRootTable)this.readObject(rec_id);
                }
            }
        }
        catch(Exception exception)
        {
            // log a message that it failed to create the root node.
        }
        
        return namedRootTable_;
    }

    /**
     * Update the persistent copy of the named root table. If there is a 
     * problem with the update then propagate the IOException
     *
     * @exception   IOExcpetion                     
     */
    protected void updateNamedRootTable()
        throws IOException, DatabaseIOException
    {
        long rec_id = db_.getRoot(ROOT_NAMES_INDEX);
        if (rec_id != 0L)
        {
            this.updateObject(rec_id, namedRootTable_);
        }
    }

    /**
     * This will print the names of all the root entries
     *
     * @return      String
     */
    public String toString()
    {
        StringBuffer buf = new StringBuffer("PageManagedDatabase ");

        buf.append("db _ = ");
        buf.append(db_.toString());
        buf.append(" opened_ = ");
        buf.append(opened_);
        buf.append(" lastAllocatedRoot_ ");
        buf.append(lastAllocatedRoot_);
        buf.append(" namedRootTable_ ");
        buf.append(namedRootTable_.toString());
        buf.append(" name_ ");
        buf.append(name_);

        return buf.toString();
    }

    /**
     * Commit the work in the record manager. This will delegate the call down
     * to the recprd manager itself. It only has page-level scope. 
     * <p>
     * It will throw an IOException if there is a problem with the commit.
     *
     * @exception   IOException
     */
    void commit() 
        throws IOException
    {
    //    db_.commit();
    }

    /**
     * Rollback the work in the record manager. This will delegate the call down
     * to the record manager itself. It only has page-level scope. 
     * <p>
     * It will throw an IOException if there is a problem with the rollback.
     *
     * @exception   IOException
     */
    void rollback() 
        throws IOException
    {
    //    db_.rollback();
    }

    /**
     * Transient reference to the record manager.
     */
    private RecordManager db_ = null;

    /**
     * Holds the state of the database. True if it is opened and false 
     * otherwise
     */
    private boolean opened_ = false;

    /**
     * This is the index of the last allocated root node. Everytime a root
     * node is allocated it is incremented
     */
    private int lastAllocatedRoot_ = 0;

    /**
     * Maintains all the root nodes for the database.
     */
    private PMDRootTable namedRootTable_ = null;
    
    /**
     * The relative or absolute path of the database
     */
    private String name_ = null;

    /**
     * LRUCache which is used to cache database objects and save on 
     * some streaming. All writes happen through the cache
     */
    private LruObjectCache objectCache_ = null;

    /**
     * Constant representing the index of the root name tab
     */
    final static private int ROOT_NAMES_INDEX = 0;
    
    /**
     * Used internally to represent a new entry in the database. If we 
     * are updating objects in the database then we use the assigned identity.
     */
    final static private int UNASSIGNED = -1;
}
