/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 1999 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: DatabaseIfc.java,v 1.3 2000/03/09 23:52:43 jima Exp $
 *
 * Date         Author  Changes
 * 1/13/2000    jima    Created
 */
package org.exolab.core.foundation;

import java.util.Enumeration;

/**
 * This interface attempts to abstract the database layer so that is can 
 * support several types of persistent storss. including file, OODBMS and RDBMS.
 * It concept of root objects is borrowed from the OODBMS  to represent named 
 * objects, 
 * <p>
 * A database works on PersistentObjects, which have unique object identifier.
 * and can be inserted, updated or removed from the database. They can also
 * be fetched if the 'id' of the particular object is known. In some instances
 * an object can be bound to a name.Similarly, the same object can be unbound
 * or looked by name. Unbinding an object does not remove the object from the
 * database since it can still be located by object id.
 *
 * @version     $Revision: 1.3 $ $Date: 2000/03/09 23:52:43 $
 * @author      <a href="mailto:jima@exoffice.com">Jim Alateras</a>
 * @see         DatabaseManagerIfc
 * @see         PersistentCapableIfc
 **/
public interface DatabaseIfc
{
    /**
     * Opens the specified database. If it fails to open the database then it
     * throws the FailedToOpenDatabase
     *
     * @exception   FailedToOpenDatabaseException
     */
    public void open()
        throws FailedToOpenDatabaseException;

    /**
     * Close this database. If it is already closed then this becomes a no-op
     */
    public void close();

    /**
     * Test if the database is open
     *
     * @return      boolean         true if it open and false otherwise
     */
    public boolean isOpen();

    /**
     * Return an enumeration of all the database root names. A database root is 
     * one that is bound by name and represents an entry point in the database.
     *
     * @return      Enumeration     collection of root names (i.e .string)
     */
    public Enumeration getRootNames();

    /**
     * Bind the specified name to an object. All bound objects are actually
     * database roots. If an object is already bound to this name the method
     * throws the ObjectNameExistsException .
     *
     * @param       name            binding name   
     * @param       object          object to bind
     * @exception   ObjectNameExistsException
     * @exception   DatabaseIOException
     */
    public void bind(String name, PersistentCapableIfc object)
        throws ObjectNameExistsException, DatabaseIOException;

    /**
     * Unbind the object associated with the specified name. IF the name 
     * does not exist then throw the ObjectNameNotFoundException
     *
     * @param       name            binding name
     * @exception   ObjectNameNotFoundException
     * @exception   DatabaseIOException
     */
    public void unbind(String name)
        throws ObjectNameNotFoundException, DatabaseIOException;

    /**
     * Lookup the object specified by name. If the object does not exists
     * then return null.
     *
     * @param       name            lookup name
     * @return      PersistentCapableIfc          
     */
    public PersistentCapableIfc lookup(String name);

    /**
     * Check if an object is already bound to the specified name
     *
     * @param       name            lookup name
     * @return      boolean         true if name is bound to object
     */
    public boolean isBound(String name);

    /**
     * Insert the specified object into the database. Throw DatabaseIOException
     * if there is any problem
     * 
     * @param       object          object to store
     * @exception   DatabaseIOException
     */
    public void insert(PersistentCapableIfc object)
        throws DatabaseIOException;

    /**
     * Delete the specified object form the database. Throw DatabaseIOException
     * if there are any errors
     *
     * @param       object          object to delete
     * @exception   DatabaseIOException
     */
    public void delete(PersistentCapableIfc object)
        throws DatabaseIOException;

    /**
     * Update the specified object form the database. Throw DatabaseIOException
     * if there are any errors
     *
     * @param       object          object to update
     * @exception   DatabaseIOException
     */
    public void update(PersistentCapableIfc object)
        throws DatabaseIOException;
}
