/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 1999 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: Lock.java,v 1.6 2003/06/09 06:28:39 tanderson Exp $
 *
 * Date         Author  Changes
 * 3/07/2000    jima    added a new method anyOtherTxWaiting to test if there
 *                      are others waiting on this lock outside the specified
 *                      transaction context.
 */


package org.exolab.core.foundation;


import org.exolab.core.util.Messages;

/**
 * Exclusive lock and lock synchronization on an object. Each object
 * (session, entity bean, etc) is required to have one <tt>Lock</tt>
 * which is served as an exclusive lock. Only one transaction can
 * acquire the lock at any given time, but multiple transactions may
 * be waiting to acquire a lock.
 * <p>
 * In order to obtain a lock, the transaction must call {@link
 * #acquire} passing itself and the lock timeout. The transaction must
 * attempt to obtain only one lock at any given time by synchronizing
 * all calls to {@link #acquire}.
 * <p>
 * Locks implement reference counting. The transaction may acquire the
 * same lock any number of times, but must also release it the same
 * number of times (in no particular order).
 * <p>
 * If a lock cannot be acquired, the transaction will hold until the
 * lock is available or timeout occurs. If timeout occured (or a dead
 * lock has been detected), {@link LockNotGrantedException} is thrown.
 * If the object has been deleted while waiting for the lock, {@link
 * ObjectDeletedException} is thrown.
 * <p>
 * When the lock is acquired, {@link #acquire} will return the locked
 * object.
 * <p>
 * The transaction must call {@link #release} when the lock is no
 * longer required, allowing other transactions to obtain a lock. The
 * transaction must release all references to the object prior to
 * calling {@link #release}.
 * <p>
 * If the object has been deleted, the transaction must call {@link
 * #delete} instead of {@link #release}.
 *
 *
 * @author <a href="arkin@exoffice.com">Assaf Arkin</a>
 * @version $Revision: 1.6 $ $Date: 2003/06/09 06:28:39 $
 * @see TransactionContext
 */
final public class Lock
{
    /**
     * The object being locked, null if the object has been deleted.
     */
    private PersistentCapableIfc _object;


    /**
     * The lock holder on this object. Refers to the transaction that has
     * acquired the lock.
     */
    private TransactionContext _lock;


    /**
     * The number of times this lock has been acquired. Incremented by {@link
     * #acquire} and decremented by {@link #release}.
     */
    private int                _lockCount;


    /**
     * List of all transactions waiting for a lock. Attempts to acquire a
     * lock while object is locked will be recorded here. When the lock is
     * released, the first waiting transaction will be able to acquire it.
     */
    private LinkedTx           _waiting;


    /**
     * Create a new lock for the specified object. Must not create two
     * locks for the same object. This will be the object returned from
     * a successful call to {@link #acquire}.
     *
     * @param obj The object to create a lock for
     */
    public Lock( PersistentCapableIfc obj )
    {
	_object = obj;
    }


    /**
     * Returns true if the transaction holds a lock on the object. This
     * method is an efficient mean to determine whether a lock is required,
     * or if the object is owned by the transaction.
     *
     * @param tx The transaction
     * @return True if the transaction has a lock on this object
     */
    public boolean hasLock( TransactionContext tx )
    {
	return ( _lock == tx );
    }

    /**
     * Return a reference to the object that this lock guards
     *
     * @return      PersistentCapableIfc
     */
    public PersistentCapableIfc getObject()
    {
        return _object;
    }

    /**
     * Acquires a lock on the object on behalf of the specified
     * transaction. If the lock cannot be acquired immediately, the
     * thread will block until the lock is made available or the
     * timeout has elapsed. If the timeout has elapsed or a dead lock
     * has been detected, a {@link LockNotGrantedException} is thrown.
     * If the object has been deleted while waiting for a lock, a {@link
     * ObjectDeletedException} is thrown. To prevent dead locks, a
     * transaction must only call this method for any given object
     * from a single thread and must mark the lock it is trying to
     * acquire and return it from a call to {@link
     * TransactionContext#getWaitingOnLock} if the call to this method
     * has not returned yet. If this method has been called multiple times
     * by the same transaction, then {@link #release} must be called the
     * same number of times to release the lock.
     * 
     * @param tx The transaction requesting the lock
     * @param timeout Timeout waiting to acquire lock (in milliseconds),
     *  zero for no waiting
     * @return The locked object
     * @throws LockNotGrantedException Lock could not be granted in
     *  the specified timeout or a dead lock has been detected
     * @throws ObjectDeletedWaitingForLockException The object has
     *  been deleted while waiting for the lock
     */
    public synchronized PersistentCapableIfc acquire( 
        TransactionContext tx, int timeout )
	    throws LockNotGrantedException, ObjectDeletedWaitingForLockException
    {
	// Note: This method must succeed even if an exception is thrown
	// in the middle. An exception may be thrown by a Thread.stop().
	// Must make sure not to lose consistency.

	if ( _lock == tx ) {
	    // Already have write lock, can continue
	    ++_lockCount;
	} else if ( _lock == null ) {
	    // No locks, can acquire immediately
	    _lock = tx;
	    ++_lockCount;
	} else {
	    // Don't wait if timeout is zero
	    if ( timeout == 0 ) {
    		throw new LockNotGrantedException( "container.lockTimeout" );
        }

	    try {
		// Detect possibility of dead-lock. Must remain in wait-on-lock
		// position until lock is granted or exception thrown.
		tx.setWaitOnLock( this );
		detectDeadlock( tx );

		// Must wait for lock and then attempt to reacquire
		_waiting = new LinkedTx( tx, _waiting );

                // Wait until notified or timeout elapses. Must detect
                // when notified but object deleted (i.e. locks released)
		// All waiting transactions are notified at once, but once
		// notified a race condition starts to acquire new lock
		long clock = System.currentTimeMillis();
		try {
		    wait( timeout * 1000 );
		} catch ( InterruptedException except ) {
		    // If the thread is interrupted, come out with the proper message
            System.err.println("SLEEP HAS BEEN INTERRUPTED");
		    throw new LockNotGrantedException( "container.lockTimeout" );
		}
		if ( _object == null ) {
		    // If object has been deleted while waiting for lock, report deletion.
		    throw new ObjectDeletedWaitingForLockException();
		}
		// Try to re-acquire lock, this time less timeout,
		// eventually timeout of zero will either succeed or fail
		// without blocking.
		timeout -= ( System.currentTimeMillis() - clock );
		if ( timeout < 0 )
		    timeout = 0;
		removeWaiting( tx );
		return acquire( tx, timeout );
	    } finally {
		// Must always remove waiting transaction.
		removeWaiting( tx );
	    }
	}
	return _object;
    }


    /**
     * Releases a lock on the object previously acquired with {@link
     * #acquire}. This method must be called one time for each time
     * {@link #acquire} was called. After the lock is fully released,
     * other transactions might acquire it.
     *
     * @param tx The transaction that holds the lock
     */
    public synchronized void release( TransactionContext tx )
    {
	try {
	    tx.setWaitOnLock( null );
	    if ( _lock == tx ) {
		--_lockCount;
		if ( _lockCount == 0 ) {
		    _lock = null;
		    if ( _waiting != null )
			notify();
		}
	    } else
		throw new IllegalStateException( Messages.message( "container.notOwnerLock" ) );
	} catch ( Throwable death ) {
	    // This operation must never fail, not even in the
	    // event of a thread death
	    release( tx );
	}
    }


    /**
     * Check to see if the current transaction context owns the lock and that
     * no other transaction is queued waiting for it.
     *
     * @return      boolean     true if no other waiters and false otherwise
     */
    public synchronized boolean anyTxWaiting()
    {
        boolean result = true;

        if (_lockCount == 0)
        {
            // no one else is waiting for this lock
            result = false;
        }

        return result;
    }
     
    /**
     * Informs the lock that the object has been deleted by the
     * transaction holding the lock. The lock on the object is
     * released and all transactions waiting for a lock will
     * terminate with an {@link ObjectDeletedException}.
     *
     * @param tx The transaction that holds the lock
     * @throws RuntimeException Attempt to delete object without
     *   acquiring a write lock
     */
    public synchronized void delete( TransactionContext tx )
    {
	if ( tx != _lock )
	    throw new IllegalStateException( Messages.message( "container.notOwnerLock" ) );
	try {
	    // Mark lock as unlocked and deleted, notify all waiting transactions
	    _object = null;
	    _lock = null;
	    _lockCount = 0;
	    notifyAll();
	} catch ( Throwable death ) {
	    // Delete operation must never fail, not even in the
	    // event of a thread death
	    release( tx );
	}
    }


    /**
     * Detects a possible dead lock involving the transaction waiting
     * to acquire this lock. If the lock is locked by any transaction
     * waiting for a lock on <tt>waitingTx</tt>, a dead lock is
     * detected and {@link LockNotGrantedException} thrown.
     *
     * @param waitingTx The transaction waiting to acquire this lock
     */
    private void detectDeadlock( TransactionContext waitingTx )
	throws LockNotGrantedException
    {
	Lock waitOn;
    
	// For each lock look at all the waiting transactions( waitOn) and
	// determine whether they are currently waiting for a lock. A transaction
	// is waiting for a lock if it has called acquire() and has not
	// returned from the call.

	// If one of these locks is locked by this transaction, a dead lock has
	// been detected. Recursion is necessary to prevent indirect dead locks
	// (A locked by B, B locked by C, C acquires lock on A)

	// Only the last lock attempt in a dead-lock situation will cancel.

	waitOn = _lock.getWaitOnLock();
	if ( waitOn != null ) {
	    LinkedTx read;
	    
	    // Is the blocked transaction blocked by the transaction locking
	    // this object? This is a deadlock.
	    if ( waitOn._lock == waitingTx ) {
		throw new LockNotGrantedException( "container.deadlock" );
	    }
	    waitOn.detectDeadlock( waitingTx );
	}
    }


    /**
     * Remove the transaction from the waiting list.
     */
    private void removeWaiting( TransactionContext tx )
    {
	try {
		tx.setWaitOnLock( null );
	    if ( _waiting != null ) {
		if ( _waiting.tx == tx ) {
		    _waiting = _waiting.next;
		} else {
		    LinkedTx wait;
		    
		    wait = _waiting;
		    while ( wait.next != null ) {
			if ( wait.next.tx == tx ) {
			    wait.next = wait.next.next;
			    break;
			}
			wait = wait.next;
		    }
		}
	    }
	} catch ( Throwable death ) {
	    // This operation must never fail, not even in the
	    // event of a thread death
	    removeWaiting( tx );
	}
    }

    /**
     * Object uses to hold a linked list of transactions holding
     * write locks or waiting for a lock.
     */
    static class LinkedTx
    {

	TransactionContext tx;

	LinkedTx           next;

	LinkedTx( TransactionContext tx, LinkedTx next )
	{
	    this.tx = tx;
	    this.next = next;
	}

    }
}


