/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 1999 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: LruObjectCache.java,v 1.1 2000/06/07 09:17:28 jima Exp $
 *
 * Date         Author  Changes
 * 06/07/2000   jima    Created
 */
package org.exolab.core.foundation;

import java.util.Hashtable;
import java.util.Enumeration;

/**
 * This class is a global object cache for database objects. It is sycn'
 * ed with the data store and is used to improve the performance of the 
 * overall system by minimizing the number of times an object is stremed in 
 * and out of the data store.
 * <p>
 * The algorithm is quite simplistic implementation of an LRU cache whose 
 * capacity is specified during object construction time.The cache has basically
 * two operations 'insert' and 'remove'. When inserting objects into the cache,
 * and the cache is full, then the 'least used object' will be evicted.
 * <p>
 * The LRU algorithm is implemented as a doubly linked list with references to
 * the first and last elements in the list. The element at the end of the list
 * is the least used element and the element at the top of a list is the most
 * used element.
 *
 * @version     $Revision: 1.1 $ $Date: 2000/06/07 09:17:28 $
 * @author      <a href="mailto:jima@exoffice.com">Jim Alateras</a>
 **/
public class LruObjectCache
{
    /**
     * Instantiate an instance of this cache with the specified number 
     * of entries. When the number of entriesis reached the eviction
     * algorithm (in this case LRU) determines which element to evict.
     *
     * @param       size            size of the cache
     */
    public LruObjectCache(int size)
    {
        maxSize_ = size;
        hits_ = 0;
        getRequests_ = 0;
        keyMap_ = new Hashtable(size);
        head_ = null;
        tail_ = null;
    }

    /**
     * Insert the object with the following key into the cache. If the 
     * key already exists then update, the associated object
     *
     * @param       key             object key
     * @param       object          object to cache
     */
    public synchronized void insert(Object key, Object object)
    {
        touchCacheEntry(key, object);
    }

    /**
     * Remove the object with the specified key from the cache. If the 
     * key does not exist then fail silently.
     *
     * @param       key             object key
     */
    public synchronized void remove(Object key)
    {
        removeFromCache(key);
    }

    /**
     * Return the object with the specified key from the cache. If the 
     * object does not exist then return null. If the object exists then
     * increment the hit count
     *
     * @param       key             object key
     * @return      Object          object value
     */
    public synchronized Object get(Object key)
    {
        CacheEntry value = (CacheEntry)keyMap_.get(key);

        if (value != null)
        {
            // reinsert into the cache
            touchCacheEntry(key, value.value_);
        }

        return (value == null) ? null : value.value_;
    }

    /**
     * Clear all the entries in the cache
     */
    public synchronized void clear()
    {
        // walk the list and remove each element from LRU 
        // list
        CacheEntry entry = head_;
        while (entry != null)
        {
            CacheEntry temp = entry.next_;

            entry.prev_ = null;
            entry.next_ = null;
            entry.key_ = null;
            entry.value_ = null;

            entry = temp;
        }

        // clear the keyMap
        keyMap_.clear();

        // reset the fields 
        head_ = null;
        tail_ = null;
        getRequests_ = 0;
        hits_ = 0;
    }

    /**
     * Return the hit:miss ratio. This is calculate by dividing the number
     * successful retrievals from the cache by the number of requested 
     * retrievals. 
     * <p>
     * Can be used to monitor the effectiveness of the cache
     *
     * @return      double
     */
    public double hitMissRatio()
    {
        return hits_ / getRequests_;
    }

    /**
     * Return the number of entries in the cache
     *
     * @return      int         number of entries in cache
     */
    public int getUsedSlots()
    {
        return keyMap_.size();
    }

    /**
     * Return the number of free slots inthe cache
     *
     * @return      int         number of free entries in cache
     */
    public int getFreeSlots()
    {
        return maxSize_ - keyMap_.size();
    }
     
    /**
     * Return an enumeration of all the elements in the cache, Most
     * recently used ones first.
     *
     * @return      Enumeration
     */
    public Enumeration elements()
    {
        return new LruCacheEnumeration(tail_);
    }

    /**
     * Return an enumeration of entries in the cache. Iterating
     * this enumeration is expensive
     
    /**
     * Add the specified key and object to the cache to the end 
     * of the LRU queue
     *
     * @param       key         object key
     * @param       object      object value
     */
    protected void touchCacheEntry(Object key, Object object)
    {
        if ((key != null) &&
            (object != null))
        {
            // adjsut the number of requests
            getRequests_++;
            CacheEntry entry = (CacheEntry)keyMap_.get(key);
            if (entry == null)
            {
                // create a new entry. check that we do not 
                // exceed the maxSize_ then remove the least used
                // entry
                if (keyMap_.size() < maxSize_)
                {
                    entry = new CacheEntry(key, object);
                }
                else
                {
                    entry = head_;
                    head_ = entry.prev_;
                    head_.next_ = null;

                    // remove it from the keyMap and reset the values
                    keyMap_.remove(entry.key_);
                    entry.value_ = object;
                    entry.key_ = key;
                    entry.prev_ = null;
                    entry.next_ = null;
                }
            }
            else
            {
                // remove the entry from the cache if it currently
                // exists. Also increments the hits_ count
                removeFromCache(entry);
                entry.value_ = object;
                hits_++;
            }

            // place the entry at the end of the list
            if (tail_ != null)
            {
                tail_.prev_ = entry;
            }

            entry.next_ = tail_;
            entry.prev_ = null;
            tail_ = entry;

            // if the head is also null then populate it with
            // the same value as the tail.
            if (head_ == null)
            {
                head_ = tail_;
            }

            // add it to the keyMap as well
            keyMap_.put(key, entry);
        }
    }
     
    /**
     * Remove the specified entry from the cache and return it
     * to the client. Adjust all the pointers and remove the entry
     * from the keyMap
     *
     * @param       entry       entry to remove
     * @return      CacheEntry  removed cache entry
     */
    protected CacheEntry removeFromCache(CacheEntry entry)
    {
        // adjust all the links in the list around the 
        // entry being removed
        if (entry.prev_ != null)
        {
            entry.prev_.next_ = entry.next_;
        }
        else
        {
            tail_ = entry.next_;
        }

        if (entry.next_ != null)
        {
            entry.next_.prev_ = entry.prev_;
        }
        else
        {
            head_ = entry.prev_;
        }

        // remove it from the keyMap as well
        keyMap_.remove(entry.key_);

        return entry;
    }

    /** 
     * Remove the element with the specified key from the cache
     * and return it to the client.If the element does not exist
     * then return null
     *
     * @param       key         key of object to remove
     * @return      CacheEntry  removed cache entry
     */
    protected CacheEntry removeFromCache(Object key)
    {
        CacheEntry entry = null;

        if (key != null)
        {
            entry = (CacheEntry)keyMap_.get(key);
            if (entry != null)
            {
                removeFromCache(entry);
                entry.key_ = null;
                entry.value_ = null;
                entry.prev_ = null;
                entry.next_ = null;
            }
        }

        return entry;
    }



    /**
     * Holds the maximum size of the cache
     */
    private int maxSize_ = 0;

    /**
     * The map holds a reference to all the keys anfd their associated 
     * cache entry.
     */
    private Hashtable keyMap_ = null;

    /**
     * Holds the number of successful hits made on the cache
     */
    private long hits_ = 0;

    /**
     * Holds the total number of 'get' requests made on the cache
     */
    private long getRequests_ = 0;

    /**
     * Points to the head of the LRU cache. The head is the least-
     * recently used entry and is the first to get evicted when the cache is full
     */
    private CacheEntry head_ = null;

    /** 
     * Points to the tail fo the LRU cache. The tail is the most recently 
     * used cache entry.
     */
    private CacheEntry tail_ = null;
}



/**
 * This inner class wraps up an object into a cache entry. It 
 * takes the key and the object value and then constructs an entry
 * which has pointers to the previous and next entries in the 
 * list
 */
class CacheEntry
{
    /**
     * Construct an instance of the cache entry give the specified
     * key and object. Reset all other fields
     *
     * @param       key     object key
     * @param       value   object vlaue
     */
    CacheEntry(Object key, Object value)
    {
        key_= key;
        value_ = value;
    }


    /**
     * Object key,
     */
    Object key_ = null;

    /** 
     * Object value
     */
    Object value_ = null;
    
    /**
     * Points to the previous entry in the cache or null if it 
     * is the head of the cache
     */
    CacheEntry prev_ = null;

    /** 
     * Points to the next etnry in the cache or nul;l if it is
     * the tail of the cache
     */
    CacheEntry next_ = null;
}



/**
 * This class implements the Enumeration interface which enables the 
 * client to enumerate through all the entries in the cache. It enumerates
 * from the most rcently used to the least recently used entry.
 * <p>
 * The Enumeration is not synchronized and not thread-safe.
 */
class LruCacheEnumeration
    implements Enumeration   
{  
    /**
     * Construct an enumeration given the following CacheEntry. It is
     * assumed that this is the last entry in the list and continously calls 
     * previous to iterate
     *
     * @param       entry           start enumerating from this point
     */
    public LruCacheEnumeration(CacheEntry entry)
    {
        current_ = entry;
    }

    // implementation of Enumeration.hasMoreElements
    public boolean hasMoreElements()
    {
        return (current_ != null);
    }

    // implementation of Enumeration.nextElement
    public Object nextElement()
    {
        CacheEntry result = current_;
        current_ = current_.next_;

        return result.value_;
    }

    /**
     * Cursor to the current element in the list
     */
    private CacheEntry current_ = null;
}
