/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 1999 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: PersistentObject.java,v 1.7 2003/06/09 06:28:39 tanderson Exp $
 *
 * Date         Author  Changes
 * 1/26/2000    jima    Created
 */
package org.exolab.core.foundation;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

/**
 * This is the base class for all persistent objects. A persistent object is
 * on that can be placed in a persistent store, is transactional, is lockable
 * etc. All persistent objects, must derive from this class; with no exception.
 * This class also has an id and version, which uniquely identifies this object 
 * within an application. 
 * <p>
 * This object is also Serializable
 *
 * @version     $Revision: 1.7 $ $Date: 2003/06/09 06:28:39 $
 * @author      <a href="mailto:jima@exoffice.com">Jim Alateras</a>
 * @see         ObjectId
 * @see         VersionId
 * @see         PersistentCapableIfc
 * @see         java.io.Serializable
  **/
abstract public class PersistentObject
    implements PersistentCapableIfc, Cloneable, Externalizable
{
    /**
     * The default constructor does nothing
     */
    public PersistentObject()
    {
        id_ = new ObjectId(0);
        version_ = new ObjectVersion(1);
    }

    /**
     * Instantiate an instance of this class with the specified object id
     * and version number.
     *
     * @param       id              unique object identifier
     * @param       version         version number
     */
    public PersistentObject(ObjectId id, ObjectVersion version)
    {
        id_ = id;
        version_ = version;
    }


    /**
     * Set the object id using the passed in identifier. 
     *
     * @param       id              unique object id
     */
    public void setId(long id)
    {
        id_.setId(id);
    }

    /**
     * Return an instance of the object identity
     *
     * @return      long
     */
    public long getId()
    {
        return id_.getId();
    }

    /**
     * Set the object id for this persistent object using the specified object.
     *
     * @param       id              object id
     */
    public void setObjectId(ObjectId id)
    {
        id_ = id;
    }

    /**
     * Return the object identifier associated with this object
     *
     * @return      ObjectId
     */
    public ObjectId getObjectId()
    {
        return id_;
    }

    /**
     * Set the object version using the passed in identifier. 
     *
     * @param       version         version
     */
    public void setVersion(long version)
    {
        version_.setVersion(version);
    }

    /**
     * Return the version of the object
     *
     * @return      long
     */
    public long getVersion()
    {
        return version_.getVersion();
    }

    /**
     * Set the object version for this persistent object using the specified
     * object.
     *
     * @param       version         object version
     */
    public void setObjectVersion(ObjectVersion version)
    {
        version_ = version;
    }

    /**
     * Return the object version associated with this object
     *
     * @return      ObjectId
     */
    public ObjectVersion getObjectVersion()
    {
        return version_;
    }

    /**
     * Set the minimum size that the object should occurpy. It is only
     * a hint.
     *
     * @param       size        minimum object size
     */
    public void setMinimumObjectSize(int size)
    {
        minimumSize_ = size;
    }

    /**
     * Return the minimum size of this object. This vould be used to 
     * preallocate the object size for more efficient storage.
     *
     * @return      int
     */
    public int getMinimumObjectSize()
    {
        return minimumSize_;
    }

    /**
     * Check if the two objects are equal. Equality implies that both
     * objects are the same type and the object id's of the objects are
     * equal
     *
     * @param       object          object to compare against
     * @return      boolean         true if identical
     */
    public boolean equals(Object object)
    {
        boolean result = false;

        if ((object instanceof PersistentObject) &&
            (this.getObjectId().equals(((PersistentObject)object).getObjectId())))
        {
            result = true;
        }

        return result;
    }

    /**
     * Clone this object and its base objects. 
     *
     * @return Obect - the cloned object
     * @exception CloneNotSupportedException
     */
    public Object clone()
        throws CloneNotSupportedException
    {
        PersistentObject clone = null;
            
        clone = (PersistentObject)super.clone();
        clone.id_ = new ObjectId(clone.id_.getId());
        clone.version_ = new ObjectVersion(clone.version_.getVersion());

        return clone;
    }

    // implementation of Externalizable.writeExternal
    public void writeExternal(ObjectOutput stream)
    throws IOException
    {
		stream.writeLong(serialVerionUID);
        stream.writeLong(id_.getId());
        stream.writeLong(version_.getVersion());
    }

    // implementation of Externalizable.writeExternal
    public void readExternal(ObjectInput stream)
    throws IOException, ClassNotFoundException
    {
		long version = stream.readLong();
		if (version == serialVerionUID) {	
            id_ = new ObjectId(stream.readLong());
            version_ = new ObjectVersion(stream.readLong());
        } else {
            throw new IOException("PersistentMessageHandle with version " +
                version + " is not supported.");
        }
    }


    private ObjectId id_ = null;
    private ObjectVersion version_ = null;
    transient int minimumSize_ = 0;

    /**
     * Used for serialization
     */
    static final long serialVerionUID = 1;
    
}
