/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 1999 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: SessionIfc.java,v 1.3 2000/05/19 00:39:54 jima Exp $
 *
 * Date         Author  Changes
 * 1/14/2000    jima    Created
 */

package org.exolab.core.foundation;

import java.util.Enumeration;

/**
 * A session defines a communication pipe for clients and servers to exchange
 * information. A session is responsible for authorisation and authentication
 * and for providing a mediating interface to the specific service. It is the
 * first point of contact for clients to your service. 
 * <p>
 * Because it is a mediating interface is supports functions from DatabaseIfc
 * and TransactionIfc.
 * <p>
 * Sessions are transient objects that can only be created by the 
 * SessionManager.
 *
 * @version     $Revision: 1.3 $ $Date: 2000/05/19 00:39:54 $
 * @author      <a href="mailto:jima@exoffice.com">Jim Alateras</a>
 * @see         SessionManager
 **/
public interface SessionIfc
{
    /**
     * Return an enumeration of all the database root names. A database root is 
     * one that is bound by name and represents an entry point in the database.
     *
     * @return      Enumeration     collection of roots
     */
    public Enumeration getRootNames();

    /**
     * Bind the specified name to an object. All bound objects are actually
     * database roots. If an object is already bound to this name the method
     * throws the ObjectNameExistsException. If there is a problem binding the
     * object to the specified name then the DataIOException is raised.
     * <p>
     * The TransactionNotInProgressException is thrown when bind is called 
     * outside a transaction.
     *
     * @param       name            binding name   
     * @param       object          object to bind
     * @exception   ObjectNameExistsException
     * @exception   DatabaseIOException
     * @exception   TransactionNotInProgressException
     */
    public void bind(String name, PersistentCapableIfc object)
        throws ObjectNameExistsException, DatabaseIOException,
            TransactionNotInProgressException;

    /**
     * Unbind the object associated with the specified name. IF the name 
     * does not exist then throw the ObjectNameNotFoundException. If there 
     * are database related problems to the unbind then  throw 
     * DatabaseIOException
     * <p>
     * The TransactionNotInProgressException is thrown when unbind is called 
     * outside a transaction.
     *
     * @param       name            binding name
     * @exception   ObjectNameNotFoundException
     * @exception   DatabaseIOException
     * @exception   TransactionNotInProgressException
     */
    public void unbind(String name)
        throws ObjectNameNotFoundException, DatabaseIOException,
            TransactionNotInProgressException;

    /**
     * Lookup the object specified by name. If the object does not exists
     * then return null.
     *
     * @param       name            lookup name
     * @return      Object          
     */
    public PersistentCapableIfc lookup(String name);

    /**
     * Check if an object is already bound to the specified name
     *
     * @param       name            lookup name
     * @return      boolean         true if name is bound to object
     */
    public boolean isBound(String name);

    /**
     * Return the current transaction for this session. This will always be 
     * a non-null object but the transaction may not be currently opened.
     *
     * @return      TransactionIfc
     */
    public TransactionIfc getCurrentTransaction();

    /**
     * Return the database that this session is associated with. This should
     * not return a non-null object but could return a DatabaseIfc instance 
     * that is invalid or closed
     *
     * @return      DatabaseIfc
     */
    public DatabaseIfc getDatabase();
    
    /**
     * Return the session manager responsible for allocating this session
     *
     * @return      SessionManager
     */
    public SessionManagerIfc getSessionManager();

    /**
     * Set the session manager that allocated this session. If a session
     * already exists then overwrite it.
     *
     * @param       manager     session manager
     */
    public void setSessionManager(SessionManagerIfc manager);
    
    /**
     * Create the specified persistent object in the database for this 
     * session. The object must be created within a transaction and is
     * only made permanent when a commit is issued. If there is not transaction
     * in progress the TransactionNotInProgressException is raised. If  there
     * is a general database problem then the DatabaseIOException is raised.
     *
     * @param           object              object to create
     * @exception       TransactionNotInProgressException
     * @exception       DatabaseIOException
     */
    public void createObject(PersistentCapableIfc object)
        throws TransactionNotInProgressException, DatabaseIOException;
    
    /**
     * Update the specified persistent object in the database for this 
     * session. The object must be updated within a transaction and is
     * only made permanent when a commit is issued. If there is not transaction
     * in progress the TransactionNotInProgressException is raised. If  there
     * is a general database problem then the DatabaseIOException is raised.
     *
     * @param           object              object to update
     * @exception       TransactionNotInProgressException
     * @exception       DatabaseIOException
     */
    public void updateObject(PersistentCapableIfc object)
        throws TransactionNotInProgressException, DatabaseIOException;
    
    /**
     * Delete the specified persistent object in the database for this 
     * session. The object must be deleted within a transaction and is
     * only made deleted from thedatabase when the transaction commits. If 
     * there is no transaction in progress the TransactionNotInProgress-
     * Exception is raised. If  there is a general database problem then
     * the DatabaseIOException is raised.
     *
     * @param       object              object to delete
     * @exception   TransactionNotInProgressException
     * @exception   DatabaseIOException
     */
    public void deleteObject(PersistentCapableIfc object)
        throws TransactionNotInProgressException, DatabaseIOException;
    
    /**
     * Retrieve the persistent object associated with the id. If the object
     * does not exist then throw a DatabaseIOException, If there is no 
     * transaction in progress then throw a TransactionNotInProgressException.
     *
     * @param       id          identity of object to retrieve
     * @return      PersistentCapableIfc
     * @exception   TransactionNotInProgressException
     * @exception   DatabaseIOException
     */
    public PersistentCapableIfc retrieveObject(long id)
        throws TransactionNotInProgressException, DatabaseIOException;

    /**
     * Acquire an exclusive lock on the specified object. This will lock out
     * all other transactions trying to access the specified object. If the 
     * session cannot acquire a lock then FailedToAcquireLockException is 
     * raised. The TransactionNotInProgressException is thrown when an attempt
     * is made to acquire a lock outside a transaction
     * <p>
     * The timeout determines how long to wait for the lock. A value of 
     * DEFAULT_LOCK_TIMEOUT is used to wait the default time.
     * <p>
     * All locks granted to a session will be automatically released in an
     * abort, commit or rollback
     *
     * @param       object          object to lock
     * @param       int             time we should wait in seconds
     * @return      Object          reference to locked object
     * @exception   FailedToAcquireLockException
     * @exception   FailedToCreateLockException
     * @exception   TransactionNotInProgressException
     */
    public void acquireLock(PersistentCapableIfc object, int timeout)
        throws FailedToAcquireLockException, FailedToCreateLockException,
            TransactionNotInProgressException;
     
    /**
     * Create a handle for the specified PersistentCapableIfc object. If it 
     * cannot create the handle then throw the FailedToCreateHandleException
     *
     * @param       object          PersistentCapableIfc
     * @exception   FailedToCreateHandleException
     * @exception   TransactionNotInProgressException
     */
    public HandleIfc createHandle(PersistentCapableIfc object)
        throws FailedToCreateHandleException, TransactionNotInProgressException;

    /**
     * Return a reference to the collection manager, which is used to allocate
     * different collection types includinng queues and vectors.
     *
     * @return      CollectionManagerIfc
     */
    public CollectionManagerIfc getCollectionManager();

    /**
     * Return the session id associated with this object
     *
     * @return      long        session identifier
     */
    public long getSessionId();

    /**
     * Check the validity of the session
     *
     * @return      boolean     true if the session is valid
     */
    public boolean isValid();


    // specifies the acquire wishes to acquire a lock within the default time
    // period. The period is in seconds
    public static final long DEFAULT_LOCK_TIMEOUT = -1;
}
