/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 1999 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: TransactionIfc.java,v 1.1 2000/03/09 23:52:43 jima Exp $
 *
 * Date         Author  Changes
 * 3/05/2000   jima    Created
 */

package org.exolab.core.foundation;



/**
 * This is an protocol for a Transaction. A Transaction is associated with one
 * and only one session (a session serialises transactions) but may have be 
 * associated with one or more threads. A transaction can be in one three 
 * possible states.open, committed or closed. You can also begin, commit,
 * abort or rollback a transaction (in this context abort and rollback are 
 * sematically identical since their is no support for savepoints). It also has
 * a method called 'setRollbackOnly' which will force the transaction to be 
 * rolledback even if a commit is issued.
 * 
 * @version     $Revision: 1.1 $ $Date: 2000/03/09 23:52:43 $
 * @author      <a href="mailto:jima@exoffice.com">Jim Alateras</a>
 **/

public interface TransactionIfc
{
    /**
     * Put the transaction in the open state. If the transaction is already 
     * opened then throw TransactionInProgressException
     *
     * @exception       TransactionInProgressException
     */
    public void begin()
        throws TransactionInProgressException;

    /**
     * Close the current transaction. If the setRollbackOnly has been called
     * while the transaction was open then a rollback is forced and the 
     * TransactionException exception is thrown. 
     * <p>
     * If the transaction is not currently opened then TransactionNotInProg-
     * ressException is raised.
     *
     * @exception       TransactionException
     * @exception       TransactionNotInProgressException
     * @exception       DatabaseIOException
     */
    public void commit()
        throws TransactionNotInProgressException, TransactionException,
            DatabaseIOException;

    /**
     * Abort the current transaction. If there is no transaction currently in
     * progress then throw TransactionNotInProgressException
     *
     * @exception       TransactionNotInProgressException
     * @exception       DatabaseIOException
     */
    public void abort()
        throws TransactionNotInProgressException, DatabaseIOException;

    /**
     * Rollback the current transaction. If there is no transaction currently 
     * in progress then throw TransactionNotInProgressException
     *
     * @exception       TransactionNotInProgressException
     * @exception       DatabaseIOException
     */
    public void rollback()
        throws TransactionNotInProgressException, DatabaseIOException;

    /**
     * Set the transaction so that the only outcome is rollback. If there is no
     * transaction in progress raise TransactionNotInProgressException.
     *
     * @param           status          true to set rollback only; false to
     *                                  clear the rollback only flag
     * @exception       TransactionNotInProgressException
     */
    public void setRollbackOnly(boolean status)
        throws TransactionNotInProgressException;

    /**
     * Test if the transaction is opened.
     *
     * @return          boolean         true if opened; false otherwise
     */
    public boolean isOpened();

    /**
     * Return the transaction context assoicated with this transaction
     *
     * @return          TransactionContext
     */
   public TransactionContext getTransactionContext();
}

