/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * 
 * Copyright 2001,2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: HttpClient.java,v 1.5 2003/03/26 11:27:52 tanderson Exp $
 *
 * Date				Author  Changes
 * Fri 21 Sep 2001	mourikis    Created
 *
 */

package org.exolab.core.http;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.security.Security;

import java.net.URL;
import java.net.MalformedURLException;
import java.net.URLConnection;

import org.exolab.core.ipc.IpcIfc;


public class HttpClient implements IpcIfc {

    /**
     * The Unique id of this object. 
     * This is not really needed for Http Requests, but is implemented
     * just to make this connector appear like a normal ipc connector.
     */
    private String _label;

    /**
     * Any send request that expects a return value is stored here and returned
     * in the receive call.
     */
    private Object _receivedData = null;

    /**
     * The URL location of the Web server that will service the OpenJMS
     * requests.
     */
    private URL _server = null;

    
    /**
     * Creates an HttpClient Object. No connection occurs at this point,
     * just some initialisation.
     *
     * @param url The URL of the server hosting the OpenJMS servlet.
     * @param label An id for this client.
     * @throws MalformedURLException if url syntax is incorrect.
     */
    public HttpClient(String url, String label) throws MalformedURLException {

        _label = label;
        /* CONNECTING BY WAY OF PROXY  
           Client needs to set these system properties if it requires to 
           go through a proxy for HTTP.
        System.setProperty("http.proxyHost", host);
        System.setProperty("http.proxyPort", port);
        */

        /* CONNECTING FOR SSL TRANSACTIONS. 
           Client needs to set these system properties if it requires to 
           go through a proxy for HTTPS.
        System.setProperty("https.proxyHost", "host");
        System.setProperty("https.proxyPort", "port");
        System.setProperty("java.protocol.handler.pkgs",
                           "com.sun.net.ssl.internal.www.protocol");
        Security.addProvider(new com.sun.net.ssl.internal.ssl.Provider());
        */

        _server = new URL(url);
    }

    /**
     * Implementation of IpcIfc.close the channel.
     */
    public void close() throws IOException {
        // todo. for http no closure is required, but need
        // to inform server at least of closure to free up resources
        // and clean up.
        _receivedData = null;
        _label = null;
    }

    /**
     * Send the specified object. 
     *
     * @param ob The object to send. Must be serializable.
     * @throws IOException If the data fails to write to the output stream.
     */
    public void send(Serializable ob, String host, String port) 
        throws IOException {
        _receivedData = null;
        sendMessage(ob, true, true, host, port);
    }

    /**
     * Send the specified object. 
     *
     * @param ob The object to send. Must be serializable.
     * @throws IOException If the data fails to write to the output stream.
     */
    public void send(Serializable ob) throws IOException {
        _receivedData = null;
        sendMessage(ob, true, true, null, null);
    }

    /**
     * Send the specified object. No response is expected by the client.
     *
     * @param ob The object to send. Must be serializable.
     * @throws IOException If the data fails to write to the output stream.
     */
    public void sendWithoutResponse(Serializable ob) throws IOException {
        _receivedData = null;
        sendMessage(ob, true, false, null, null);
    }

    /**
     * Send the specified object. No response is expected by the client.
     *
     * @param ob The object to send. Must be serializable.
     * @throws IOException If the data fails to write to the output stream.
     */
    public void sendWithoutResponse(Serializable ob, String host, String port)
        throws IOException {
        _receivedData = null;
        sendMessage(ob, true, false, host, port);
    }

    /**
     * Return the value of any send reply.
     *
     * @return the response
     */
    public Object receive() {
        return _receivedData;
    }

    /**
     * Return the server URL
     */
    public URL getURL() {
        return _server;
    }

    /**
     * Returns a unique id for this connection.
     *
     * @return the id of this connection
     */
    public String toString() {
        return ("HttpClient - " + _label);
    }

    /**
     * Send the message to the servlet though an http request.
     *
     *<P>
     * <b>Note:</b>
     * I have added as much comment as I can in the method to help me remember
     * why I did things the way I did without constantly having to look things
     * up. Also makes it easier for others modifying this stuff to understand.
     * <b>Jim M.</b>
     *
     * @param data The data to be sent.
     * @param op True if this is a post operation
     * @param expectResponse Set an item in the header, to inform servlet
     * when it sends the message to wait for a response, and return object.
     * @param host Set the host address the servlet needs to connect to.
     * @param port Set the port address the servlet needs to connect to.
     */
    private void sendMessage(Object data, boolean op, boolean expectResponse,
                             String host, String port) 
        throws IOException {

        // send request
        ObjectInputStream response = null;
        Object result = null;
        try {
            // open the url connection, use URLConnection instead of 
            // HttpURLConnection in case this will allow us to use HTTPS 
            // later on. HttpURLConnection 
            // does not support HTTPS.  
            URLConnection con = _server.openConnection();

            
            // if(con instanceof com.sun.net.ssl.HttpsURLConnection) {
            //    System.out.println("Using HTTPS connection");
            // }

            // if the DoOutput flag is set to true inform URL connection 
            // that we are sending data out.  The default behavior is false. 
            // This will change the request from GET to a POST.  
            con.setDoOutput(op);

            // Setting the UseCaches to false allows the application 
            // "tunnel through" and ignore the caches.  
            // If the UseCaches flag on a connection is true, 
            // the connection is allowed to use whatever caches it can.  
            con.setUseCaches(false);

            // setRequestProperty is a cool feature to keep in mind.  
            // Basically, the http header that goes to and from the server 
            // is in the form of a 'name-value' pair.  setRequestProperty 
            // sets and adds name value pairs.  The servlet can look at 
            // values sent to the server, and set values in the response 
            // back to the client.  This in its self  is a backdoor approach 
            // to sending and receiving additional information separate from 
            // object serialization.
            //<p>
            // Setting the 'Content-Type' type to 'application' says that 
            // stream is "intended for the transmission of data to be 
            // interpreted by locally-installed programs". 
            // <p>
            // Refer http://www.io.com/docs/MIME_html/1521/Appendix_F.html
            // for more on content types.
            con.setRequestProperty("Content-Type","application/octet-stream");
            if (expectResponse) {
                con.setRequestProperty("jms-response", "yes");
            }
            if (host != null) {
                con.setRequestProperty("jms-host", host);
            }
            if (port != null) {
                con.setRequestProperty("jms-port", port);
            }
            

            // Get the output stream and put in a form that is easy to use and 
            // efficient.
            ObjectOutputStream request =
                new ObjectOutputStream(new BufferedOutputStream
                    (con.getOutputStream()));

            request.writeObject(data);
            request.flush();
            request.close();

            // get the result input stream 
            response = new ObjectInputStream(new BufferedInputStream
                (con.getInputStream()));

            if (expectResponse) {
                // read response back from the server
                _receivedData = response.readObject();
            }
        } catch (ClassNotFoundException err) {
            System.err.println(err);
            throw new IOException(err.getMessage());
        } finally {
            try {
                if( response!=null ) {
                    response.close();
                }
            } catch(IOException ex) {}
        }
    }
}
