/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2000 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: Client.java,v 1.2 2001/05/20 10:49:35 jima Exp $
 *
 * Date         Author  Changes
 * $Date	    jimm    Created
 */


package org.exolab.core.ipc;

import java.io.Serializable;
import java.io.IOException;
import java.lang.ClassNotFoundException;
import java.rmi.UnknownHostException;

/**
 * Client side ipc connections, must connect with a host and port option.
 * An exception is thrown if the connection fails.
 * Currently all sends have a 10 sec timeout, receives will block
 * until data is available.
 *
 * <P> For the initial release a positive acknowlegment is used. Clients that
 * initiate a send should then call receive to ensure an response is received.
 * 
 *
 * @version     $Revision: 1.2 $ $Date: 2001/05/20 10:49:35 $
 * @author      <a href="mailto:mourikis@exolab.org">Jim Mourikis</a>
 * @see         org.exolab.core.ipc.TcpService
 *
 **/


public class Client 
    implements Serializable {
    // The ipc service used for all connections.
    private transient IpcIfc          service_;

    /**
     * Connect to a server with the given port on the given host.
     *
     * @param host The host the service is running on.
     * @param port The connection port the server is listening on.
     * @exception UnknownHostException Cannot fine the given host.	
     * @exception IOException Failed to connect to the server.
     *
     */
    public Client(String host, int port) 
    throws UnknownHostException, IOException
    {
        service_ = new TcpService(host, port);
    }


    /**
     * Send the specified object.
     *
     * @param ob The object to send. Must be serializable.
     * @exception IOException If the data fails to write to the output stream.
     *
     */
    public void send(Serializable ob) throws IOException
    {
        service_.send(ob);
    }


    /**
     * Wait for an Object to be received.
     *
     * @return Object The object received.
     * @exception IOException If the data cannot be received.
     * @exception ClassNotFoundException If the object received is unknown.
     *
     */
    public Object receive() throws IOException, ClassNotFoundException
    {
        return service_.receive();
    }


    /**
     * Close the existing service. Connections can no longer be 
     * sent or received.
     *
     * @exception IOException If the service fails to close.
     *
     */
    public void close() throws IOException
    {
        service_.close();
    }

} // End Client
