/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2000 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: Notifier.java,v 1.3 2000/06/27 01:23:32 mourikis Exp $
 *
 * Date         Author  Changes
 * $Date	    jimm    Created
 */


package org.exolab.core.ipc;

import java.io.Serializable;
import java.io.IOException;
import java.lang.ClassNotFoundException;
import java.lang.Thread;
import java.lang.Runnable;
import java.lang.InterruptedException;
import java.util.HashMap;


/**
 * This class is given a new ipc connection and the clients callback handle.
 * The object creates an internal thread which is used to exclusively monitor
 * this ipc connection, and inform the client via the handle of incoming
 * requests. 
 *
 * If the client handle returns a non-null object, this is passed back down
 * the ipc connection object.
 *
 * @version     $Revision: 1.3 $ $Date: 2000/06/27 01:23:32 $
 * @author      <a href="mailto:mourikis@exolab.org">Jim Mourikis</a>
 * @see         org.exolab.core.ipc.TcpService
 * @see         org.exolab.core.ipc.TcpServerService
 * @see         org.exolab.core.ipc.NotifierIfc
 *
 **/


public class Notifier 
{
	// The receiving client reference.
	NotifierIfc		client_;

	// The ipc service connection.
	IpcIfc			ipc_;
	
	// The main loop controller for this thread.
	boolean			stopRequested_ = false;
	
	// The controlling thread.
	Thread			thread_;
	
	// A reference to the pool all notifier objects belong to.
	HashMap			pool_;

	// The id od this connection.
	String			id_;
	
	
	/**
	 * Construct the notifier with a reference to the client interested
	 * in receiving messages. Create an internal thread, and begin listening
	 * to the ipc connection.
	 *
	 * @param client The receiving client.
	 * @param ipc The Ipc connection established.
	 * @param pool_ The pool this notifier belongs to.
	 *
	 */
	public Notifier(NotifierIfc client, IpcIfc ipc, HashMap pool)
	{
		client_ = client;
		ipc_ = ipc;
		pool_ = pool;
		id_ = ipc.toString();
		Runnable r = new Runnable()
			{
				public void run()
				{
					start();
				}
			};
		thread_ = new Thread(r);
		thread_.start();
		// Remove this Object from pool.
		synchronized(pool_)
		{
			pool_.put(thread_.getName(), this);
		}
	}


	/**
	 * An internal thread is started, and the ipc connection is monitored for
	 * incoming messages. Any received message is passed back to the client.
	 * If the client returns a non-null object, it is passed back down the
	 * ipc channel.
	 *
	 */
	private void start()
	{
		Object ob;
		Serializable returnValue;
		
		while (!stopRequested_)
		{
			try
			{
				ob = ipc_.receive();
				returnValue = client_.notify(ob, id_);
				if (returnValue != null)
				{
					ipc_.send(returnValue);
				}
			}
			catch(IOException err)
			{
				stopRequested_ = true;
			}
			catch(ClassNotFoundException err)
			{
				System.err.println("Unknown Class Ignoring\n" + err);
			}
			catch(Exception err)
			{
				System.err.println("Shutdown request");
			}
		}
		client_.disconnection(id_);
		// Remove this Object from pool.
		synchronized(pool_)
		{
			if (ipc_ != null)
			{
				try
				{
					ipc_.close();
				}
				catch(IOException err) {/*May already be closed.*/}
				ipc_ = null;
			}
			pool_.remove(thread_.getName());
		}
	}




	/**
	 * Stop execution of this thread. And close all ipc connections in use.
	 *
	 */
	public void stop()
	{
		try
		{
			stopRequested_ = true;
			thread_.interrupt();
			ipc_.close();
			ipc_ = null;
		}
		catch(IOException err)
		{
			// if ipc is already dead.
		}
	}
	
} // End Notifier
