/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2000,2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: Server.java,v 1.3 2003/06/19 05:58:52 tanderson Exp $
 *
 * Date         Author  Changes
 * $Date	    jimm    Created
 */
package org.exolab.core.ipc;

import java.io.IOException;
import java.net.InetAddress;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;


/**
 * The main entry point for Ipc server connections.
 * Create and start the appropriate service for a server.
 *
 * @version     $Revision: 1.3 $ $Date: 2003/06/19 05:58:52 $
 * @author      <a href="mailto:mourikis@exolab.org">Jim Mourikis</a>
 * @see         org.exolab.core.ipc.TcpServerService
 */
public class Server implements Runnable {

    /**
     * The current active IPC service
     */
    private IpcServerIfc _service;

    /**
     *  The logger
     */
    private static final Log _log = LogFactory.getLog(Server.class);


    /**
     * Instantiate an appropriate service.
     * Pass in a notifier handle to receive notifications of messages received.
     *
     * @param notifier The client handle to call when data is received.
     */
    public Server(NotifierIfc notifier) throws IOException {
        this(0, notifier);
    }

    /**
     * Construct an appropriate service with the given port number.
     * Pass in a notifier handle to receive notifications of messages received.
     *
     * @param port The port to connect on.
     * @param notifier The client handle to call when data is received.
     */
    public Server(int port, NotifierIfc notifier) throws IOException {
        this(null, port, notifier);
    }

    /**
     * Construct an appropriate service with the given port number.
     * Pass in a notifier handle to receive notifications of messages received.
     *
     * @param address the local IP address to bind to. May be null.
     * @param port the port to connect on.
     * @param notifier the client handle to call when data is received.
     */
    public Server(InetAddress address, int port, NotifierIfc notifier) 
        throws IOException {
        _service = new TcpServerService(address, port, notifier);
    }

    /**
     * Shutdown the service.
     */
    public void stop() {
        try {
            _service.shutdownAll();
        } catch(IOException exception) {
            _log.debug(exception.getMessage(), exception);
	}
    }

    /**
     * Return the port number this server is listening on.
     * 
     * @return the port number the server is using.
     */
    public int getPort() {
        return _service.getPort();
    }

    /**
     * Return the machines address.
     * 
     * @return the machines IP address.
     */
    public String getHost() {
        return _service.getHost();
    }

    /**
     * Start the service running. This call never returns until the service
     * is shutdown.
     */
    public void run() {
        _log.debug("IPC Service Started");
        _service.start();
        _log.debug("IPC Service Stopped");
    }

} // --Server
