/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2000,2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: TcpServerService.java,v 1.7 2003/07/12 00:23:57 tanderson Exp $
 *
 * Date         Author  Changes
 * $Date	    jimm    Created
 */
package org.exolab.core.ipc;

import java.io.IOException;
import java.io.Serializable;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.HashMap;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;


/**
 * Create a listener socket and wait for requests to come in. When a connection
 * is made add the new object to the list of active connections and spawn a 
 * new thread to listen to the connection.
 *
 * @version     $Revision: 1.7 $ $Date: 2003/07/12 00:23:57 $
 * @author      <a href="mailto:mourikis@exolab.org">Jim Mourikis</a>
 * @see         org.exolab.core.ipc.IpcServerService
 * @see         org.exolab.core.ipc.TcpService
 * @see         org.exolab.core.ipc.Server
 */
public class TcpServerService implements IpcServerIfc {

    /**
     * The main server socket to listen for incoming calls.
     */
    private ServerSocket _server;
	
    /**
     * The list of all current connections to the server.
     */
    private HashMap _connections;

    /**
     * While the active flag is set, the server continues to run.
     */
    private boolean _active = true;

    /**
     * The client callback, called for all data received.
     */
    private NotifierIfc _notifier = null;

    /**
     *  The logger
     */
    private static final Log _log = LogFactory.getLog(TcpServerService.class);


    /**
     * Sets up a listener port on the next free port as allocated by the 
     * system and adds a notifier to call with all received messages.
     *
     * @param notifier the client handle to call when data is received.
     */
    public TcpServerService(NotifierIfc notifier) throws IOException {
        this(0, notifier);
    }

    /**
     * Sets up a listener port and adds a notifier to call with all received
     * messages.
     *
     * @param port the port to accept connections on.
     * @param notifier the client handle to call when data is received.
     */
    public TcpServerService(int port, NotifierIfc notifier) 
        throws IOException {
        this(null, port, notifier);
    }

    /**
     * Sets up a listener port and adds a notifier to call with all received
     * messages.
     *
     * @param address the local IP address to bind to. May be null.
     * @param port the port to accept connections on.
     * @param notifier the client handle to call when data is received.
     */
    public TcpServerService(InetAddress address, int port, 
                            NotifierIfc notifier) throws IOException {
        // Give it a reasonable number for performance.
        _connections = new HashMap(50);
        _notifier = notifier;
        _server = new ServerSocket(port, 50, address);
    }

    /**
     * Start the service running. This call never returns until the service
     * is shutdown.
     *
     * When a connection is made, the new socket is added to the connections
     * list, and a new thread is spawned to receive messages from the socket.
     *
     * @throws IOException If a failure on connection occurs.
     */
    public void start() {
        if (_log.isDebugEnabled()) {
            _log.debug("Server is ready, on port " + _server.getLocalPort());
        }
        while (_active) {
            try {
                Socket s = _server.accept();
                new Notifier(_notifier, new TcpService(s), _connections);
            } catch(IOException exception) {
                _log.debug(exception.getMessage(), exception);
            }
        }
    }

    /**
     * Shutdown all active connections, including the server connection.
     * 
     * @param ipc The service to shutdown.
     * @throws IOException If the service reports an error during shutdown.
     */
    public void shutdownAll() throws IOException {
        Object[] connections = _connections.values().toArray();

        for (int i = 0, j = _connections.size(); i < j; i++) {
            ((Notifier)connections[i]).stop();
        }
        _connections.clear();
        _active = false;
        _server.close();
    }

    /**
     * Return the port number this server is listening on.
     * 
     * @return int The port number the server is using.
     */
    public int getPort() {
        return _server.getLocalPort();
    }

    /**
     * Return the machines address.
     * 
     * @return String The machines Ip address.
     */
    public String getHost() {
        String host = TcpService.LOCAL_HOST;
        
        try {
            host = InetAddress.getLocalHost().getHostAddress();
        } catch(UnknownHostException exception) {
            _log.debug("Failed to get IP address for this host");
        }
        return host;
    }

} //-- TcpServerService
