/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2000 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: TcpService.java,v 1.5 2001/05/20 10:49:35 jima Exp $
 *
 * Date         Author  Changes
 * $Date	    jimm    Created
 */


package org.exolab.core.ipc;

import java.net.Socket;
import java.net.SocketException;
import java.io.DataOutputStream;
import java.io.DataInputStream;
import java.io.ByteArrayOutputStream;
import java.io.ObjectOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.io.IOException;
import java.lang.ClassNotFoundException;
import java.rmi.UnknownHostException;


/**
 * Implement a standard Tcp service, with one send and one receive buffer.
 * add a 10 second timeout on the send buffer in case of failures.
 * Receive buffers are blocking, but can have any parameters set as required
 * by getting the underlying soket and setting parameters as required.
 *
 * @version     $Revision: 1.5 $ $Date: 2001/05/20 10:49:35 $
 * @author      <a href="mailto:mourikis@exolab.org">Jim Mourikis</a>
 * @see         org.exolab.core.ipc.IpcIfc
 * @see			org.exolab.core.ipc.Client
 * @see			org.exolab.core.ipc.TcpServerService
 *
 **/


public class TcpService implements IpcIfc {
    // The Tcp socket
    private Socket              sock_;

    // The input stream
    private DataInputStream     input_ = null;

    // The output stream
    private DataOutputStream    output_ = null;

    // Timeout on socket set to 10 secs
    public static final int     TIMEOUT = 10000;

    // The local host uses loopback
    public static final String  LOCAL_HOST = "127.0.0.1";

    // Default port number
    public static final int     DEFAULT_PORT = 3030;

    /**
     * Default Constructor connects to the local machine on the default port
     * address.
     *
     */
    public TcpService()
    throws UnknownHostException, IOException
    {
        this(LOCAL_HOST, DEFAULT_PORT);
    }

    /**
     * Attempts to connect to the given host on the given port address.
     *
     * @param host The host to connect to.
     * @param port The port to connect on.
     * @exception TcpConnectionError if the connection fails.
     *
     */
    public TcpService(String host, int port)
    throws UnknownHostException, IOException
    {
        this(new Socket(host, port));
        try {
            sock_.setSoTimeout(TIMEOUT);
            sock_.setSoLinger(true, 30);
        } catch (SocketException err) {
            System.err.println("Failed to set socket options: " + 
                               err.getMessage());
        }
    }


    /**
     * Constructor keeps socket reference, and initialises input output
     * streams for the socket.
     *
     * @param sock The socket to monitor.
     *
     */
    public TcpService(Socket sock) throws IOException
    {
        sock_ = sock;
        input_ = new DataInputStream(sock_.getInputStream());
        output_ = new DataOutputStream(sock_.getOutputStream());
    }


    /**
     * Return the socket reference. This should only be used for setting
     * socket options. DO not shutdown the socket directly. Use the close
     * option below.
     *
     * @return The socket reference this TcpService object contains.
     *
     */
    public Socket getSocket()
    {
        return sock_;
    }


    /**
     * Close the socket connection, and perform any necessary cleanup.
     *
     * @exception IOException If the socket encounters an error.
     *
     */
    public void close() throws IOException
    {
        sock_.close();
    }


    /**
     * Send the specified object. 
     * First serialize the object, then send the length of the data followed 
     * by the data stream.
     *
     * @param ob The object to send. Must be serializable.
     * @exception IOException If the data fails to write to the output stream.
     *
     */
    public void send(Serializable ob) throws IOException
    {
        ByteArrayOutputStream bytes = new ByteArrayOutputStream();
        ObjectOutputStream out = new ObjectOutputStream(bytes);
        out.writeObject(ob);
        byte[] buffer = bytes.toByteArray();
        out.close();
        output_.writeInt(buffer.length);
        output_.write(buffer);
    }


    /**
     * An object is received on the socket.
     * Read the size, then read the data.
     *
     * @return Object The object received.
     * @exception IOException If the data cannot be received.
     * @exception ClassNotFoundException If the class cannot be re-constituted.
     *
     */
    public Object receive() throws IOException, ClassNotFoundException
    {
        Object ob = null;

        int n = input_.readInt();
        byte[] buffer = new byte[n];
        input_.readFully(buffer);
        ByteArrayInputStream bytes = new ByteArrayInputStream(buffer);
        ObjectInputStream in = new ObjectInputStream(bytes);
        ob = in.readObject();
        in.close();
       return ob;
    }



    /**
     * Retrieve a unique id for this connection.
     *
     * @param String The unique string id
     */
    public String toString()
    {
        return String.valueOf(sock_.getPort());
    }

} // End TcpService
