/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2002 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: Channel.java,v 1.3 2002/02/20 10:23:48 tima Exp $
 */
package org.exolab.core.messenger;

// java
import java.rmi.RemoteException;


/**
 *
 * @version     $Revision: 1.3 $ $Date: 2002/02/20 10:23:48 $
 * @author      <a href="mailto:tima@intalio.com">Tim Anderson</a>
 * @see         Connection
 * @see         ConnectionFactory
 * @see         Messenger
 */
public interface Channel {

    /**
     * Send a message
     *
     * @param message the object to send
     * @throws RemoteException if the message can't be sent
     */
    public void send(Object message) throws RemoteException;

    /**
     * Receive a message. This method blocks until a message is received.
     * 
     * @return the received message
     * @throws ClassNotFoundException if the class of a serialized object 
     * cannot be found
     * @throws RemoteException if the message can't be received
     */
    public Object receive() throws ClassNotFoundException, RemoteException;

    /**
     * Receive the next message that arrives within the specified timeout 
     * interval.<p>
     * This call blocks until a message arrives, the timeout expires, or 
     * the channel is closed. A timeout of zero never expires and the call 
     * blocks indefinitely.
     * 
     * @param timeout the timeout value (in milliseconds)
     * @return the received message
     * @throws RemoteException if a message can't be received
     * @throws ReceiveTimeoutException if no message is available within
     * the specified interval
     */
    public Object receive(long timeout) throws RemoteException;

    /**
     * Send a message and wait for a response.
     *
     * @param request the object to send
     * @return the response
     * @throws Exception if the server throws an exception
     * @throws RemoteException if the message can't be sent or the response 
     * can't be received
     */
    public Object invoke(Object request) throws Exception, RemoteException;

    /**
     * Register a listener to process messages asynchronously.<br/>
     * Only a single listener may be registered at any one time.<br/>
     * Setting the listener to null deregisters the listener.<p/>>
     * NOTE: the listener will be invoked in a separate thread - to maintain 
     * the single threaded nature of channels, it is an error to call 
     * {@link #send}, {@link #receive}, or {@link #invoke} from any other
     * thread other than that running the listener, while the listener is 
     * registered.
     *
     * @param listener the listener to process messages, or null,
     * to deregister the current listener
     */
    public void setChannelListener(ChannelListener listener);

    /**
     * Close the channel
     *
     * @throws RemoteException if an error occurs closing the channel
     */
    public void close() throws RemoteException;

} //-- Channel
