/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2002 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: Connection.java,v 1.4 2002/02/21 09:49:41 tima Exp $
 */
package org.exolab.core.messenger;

// java
import java.rmi.RemoteException;


/**
 * A <code>Connection</code> represents a logical connection over a shared
 * physical connection, managed by a {@link ManagedConnection}
 *
 * @version     $Revision: 1.4 $ $Date: 2002/02/21 09:49:41 $
 * @author      <a href="mailto:tima@intalio.com">Tim Anderson</a>
 * @see         Channel
 * @see         ChannelHandler
 * @see         ManagedConnection
 */
public interface Connection {

    /**
     * Open a channel
     *
     * @param name the name of the channel
     * @return the open channel
     * @throws RemoteException if the channel can't be opened
     * @throws InvalidChannelException if the channel doesn't exist
     */
    public Channel open(String name) throws RemoteException;

    /**
     * Register a handler to accept channel connections for the specified name.
     * <br/>
     * This method returns immediately.
     *
     * @param name the channel name to listen on
     * @param handler the handler to pass new channels to
     * @throws RemoteException if an error occurs accepting channels
     * @throws IllegalArgumentException if any argument is null
     */
    public void accept(String name, ChannelHandler handler) 
        throws RemoteException;

    /**
     * Register a handler to accept a single channel connection for the 
     * specified name, and bind the listener to it <br/>
     * This method returns immediately.
     *
     * @param name the channel name to listen on
     * @param listener the listener to handle requests on the channel
     * @throws RemoteException if an error occurs accepting channels
     * @throws IllegalArgumentException if any argument is null
     */
    public void accept(String name, ChannelListener listener) 
        throws RemoteException;

    /**
     * Stop accepting requests on the specified channel
     *
     * @param name the channel name
     * @throws IllegalArgumentException if <code>name</code> is null
     */
    public void close(String name);

    /**
     * Close the connection
     * <p>
     * This closes all channels open on the connection, prior to notifying
     * registered {@link ConnectionEventListener} instances
     *
     * @throws RemoteException if an error occurs closing the connection
     */
    public void close() throws RemoteException;

    /**
     * Add a listener for connection events
     *
     * @param listener the listener to add
     */
    public void addConnectionEventListener(ConnectionEventListener listener);

    /**
     * Remove a listener of connection events
     *
     * @param listener the listener to remove
     */
    public void removeConnectionEventListener(
        ConnectionEventListener listener);
    
} //-- Connection
