/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2002-2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: ConnectionDispatcher.java,v 1.5 2003/06/09 06:28:40 tanderson Exp $
 */
package org.exolab.core.messenger;

import java.net.MalformedURLException;
import java.rmi.RemoteException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;


/**
 * This class listens for connections using an {@link ConnectionAcceptor} and 
 * notifies the registered {@link ConnectionHandler} when a connection is 
 * made.
 *
 * @version     $Revision: 1.5 $ $Date: 2003/06/09 06:28:40 $
 * @author      <a href="mailto:tima@intalio.com">Tim Anderson</a>
 * @see         ConnectionAcceptor
 * @see         ConnectionHandler
 */
class ConnectionDispatcher implements Runnable {

    /**
     * The connection acceptor
     */
    private ManagedConnectionAcceptor _acceptor;

    /**
     * The dispatcher thread
     */
    private Thread _thread;

    /**
     * Flag to indicate that the dispatcher has been closed
     */
    private boolean _closed = false;

    /**
     * The logger
     */
    private static final Log _log =
        LogFactory.getLog(ConnectionDispatcher.class);


    /**
     * Construct a new <code>ConnectionDispatcher</code>
     *
     * @param acceptor the acceptor to listen for connections
     * @throws IllegalArgumentException if <code>acceptor</code> is null
     */
    public ConnectionDispatcher(ManagedConnectionAcceptor acceptor) {
        if (acceptor == null) {
            throw new IllegalArgumentException("Argument 'acceptor' is null");
        }
        _acceptor = acceptor;
    }
    
    /**
     * Listen for connections, and dispatch them to the registered handler.<p>
     * If an error occurs accepting a connection, the handler will be notified
     * and the dispatcher will terminate
     */
    public synchronized void dispatch() {
        if (_thread == null) {
            String name = "ConnectionDispatcher: " + _acceptor.getURI();
            _thread = new Thread(this, name);
            _thread.start();
        }
    }

    /**
     * Register a connection handler to accept new connections
     *
     * @param uri the URI that the handler accepts connections for
     * @param handler the connection handler
     * @throws MalformedURLException if <code>uri</code> is an invalid URI
     * @throws RemoteException if the handler cannot be registered
     * @throws IllegalArgumentException if any argument is null
     */
    public synchronized void accept(String uri, ConnectionHandler handler) 
        throws MalformedURLException, RemoteException {
        if (_closed) {
            throw new MessengerException("Connection dispatcher is closed");
        }
        _acceptor.accept(uri, handler);
    }

    /**
     * Stop accepting connections for the specified URI
     *
     * @param uri the URI to close
     * @throws RemoteException if an error occurs closing the connection
     * @throws IllegalArgumentException if <code>uri</code> is null
     */
    public synchronized void close(String uri) 
        throws MalformedURLException, RemoteException {
        if (!_closed) {
            _acceptor.close(uri);
        }
    }

    /**
     * Accept connections and pass them to the registered handler
     */
    public void run() {
        while (!closed()) {
            try {
                ManagedConnection connection = _acceptor.accept();
            } catch (RemoteException exception) {
                // stopping the acceptor will typically generate an exception,
                // so only log errors that haven't arisen as a result
                synchronized (this) {
                    if (!_closed) {
                        _log.debug(exception, exception);
                        _closed = true;
                    }
                }
            }
        }
        // clean up
        _acceptor = null;
    }

    /**
     * Close the dispatcher, freeing all associated resources
     */
    public synchronized void close() {
        if (!_closed) {
            _closed = true;
            try {
                _acceptor.close();
            } catch (RemoteException exception) {
                _log.debug(exception, exception);
            }
        }
    }

    /**
     * Returns <code>true</code> if the dispatcher is closed
     */
    public synchronized boolean closed() {
        return _closed;
    }

} //-- ConnectionDispatcher
