/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2002 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: ConnectionEventNotifier.java,v 1.3 2002/02/20 10:23:48 tima Exp $
 */
package org.exolab.core.messenger;

// java
import java.util.LinkedList;
import java.util.List;


/**
 * A helper class for managing {@link ConnectionEventListener} instances
 * and notifying them of any events
 *
 * @version     $Revision: 1.3 $ $Date: 2002/02/20 10:23:48 $
 * @author      <a href="mailto:tima@intalio.com">Tim Anderson</a>
 * @see         Connection
 */
public class ConnectionEventNotifier {

    /**
     * The event listeners
     */
    private List _listeners = new LinkedList();


    /**
     * Add a listener for connection events
     *
     * @param listener the listener to add
     */
    public void addConnectionEventListener(ConnectionEventListener listener) {
        synchronized (_listeners) {
            _listeners.add(listener);
        }
    }

    /**
     * Remove a listener of connection events
     *
     * @param listener the listener to remove
     */
    public void removeConnectionEventListener(
        ConnectionEventListener listener) {
        synchronized (_listeners) {
            _listeners.remove(listener);
        }
    }

    /**
     * Invoked when the connection is closed
     *
     * @param connection the closed connection
     */
    public void closed(Connection connection) {
        Object[] listeners;
        synchronized (_listeners) {
            listeners = _listeners.toArray();
        }

        // notify the listeners
        for (int i = 0; i < listeners.length; ++i) {
            ConnectionEventListener listener = 
                (ConnectionEventListener) listeners[i];
            listener.closed(connection);
        } 
    }

    /**
     * Invoked when a fatal connection error occurs, just before an Exception
     * is thrown to the application
     *
     * @param connection the connection the error occurred on
     * @param exception the error
    */
    public void error(Connection connection, Exception exception) {
        Object[] listeners;
        synchronized (_listeners) {
            listeners = _listeners.toArray();
        }

        // notify the listeners
        for (int i = 0; i < listeners.length; ++i) {
            ConnectionEventListener listener = 
                (ConnectionEventListener) listeners[i];
            listener.error(connection, exception);
        } 
    }

} //-- ConnectionEventListener
