/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2002 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: ManagedConnection.java,v 1.2 2002/02/21 09:49:41 tima Exp $
 */
package org.exolab.core.messenger;

// java
import java.rmi.RemoteException;


/**
 * A <code>ManagedConnection</code> instance represents a physical connection 
 * over a low level transport protocol.<p/>
 * It is responsible for establishing the connection, and managing its 
 * lifecyle. The physical connection is shared amongst multiple 
 * {@link Connection} instances.
 *
 * @version     $Revision: 1.2 $ $Date: 2002/02/21 09:49:41 $
 * @author      <a href="mailto:tima@intalio.com">Tim Anderson</a>
 * @see         Connection
 * @see         Messenger
 */
public interface ManagedConnection {

    /**
     * Creates a new connection handle for the underlying physical connection
     * represented by the ManagedConnection instance
     *
     * @param name the name of the connection
     * @param properties connection properties (may be null)
     * @return a new conneciton handle
     * @throws ConnectException if no handler is registered for 
     * <code>name</code>
     * @throws RemoteException if a new connection handle cannot be created
     */
    public Connection open(String name, ConnectionProperties properties) 
        throws RemoteException;

    /**
     * Register a handler to accept connections for the specified name.
     * <br/>
     * This method returns immediately.
     *
     * @param name the connection name to listen on
     * @param handler the handler to pass new connections to
     * @throws RemoteException if an error occurs registering the handler
     * @throws IllegalArgumentException if any argument is null
     */
    public void accept(String name, ConnectionHandler handler) 
        throws RemoteException;

    /**
     * Stop accepting connect requests for the specified connection
     *
     * @param name the connection name
     * @throws IllegalArgumentException if <code>name</code> is null
     */
    public void close(String name);

    /**
     * Destroys the physical connection
     *
     * @throws RemoteException if an error occurs closing the connection
     */
    public void close() throws RemoteException;

    /**
     * Set the listener for connection events
     *
     * @param listener the event listener. Setting it to <code>null</code>
     * removes the listener
     */
    public void setManagedConnectionEventListener(
        ManagedConnectionEventListener listener);
    
} //-- ManagedConnection
