/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2002 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: ManagedConnectionAcceptor.java,v 1.1 2002/02/21 09:46:42 tima Exp $
 */
package org.exolab.core.messenger;

// java
import java.net.MalformedURLException;
import java.rmi.RemoteException;


/**
 * A <code>ManagedConnectionAcceptor</code> waits for a connection request to 
 * come in over the network, and constructs a {@link ManagedConnection} 
 * object to handle it.
 *
 * @version     $Revision: 1.1 $ $Date: 2002/02/21 09:46:42 $
 * @author      <a href="mailto:tima@intalio.com">Tim Anderson</a>
 * @see         ManagedConnection
 * @see         Messenger
 */
public interface ManagedConnectionAcceptor {

    /**
     * Returns the URI that connections are being accepted on
     *
     * @return the URI that connections are being accepted on
     */
    public String getURI();

    /**
     * Register a connection handler to accept new {@link Connection}
     * instances
     *
     * @param uri the URI that the handler accepts connections for
     * @param handler the connection handler
     * @throws MalformedURLException if <code>uri</code> is an invalid URI
     * @throws RemoteException if the handler cannot be registered
     * @throws IllegalArgumentException if any argument is null
     */
    public void accept(String uri, ConnectionHandler handler) 
        throws MalformedURLException, RemoteException;

    /**
     * Stop accepting connections for the specified URI
     *
     * @param uri the URI to close
     * @throws MalformedURLException if <code>uri</code> is an invalid URI
     * @throws RemoteException if an error occurs closing the connection
     * @throws IllegalArgumentException if <code>uri</code> is null
     */
    public void close(String uri) 
        throws MalformedURLException, RemoteException;

    /**
     * Listens for a connection to be made to this acceptor and accepts it
     *
     * @return the new connection
     * @throws RemoteException if an error occurs when waiting for a 
     * connection
     */
    public ManagedConnection accept() throws RemoteException;

    /**
     * Close the acceptor, freeing all associated resources
     *
     * @throws RemoteException if an error occurs
     */
    public void close() throws RemoteException;

    /**
     * Add a listener for connection acceptor events
     *
     * @param listener the listener to add
     */
    public void addAcceptorEventListener(AcceptorEventListener listener);

    /**
     * Remove a listener of connection acceptor events
     *
     * @param listener the listener to remove
     */
    public void removeAcceptorEventListener(AcceptorEventListener listener);

} //-- ManagedConnectionAcceptor
