/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2002,2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: ManagedPacketConnection.java,v 1.3 2003/06/09 06:28:40 tanderson Exp $
 */
package org.exolab.core.messenger;

import java.rmi.ConnectIOException;
import java.rmi.RemoteException;
import java.rmi.ServerException;
import java.io.Serializable;
import java.util.HashMap;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.exolab.core.threadPool.ThreadPool;


/**
 * A <code>ManagedPacketConnection</code> implements an efficient packet 
 * based protocol over a low level transport protocol.
 *
 * @version     $Revision: 1.3 $ $Date: 2003/06/09 06:28:40 $
 * @author      <a href="mailto:tima@intalio.com">Tim Anderson</a>
 * @see         PacketConnection
 * @see         Messenger
 */
public class ManagedPacketConnection 
    implements ManagedConnection, ConnectionEventListener {

    /**
     * The multiplexer factory
     */
    private MultiplexerFactory _factory;

    /**
     * The multiplexer
     */
    private Multiplexer _multiplexer;

    /**
     * The demultiplexer
     */
    private Demultiplexer _demultiplexer;

    /**
     * The system connection
     */
    private PacketConnection _system;

    /**
     * The channel for opening and closing connections
     */
    private PacketChannel _connectionChannel;

    /**
     * The channel for accepting connections
     */
    private PacketChannel _connectionAcceptChannel;

    /**
     * The handshaking channel
     */
    private PacketChannel _handshakeChannel;

    /**
     * The set of open connections, keyed on connection id
     */
    private HashMap _connections = new HashMap();

    /**
     * The queue of outbound packets
     */
    private PacketQueue _queue;

    /**
     * The thread pool
     */
    private ThreadPool _threads;

    /**
     * The pool of packets
     */
    private PacketPool _pool;

    /**
     * The connection event listener
     */
    private ManagedConnectionEventListener _listener;

    /**
     * The connection acceptor
     */
    private PacketConnectionAcceptor _acceptor;

    /**
     * Flag to mark this connection as closed
     */
    private boolean _closed = false;

    /**
     * Flag to mark this connection as invalid. This occurs when the 
     * multiplexer/demultiplexer terminate unexpectedly
     */
    private boolean _invalid = false;

    /**
     * The seed for connection identifiers
     */
    private int _connectionIdSeed = SYSTEM_CONNECTION_ID + 1;

    /**
     * The system connection identifier
     */
    private static final int SYSTEM_CONNECTION_ID = 0;

    /**
     * The logging category
     */
    private static final Log _log =
        LogFactory.getLog(ManagedPacketConnection.class);


    /**
     * Construct a new <code>ManagedPacketConnection</code>
     *
     * @param context the connection context
     * @param factory the factory for multiplexers/demultiplexers
     * @throws IllegalArgumentException if any argument is null
     */
    public ManagedPacketConnection(Context context, 
                                   MultiplexerFactory factory) {
        if (context == null) {
            throw new IllegalArgumentException("Argument 'context' is null");
        }
        if (factory == null) {
            throw new IllegalArgumentException("Argument 'factory' is null");
        }

        _queue = new PacketQueue(1024);
        _threads = context.getThreads();
        _pool = new PacketPool(1024);
        _factory = factory;
        _acceptor = new PacketConnectionAcceptor(this);
    }


    /**
     * Creates a new connection handle for the underlying physical connection
     * represented by the ManagedConnection instance
     *
     * @param name the name of the connection
     * @param properties connection properties (may be null)
     * @return a new conneciton handle
     * @throws ConnectException if no handler is registered for 
     * <code>name</code>
     * @throws RemoteException if a new connection handle cannot be created
     */
    public Connection open(String name, ConnectionProperties properties) 
        throws RemoteException {
        PacketConnection result = null;
        ConnectRequest request = new ConnectRequest(name);
        Channel channel;
        synchronized (this) {
            ensureOpen();
            channel = _connectionChannel;            
        }
        // can't do the following in a synchronized block, as the invocation
        // may take a long time
        ConnectResponse response;
        try {
            response = (ConnectResponse) channel.invoke(request);
        } catch (RemoteException exception) {
            throw exception;    
        } catch (Exception exception) {
            throw new ServerException(exception.getMessage(), exception);
        }

        int connectionId = response.getConnectionId();
        
        synchronized (this) {
            ensureOpen(); // make sure the connection is still available
            result = new PacketConnection(this, connectionId);
            _connections.put(new Integer(connectionId), result);
            result.addConnectionEventListener(this);
        }
        return result;
    }

    /**
     * Register a handler to accept new connections.<br/>
     * This method returns immediately.
     *
     * @param handler the handler to pass new connections to
     * @throws RemoteException if an error occurs registering the handler
     * @throws IllegalArgumentException if any argument is null
     */
    public synchronized void accept(String name, ConnectionHandler handler) 
        throws RemoteException {
        ensureOpen();
        _acceptor.accept(name, handler);
    }

    /**
     * Stop accepting connect requests for the specified connection
     *
     * @param name the connection name
     * @throws IllegalArgumentException if <code>name</code> is null
     */
    public synchronized void close(String name) {
        if (_acceptor != null) {
            _acceptor.close(name);
        }
    }

    /**
     * Destroys the physical connection, prior to notifying any registered
     * {@link ManagedConnectionEventListener} instance
     *
     * @throws RemoteException if an error occurs closing the connection
     */
    public void close() throws RemoteException {
        ManagedConnectionEventListener listener;
        synchronized (this) {
            listener = _listener;
            doClose();
        }
        if (listener != null) {
            listener.closed(this);
        }
    }

    /**
     * Set the listener for connection events
     *
     * @param listener the event listener. Setting it to <code>null</code>
     * removes the listener
     */
    public synchronized void setManagedConnectionEventListener(
        ManagedConnectionEventListener listener) {
        _listener = listener;
    }

    /**
     * Invoked when a connection is closed
     *
     * @param connection the closed connection
     */
    public void closed(Connection connection) {
        boolean destroy = false;
        synchronized (this) {
            if (!_closed) {
                int connectionId = 
                    ((PacketConnection) connection).getConnectionId();
                _connections.remove(new Integer(connectionId));
                if (_connections.size() <= 1) {
                    // if no connections are left, or the only connection 
                    // remaining connection is the system connection, destroy 
                    // the physical connection.
                    destroy = true;
                }
            }
        }
        if (destroy) {
            try {
                close();
            } catch (RemoteException ignore) {
                _log.debug("Caught exception during close", ignore);
            }
        }
    }

    /**
     * Invoked when a fatal connection error occurs, just before an Exception
     * is thrown to the application
     *
     * @param connection the connection the error occurred on
     * @param exception the error
    */
    public void error(Connection connection, Exception exception) {
        _log.error("Error on connection", exception);
    }

    /**
     * Helper to return a channel for a particular connection
     *
     * @param connectionId the connection identifier
     * @param channelId the channel identifier
     * @return the channel for the supplied identifiers, or null, if no such
     * channel exists
     */
    protected Channel getChannel(int connectionId, int channelId) {
        PacketConnection connection = null;
        synchronized (this) {
            if (!_closed) {
                connection = (PacketConnection) _connections.get(
                    new Integer(connectionId));
            }
        }
        return (connection != null) ? connection.getChannel(channelId) : null;
    }

    /**
     * Initialises the connection
     *
     * @throws RemoteException if the connection can't be initialised
     */
    protected void initialise() throws RemoteException {
        _multiplexer = _factory.createMultiplexer(this);
        _demultiplexer = _factory.createDemultiplexer(this);

        // create the system connection
        _system = new PacketConnection(this, SYSTEM_CONNECTION_ID);
        _connections.put(new Integer(SYSTEM_CONNECTION_ID), _system);

        _connectionChannel = _system.createChannel(
            PacketChannel.CONNECTION_CHANNEL_ID, 
            PacketChannel.CONNECTION_ACCEPT_CHANNEL_ID);

        // register an acceptor to handle requests to open/close connections
        _connectionAcceptChannel = _system.createChannel(
            PacketChannel.CONNECTION_ACCEPT_CHANNEL_ID,
            PacketChannel.CONNECTION_CHANNEL_ID);
        _connectionAcceptChannel.setChannelListener(_acceptor);

        _handshakeChannel = _system.createChannel(
            PacketChannel.HANDSHAKE_CHANNEL_ID, 
            PacketChannel.HANDSHAKE_CHANNEL_ID);
                
        // start the multiplexer and demultiplexer. 
        try {
            synchronized (this) {
                // synchronize to make sure no close/error notifications are 
                // received while setting up
                _multiplexer.start();
                _demultiplexer.start();
            }

            // verify that the opposing connection is valid
            handshake(1000 * 10);
        } catch (RemoteException exception) {
            // make sure that the multiplexer/demultiplexer threads are stopped
            closeInternal();
            throw new RemoteException("guck", exception);
        }
    }

    /**
     * Performs the actual connection closure, without notifying registered
     * listeners
     *
     * @throws RemoteException if an error occurs closing the connection
     */
    protected synchronized void doClose() throws RemoteException {
        if (!_closed) {
            _closed = true; 
            closeInternal();
            _factory = null;
            if (_connectionChannel != null) {
                _connectionChannel.close();
                _connectionChannel = null;
            }
            if (_handshakeChannel != null) {
                _handshakeChannel.close();
                _handshakeChannel = null;
            }
            if (_system != null) {
                _system.close();
                _system = null;
            }
            if (_pool != null) {
                _pool.close();
                _pool = null;       
            }                
            _threads = null;
            _listener = null;
        }
    }

    /**
     * Returns the queue of packets to be streamed down the physical connection
     */
    protected synchronized PacketQueue getOutputQueue() {
        return _queue;
    }

    /**
     * Returns the pool of free packets
     */
    protected synchronized PacketPool getPool() {
        return _pool;
    }

    /**
     * Returns the thread pool
     */
    protected synchronized ThreadPool getThreadPool() {
        return _threads;
    }

    /**
     * Create a new connection to serve a client
     *
     * @param name the connection name
     * @throws RemoteException if the connection cannot be created
     * @throws IllegalArgumentException if <code>name</code> is null
     */
    protected synchronized PacketConnection createConnection(String name)
        throws RemoteException {
        if (name == null) {
            throw new IllegalArgumentException("Argument 'name' is null");
        }

        int connectionId = ++_connectionIdSeed;
        PacketConnection connection = new PacketConnection(this, connectionId);
        _connections.put(new Integer(connectionId), connection);
        return connection;
    }

    /**
     * Validate the connection protocol. This sends a handshake message
     * to the opposing connection, and waits for a response, for up to
     * <code>timeout</code> milliseconds. 
     *
     * @param timeout the maximum time to wait, in milliseconds
     * @throws RemoteException if no reponse is received in the specified time
     * interval, or an invalid packet is received
     */
    private void handshake(long timeout) throws RemoteException {
        _handshakeChannel.send(new Handshake());
        Object response = _handshakeChannel.receive(timeout);
        if (!(response instanceof Handshake)) {
            _log.debug("Invalid connection protocol: " +
                               "handshake on managed connection received a " +
                               response.getClass().getName());
            throw new MessengerException("Invalid connection protocol");
        }
    }

    /**
     * Close internal resources
     */
    private synchronized void closeInternal() {
        if (_multiplexer != null) {
            _multiplexer.close();
            _multiplexer = null;
        }
        if (_demultiplexer != null) {
            _demultiplexer.close();
            _demultiplexer = null;
        }
        if (_queue != null) {
            _queue.close();
            _queue = null;
        } 
    }

    protected synchronized void ensureOpen() throws RemoteException {
        if (_closed || _invalid) {
            throw new RemoteException("Connection is closed");
        }
    }

    /**
     * Handle a close event, raised by the Multiplexer or Demultiplexer
     */
    protected synchronized void streamClosed() {
        _log.debug("Multiplexer/Demultiplexer terminating");
        synchronized (this) {
            _invalid = true;
            closeInternal();
        }
    }

    /**
     * Handle an error event, raised by the Multiplexer or 
     * Demultiplexer. This closes the connection, prior to notifying any
     * registered listeners of the error.
     *
     * @param connection this connection
     * @param reason the exception describing the cause
     */
    protected void errorOnStream(Exception reason) {
        _log.debug("Multiplexer/Demultiplexer terminating on error", reason);
        synchronized (this) {
            _invalid = true;
            closeInternal();
        }
    }

    public static final class ConnectRequest implements Serializable {
        static final long serialVersionUID = 1;

        private String _name;

        public ConnectRequest(String name) {
            _name = name;
        }

        public String getName() {
            return _name;
        }

    } //-- ConnectRequest

    public static final class ConnectResponse implements Serializable {
        static final long serialVersionUID = 1;

        private int _connectionId = -1;

        public ConnectResponse() {
        }

        public ConnectResponse(int connectionId) {
            _connectionId = connectionId;
        }

        public int getConnectionId() {
            return _connectionId;
        }

    } //-- ConnectResponse

    private static final class Handshake implements Serializable {
        static final long serialVersionUID = 1;

        public Handshake() {
        }

    } //-- Handshake
    
} //-- ManagedPacketConnection

