/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2002,2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: Messenger.java,v 1.5 2003/06/09 06:28:40 tanderson Exp $
 */
package org.exolab.core.messenger;

import java.net.MalformedURLException;
import java.rmi.RemoteException;
import java.util.HashMap;

import org.apache.commons.logging.LogFactory;
import org.apache.commons.logging.Log;
import org.exolab.core.messenger.tcp.TcpConnector;


/**
 * 
 *
 * @version     $Revision: 1.5 $ $Date: 2003/06/09 06:28:40 $
 * @author      <a href="mailto:tima@intalio.com">Tim Anderson</a>
 */
public class Messenger {

    /**
     * The singleton instance
     */
    private static Messenger _instance = new Messenger();

    /**
     * Temporary hack
     */
    private Connector _tcp = new TcpConnector();

    /**
     * The set of active ConnectionDispatchers, keyed on 
     * ManagedConnectionAcceptor
     */
    private HashMap _dispatchers = new HashMap();

    /**
     * The set of ManagedConnectionAcceptors, keyed on URI
     */
    private HashMap _acceptors = new HashMap();
    
    /**
     * The logger
     */
    private static final Log _log =
        LogFactory.getLog(Messenger.class);

    /**
     * Construct a new Messenger
     */
    protected Messenger() {
    }

    /**
     * Returns the singleton instance of the Messenger
     *
     * @return the singleton instance
     */
    public static Messenger getInstance() {
        return _instance;
    }

    /**
     * Establish a connection to the specified URI.<p>
     * If a connection to the URI already exists, it will be returned.
     *
     * @param uri the URI to connect to
     * @return the new Connection
     * @throws MalformedURLException if <code>uri</code> is an invalid URI
     * @throws RemoteException if the connection cannot be opened
     * @throws IllegalArgumentException if <code>uri</code> is null
     */
    public Connection open(String uri) 
        throws MalformedURLException, RemoteException {
        return open(uri, null);
    }

    /**
     * Establish a connection to the specified URI.<p>
     * If a connection to the URI already exists, it will be returned.
     *
     * @param uri the URI to connect to
     * @param properties connection properties (may be null)
     * @return the new Connection
     * @throws MalformedURLException if <code>uri</code> is an invalid URI
     * @throws RemoteException if the connection cannot be opened
     * @throws IllegalArgumentException if <code>uri</code> is null
     */
    public Connection open(String uri, ConnectionProperties properties) 
        throws MalformedURLException, RemoteException {
        if (uri == null) {
            throw new IllegalArgumentException("Argument 'uri' is null");
        }
        Connection connection = null;
        if (uri.startsWith(_tcp.getScheme())) {
            connection = _tcp.getConnection(uri, properties);
        } else {
            throw new MalformedURLException("Unrecognised protocol for URI");
        }
        return connection;
    }


    /**
     * Accept connections for the specified URI, passing new connections
     * to the handler. <b>
     * This method returns immediately.
     *
     * @param uri the URI to listen on
     * @param handler the handler to pass new connections to
     * @throws MalformedURLException if <code>uri</code> is an invalid URI
     * @throws RemoteException if an error occurs accepting connections
     * @throws IllegalArgumentException if <code>uri</code> or 
     * <code>handler</code> are null
     */
    public void accept(String uri, ConnectionHandler handler) 
        throws MalformedURLException, RemoteException {
        accept(uri, handler, null);
    }
    
    /**
     * Accept connections for the specified URI, passing new connections
     * to the handler. <b>
     * This method returns immediately.
     *
     * @param uri the URI to listen on
     * @param handler the handler to pass new connections to
     * @param properties connection properties (may be null)
     * @throws AcceptorAlreadyRegisteredException if an 
     * {@link ConnectionAcceptor} is already listing on <code>uri</code>
     * @throws MalformedURLException if <code>uri</code> is an invalid URI
     * @throws RemoteException if an error occurs accepting connections
     * @throws IllegalArgumentException if <code>uri</code> or 
     * <code>handler</code> are null
     */
    public void accept(String uri, ConnectionHandler handler,
                       ConnectionProperties properties) 
        throws MalformedURLException, RemoteException {
        if (uri == null) {
            throw new IllegalArgumentException("Argument 'uri' is null");
        }
        Connection connection = null;
        if (!uri.startsWith(_tcp.getScheme())) {
            throw new MalformedURLException("Unrecognised protocol for URI");
        }

        ManagedConnectionAcceptor acceptor = _tcp.getManagedConnectionAcceptor(
            uri, properties);
        ConnectionDispatcher dispatcher;
        synchronized (_dispatchers) {
            dispatcher = (ConnectionDispatcher) _dispatchers.get(acceptor);
            if (dispatcher == null) {
                // the acceptor is new. Need to register it and construct
                // a new ConnectionDispatcher
                acceptor.addAcceptorEventListener(new EventHandler());
                dispatcher = new ConnectionDispatcher(acceptor);
                _acceptors.put(uri, acceptor);
                _dispatchers.put(acceptor, dispatcher);
                
                // start the dispatcher
                dispatcher.dispatch();
            }
        }
        dispatcher.accept(uri, handler);
    }

    /**
     * Stop accepting connections for the specified URI
     *
     * @param uri the URI to close
     * @throws MalformedURLException if <code>uri</code> is an invalid URI
     * @throws RemoteException if an error occurs closing the connection
     * @throws IllegalArgumentException if <code>uri</code> is null
     */
    public void close(String uri) 
        throws MalformedURLException, RemoteException {
        synchronized (_dispatchers) {
            ManagedConnectionAcceptor acceptor = 
                (ManagedConnectionAcceptor) _acceptors.get(uri);
            if (acceptor != null) {
                ConnectionDispatcher dispatcher = 
                    (ConnectionDispatcher) _dispatchers.get(acceptor);
                // TODO - close the acceptor if this is the URI represents
                // the last connection handler
                if (dispatcher != null) {                
                    dispatcher.close(uri);
                }
            }
        }
    }

    /**
     * Remove a connection acceptor and its associated dispatcher
     *
     * @param acceptor the connection acceptor to remove
     */
    private void remove(ManagedConnectionAcceptor acceptor) {
        synchronized (_dispatchers) {
            _acceptors.remove(acceptor.getURI());
            ConnectionDispatcher dispatcher = 
                (ConnectionDispatcher) _dispatchers.remove(acceptor);
            if (dispatcher != null) {
                dispatcher.close();
            }
        }
    }

    private class EventHandler implements AcceptorEventListener {
        
        public void closed(ManagedConnectionAcceptor acceptor) {
            remove(acceptor);
        }

        public void error(ManagedConnectionAcceptor acceptor, 
                          Exception exception) {
            remove(acceptor);
        }

    } //-- EventHandler

} //-- Messenger
