/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2002 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: MessengerException.java,v 1.2 2002/02/17 20:46:53 tima Exp $
 */
package org.exolab.core.messenger;

// java
import java.io.PrintStream;
import java.io.PrintWriter;
import java.rmi.RemoteException;


/**
 * Signals that a Messenger exception of some sort has occurred. 
 * 
 * @version     $Revision: 1.2 $ $Date: 2002/02/17 20:46:53 $
 * @author      <a href="mailto:tima@intalio.com">Tim Anderson</a>
 * @see         Messenger
 */
public class MessengerException extends RemoteException {

    /**
     * A reference to the nested exception. May be null.
     */
    private Throwable _exception;

    /**
     * Constructs a new exception with no detail message
     */
    public MessengerException() {
    }

    /**
     * Constructs a new exception with the specified error 
     * message. 
     *
     * @param message the error message
     */
    public MessengerException(String message) {
        this(message, null);
    }

    /**
     * Construct an exception with a link to the root cause of the exception. 
     * This is used when one exception is rethrown as another.
     *
     * @param reason the reason for the exception
     * @param exception the root cause of the exception
     */
    public MessengerException(Throwable exception) {
        this(null, exception);
    }
    
    /**
     * Construct an exception with the specified reason and a link to the
     * root cause of the exception. This is used when one exception is 
     * rethrown as another.
     *
     * @param message the error message
     * @param exception the root cause of the exception
     */
    public MessengerException(String message, Throwable exception) {
        super(message);
        _exception = exception;
    }

    /**
     * Return a reference to the root cause of the exception
     *
     * @return the root cause exception, or null
     */
    public Throwable getException() {
        return _exception;
    }

    // override Exception.printStackTrace()
    public void printStackTrace() {
        // print the stack trace first
        super.printStackTrace();

        // if there is a nested exception then print it's stack trace
        if (_exception != null) {
            _exception.printStackTrace();
        }
    }

    // override Exception.printStackTrace(PrintStream)
    public void printStackTrace(PrintStream stream) {
        // print the stack trace first
        super.printStackTrace(stream);

        // if there is a nested exception then print it's stack trace
        if (_exception != null) {
            _exception.printStackTrace(stream);
        }
    }

    // override Exception.printStackTrace(PrintWriter)
    public void printStackTrace(PrintWriter writer) {
        // print the stack trace first
        super.printStackTrace(writer);

        // if there is a nested exception then print it's stack trace
        if (_exception != null) {
            _exception.printStackTrace(writer);
        }
    }

    // override Exception.getMessage
    public String getMessage() {
        String message = super.getMessage();
        if (_exception != null && _exception.getMessage() != null) {
            if (message == null) {
                message = _exception.getMessage(); 
            }  else {
                message += " " + _exception.getMessage();
            }
        }
        return message;
    }

} //-- MessengerException
