/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2002,2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: Packet.java,v 1.4 2003/06/09 06:28:40 tanderson Exp $
 */
package org.exolab.core.messenger;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;


/**
 * A Packet contains a byte stream of serialized data, and is used to 
 * read/write chunks of data from/to streams. <br/>
 * It is used to break large serialized objects up into smaller pieces so they
 * can be sent on the wire, using the minimum number of resources. <p>
 * A Packet is re-useable, and is allocated and maintained by an 
 * {@link PacketPool}. This avoids large numbers of temporary objects being
 * created when serializing and deserializing objects.<p>
 * For efficiency, the byte stream is publicly accessible. It is the 
 * reponsibility of the caller to correctly populate it to avoid stream
 * corruption.
 *
 * @version     $Revision: 1.4 $
 * @author      <a href="mailto:tima@intalio.com">Tim Anderson</a>
 * @see         PacketPool
 * @see         PacketOutputStream
 * @see         PacketInputStream
 */
class Packet {

    /**
     * Type identifier for a 'send' or 'receive' packet
     */
    public static final byte TC_DATA = 0;

    /**
     * Type identifier for an 'invocation' packet
     */
    public static final byte TC_REQUEST = 1;

    /**
     * Type identifier for an 'invocation response' packet
     */
    public static final byte TC_RESPONSE = 2;

    /**
     * Type identifier for an 'exception' packet
     */
    public static final byte TC_EXCEPTION = 3;

    /**
     * The magic no. written at the start of the packet.
     */
    protected static final short PACKET_MAGIC = (short) 0xCAFE;

    /**
     * The packet version no.
     */
    protected static final short PACKET_VERSION = 1;

    protected static final int MAGIC_OFFSET = 0;
    protected static final int VERSION_OFFSET = 2;
    protected static final int SIZE_OFFSET = VERSION_OFFSET + 2;
    protected static final int TYPE_OFFSET = SIZE_OFFSET + 4;
    protected static final int CONNECTION_OFFSET = TYPE_OFFSET + 1;
    protected static final int CHANNEL_OFFSET = CONNECTION_OFFSET + 4;
    protected static final int DESTINATION_OFFSET = CHANNEL_OFFSET  + 4;
    protected static final int DATA_OFFSET = DESTINATION_OFFSET + 4;

    /**
     * The pool that owns this packet
     */
    private PacketPool _pool;

    /**
     * The serialized data
     */
    private byte[] _data;

    /**
     * The logging category
     */
    private static final Log _log =
        LogFactory.getLog(Packet.class);


    /**
     * Create a packet
     *
     * @param size the number of bytes to allocate to the serialized data
     * @param pool the owning pool (may be null)
     */
    protected Packet(int size, PacketPool pool) {
        _data = new byte[size];
        _pool = pool;
        setShort(MAGIC_OFFSET, PACKET_MAGIC);
        setShort(VERSION_OFFSET, PACKET_VERSION);
    }

    /**
     * Set the packet type
     *
     * @param type the packet type
     * @throws IllegalArgumentException if <code>type</code> is not a valid 
     * type
     */
    public void setType(byte type) {
        if (type < TC_DATA || type > TC_EXCEPTION) {
            throw new IllegalArgumentException("Argument 'type' is invalid");
        }
        _data[TYPE_OFFSET] = type;
    }

    /**
     * Get the packet type
     *
     * @return the packet type
     */
    public byte getType() {
        return _data[TYPE_OFFSET];
    }

    /**
     * Set the connection identifier
     *
     * @param id the connection identifier
     */
    public void setConnectionId(int id) {
        setInt(CONNECTION_OFFSET, id);
    }

    /**
     * Get the connection identifier
     *
     * @return the connection identifier
     */
    public int getConnectionId() {
        return getInt(CONNECTION_OFFSET);
    }

    /**
     * Set the channel identifier of the sender
     *
     * @param id the identifier of the sending channel
     */
    public void setChannelId(int id) {
        setInt(CHANNEL_OFFSET, id);
    }

    /**
     * Get the channel identifier of the sender
     *
     * @return the identifier of the sending channel
     */
    public int getChannelId() {
        return getInt(CHANNEL_OFFSET);
    }

    /**
     * Set the channel identifier of the receiver
     *
     * @param id the identifier of the receiving channel
     */
    public void setDestinationId(int id) {
        setInt(DESTINATION_OFFSET, id);
    }

    /**
     * Get the channel identifier of the receiver
     *
     * @return the identifier of the receiving channel
     */
    public int getDestinationId() {
        return getInt(DESTINATION_OFFSET);
    }

    /**
     * Set the size of the serialized data. This may be shorter
     * than the length of the array returned by {@link #getData}
     *
     * @param size the size of the serialized data, including the header
     * @throws IllegalArgumentException if size is greater than the length
     * of the byte array
     */
    public void setSize(int size) {
        if (size > _data.length) {
            throw new IllegalArgumentException("Argument 'size' is invalid");
        }
        setInt(SIZE_OFFSET, size);
    }

    /**
     * Get the length of the serialized data. This may be shorter
     * than the length of the array returned by {@link #getData}
     *
     * @return the length of the serialized data
     */
    public int getSize() {
        return getInt(SIZE_OFFSET);
    }

    /**
     * Get the serialized data. This contains the packet header as well
     * as user data. User data starts at the offset returned by 
     * {@link #getHeaderSize}
     *
     * @return the serialized data
     */
    public byte[] getData() {
        return _data;
    }

    /**
     * Get the size of the packet header
     *
     * @return the size of the packet header
     */
    public int getHeaderSize() {
        return DATA_OFFSET;
    }

    /**
     * Return this packet to the pool
     */
    public void release() {
        if (_pool != null) {
            _pool.release(this);
        }
    }

    /**
     * Read the contents of a packet from a stream
     *
     * @param stream the stream to read the packet from
     * @return the number of bytes actually read, or -1 for end-of-file
     * @throws IOException for any I/O error
     */
    public int read(InputStream stream) throws IOException {
        // read in the header
        int headerSize = getHeaderSize();
        if (headerSize > _data.length) {
            throw new IOException("Cannot destream packet. Packet too large");
        }
        int count = stream.read(_data, 0, headerSize);
        if (count == -1) {
            // end of file
        } else if (count != headerSize) {
            throw new IOException("Failed to read packet header. " +
                                  "Expected header size=" + headerSize + 
                                  ", but read=" + count);
        } else {
            checkHeader();
            int size = getSize();
            if (size > _data.length) {
                throw new IOException(
                    "Cannot destream packet. Packet too large");
            } else if (size < headerSize) {
                throw new IOException("Packet size is invalid");
            }

            // read the remaining packet data in
            int expect = size - headerSize;
            int actual = stream.read(_data, headerSize, expect);
            if (actual != expect && actual != -1) {
                int tmp = stream.read(_data, headerSize + actual, expect - actual);
                if ((tmp + actual) != expect) {
                    throw new IOException("Failed to read packet. " +
                                          "Expected to read size=" + expect + 
                                          ", but read=" + actual);
                }
            }
            count += expect;
        }
        return count;
    }

    /**
     * Write the contents of the packet to a stream
     *
     * @param stream the stream to write the packet to
     * @throws IOException for any I/O error
     */
    public void write(OutputStream stream) throws IOException {
        stream.write(_data, 0, getSize());
    }

    /**
     * Check the packet header
     *
     * @throws IOException if the packet header is invalid
     */
    protected void checkHeader() throws IOException {
        short magic = getShort(MAGIC_OFFSET);
        short version = getShort(VERSION_OFFSET);

        if (magic != PACKET_MAGIC) {
            throw new IOException("Stream does not contain a packet");
        }
        if (version != PACKET_VERSION) {
            throw new IOException("Packet version mismatch. Expected " +
                                  "version=" + PACKET_VERSION + 
                                  ", but got version=" + version);
        }
    }

    /**
     * Encodes a short value in the byte stream
     */
    protected void setShort(int offset, short value) {
        _data[offset] = (byte) ((value >>> 8) & 0xFF);
        _data[offset + 1] = (byte) ((value >>> 0) & 0xFF);
    }

    /**
     * Decodes a short value from the byte stream
     *
     * @param offset the offset in the byte stream 
     * @return the value at the specified offset
     */
    protected short getShort(int offset) {
        return (short) (((_data[offset] & 0xFF) << 8) +
            ((_data[offset + 1] & 0xFF) << 0));
    }

    /**
     * Encodes an integer value in the byte stream
     *
     * @param offset the offset in the byte stream 
     * @param value the value to set
     */
    protected void setInt(int offset, int value) {
	_data[offset] = (byte) ((value >>> 24) & 0xFF);
        _data[offset + 1] = (byte) ((value >>> 16) & 0xFF);
        _data[offset + 2] = (byte) ((value >>> 8) & 0xFF);
        _data[offset + 3] = (byte) ((value >>> 0) & 0xFF);
    }

    /**
     * Decodes an integer value from the byte stream
     *
     * @param offset the offset in the byte stream 
     * @return the value at the specified offset
     */
    protected int getInt(int offset) {
        return ((_data[offset] & 0xFF) << 24) + 
            ((_data[offset + 1] & 0xFF) << 16) +
            ((_data[offset + 2] & 0xFF) << 8) +
            ((_data[offset + 3] & 0xFF) << 0);
    }

    protected static void dump(String prefix, Packet packet) {
        int size = packet.getSize();
        byte[] data = packet.getData();
        StringBuffer buffer = new StringBuffer();
        buffer.append(prefix);
        buffer.append(" packet length=" + size + ", type=" + packet.getType() +
                      ", channel=" + packet.getChannelId() + ", data=");
        for (int i = 0; i < size; ++i) {
            int value = data[i] & 0xff;
            buffer.append(Integer.toHexString(value));
            buffer.append(" ");
        }
        _log.debug(buffer.toString());
    }

} //-- Packet
