/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2002,2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: PacketChannel.java,v 1.5 2003/06/09 06:28:40 tanderson Exp $
 */
package org.exolab.core.messenger;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.rmi.MarshalException;
import java.rmi.RemoteException;
import java.rmi.ServerError;
import java.rmi.UnexpectedException;
import java.rmi.UnmarshalException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.exolab.core.threadPool.CompletionListener;
import org.exolab.core.threadPool.ThreadPool;


/**
 * A PacketConnection is a {@link Connection} that minimises resource usage.<p>
 * This is achieved by:
 * <ul>
 *  <li>Restricting the number of physical connections by sharing them
 *      between multiple {@link Channel} instances</li>
 *  <li>Breaking large messages into packets, allowing multiple channels
 *      to use the connection at the same time</li>
 *  <li>Optionally restricting the number of packets used by the connection,
 *      to avoid OutOfMemoryError exceptions being thrown during high activity
 *      on the connection</li>
 * </ul>
 *
 * @version     $Revision: 1.5 $ $Date: 2003/06/09 06:28:40 $
 * @author      <a href="mailto:tima@intalio.com">Tim Anderson</a>
 * @see         PacketConnection
 * @see         PacketPool
 * @see         PacketInputStream
 * @see         PacketOutputStream
 */
public class PacketChannel implements Channel {

    /**
     * The identifier for the connection channel
     */
    public static final int CONNECTION_CHANNEL_ID = 0;

    /** 
     * The identifier for the connection acceptor channel
     */
    public static final int CONNECTION_ACCEPT_CHANNEL_ID = 1;

    /**
     * The identifier for the system channel
     */
    public static final int SYSTEM_CHANNEL_ID = 2;

    /**
     * The identifier for the channel to accept open and close requests
     */
    public static final int ACCEPTOR_CHANNEL_ID = 3;

    /**
     * The identifier for the handshaking channel
     */
    public static final int HANDSHAKE_CHANNEL_ID = 4;

    /**
     * The last system channel identifier
     */
    public static final int LAST_SYSTEM_ID = HANDSHAKE_CHANNEL_ID;

    /**
     * The underlying connection
     */
    private PacketConnection _connection;

    /**
     * The channel identifier
     */
    private final int _channelId;

    /**
     * The destination channel identifier
     */
    private int _destinationId;

    /**
     * The packet queue for outbound packets. This is owned by the connection.
     */
    private PacketQueue _outQueue;

    /**
     * The packet queue for inbound packets. This is allocated by the 
     * connection, who passes ownership to the channel.
     */
    private PacketQueue _inQueue;

    /**
     * The pool of free packets. This is owned by the connection.
     */
    private PacketPool _pool;

    /**
     * The output stream to send packets to
     */
    private PacketOutputStream _out;

    /**
     * The input stream to receive packets on
     */
    private PacketInputStream _in;

    /**
     * The listener to process messages asynchronously
     */
    private ChannelListener _listener;

    /**
     * The thread pool, for asynchronously dispatching requests
     */
    private ThreadPool _threads;

    /**
     * Flag to indicate if the channel is closed
     */
    private boolean _closed = false;

    /**
     * Synchronization helper used to coordinate close(), and enqueue()
     */
    private final Object _lock = new Object();

    /**
     * Flag to indicate if the handler is scheduled to process a request
     */
    private boolean _scheduled = false;

    /**
     * Flag to indicate if a {@link ChannelListener#invoke} is in progress.
     * This is used to prevent the channel being used during an invocation.
     */
    private boolean _invoke = false;

    /**
     * The logger
     */
    private static final Log _log =
        LogFactory.getLog(PacketChannel.class);


    /**
     * Create a new channel
     *
     * @param connection the owning connection
     * @param channelId the channel identifier
     * @param destinationId the destination channel identifier
     * @param inputQueue the inbound packet queue, to deserialize messages from
     * @throws IllegalArgumentException if any argument is null
     */
    public PacketChannel(PacketConnection connection, int channelId, 
                         int destinationId, PacketQueue inputQueue) {
        if (connection == null) {
            throw new IllegalArgumentException(
                "Argument 'connection' is null");
        }
        if (inputQueue == null) {
            throw new IllegalArgumentException(
                "Argument 'inputQueue' is null");
        }

        _connection = connection;
        _channelId = channelId;
        _destinationId = destinationId; 
        _inQueue = inputQueue;
        _outQueue = _connection.getOutputQueue();
        _pool = _connection.getPool();
        _threads = _connection.getThreadPool();

        _out = new PacketOutputStream(_connection.getConnectionId(), 
                                      _channelId, _destinationId, _outQueue, 
                                      _pool);
        _in = new PacketInputStream(_inQueue);
    }

    /**
     * Returns the channel identifier. This is unique to a connection
     *
     * @return the channel identifier
     */
    public int getChannelId() {
        return _channelId;
    }

    /**
     * Send a message
     *
     * @param message the object to send
     * @throws RemoteException if the message can't be sent
     */
    public void send(Object message) throws RemoteException {
        send(message, Packet.TC_DATA, _destinationId);
    }

    /**
     * Receive a message. This method blocks until a message is received.
     * 
     * @return the received message
     * @throws RemoteException if the message can't be received
     */
    public Object receive() throws RemoteException {
        return receive(0);
    }

    /**
     * Receive the next message that arrives within the specified timeout 
     * interval.<p>
     * This call blocks until a message arrives, the timeout expires, or 
     * the channel is closed. A timeout of zero never expires and the call 
     * blocks indefinitely.
     * 
     * @param timeout the timeout value (in milliseconds)
     * @return the received message
     * @throws RemoteException if a message can't be received
     * @throws ReceiveTimeoutException if no message is available within
     * the specified interval
     */
    public synchronized Object receive(long timeout) throws RemoteException {
        if (_listener != null) {
            // if a listener is registered, throw an exception to avoid
            // synchronization problems.
            throw new MessengerException(
                "Cannot receive - a ChannelListener is registered");
        }

        Object message = doReceive(timeout);

        // successfully deserialized object - make sure it is of the
        // correct type
        byte type = _in.getType();
        if (type == Packet.TC_REQUEST) {
            throw new MessengerException(
                "Expected data packet, but got an invocation request");
        } else if (type == Packet.TC_RESPONSE) {
            throw new MessengerException(
                "Expected data packet, but got an invocation response");
        } else if (type == Packet.TC_EXCEPTION) {
            throw (RemoteException) message;
        }
        return message;
    }

    /**
     * Send a message and wait for a response
     *
     * @param request the object to send
     * @return the response
     * @throws Exception if the server throws an exception
     * @throws RemoteException if the message can't be sent or the response 
     * can't be received
     */
    public synchronized Object invoke(Object request) 
        throws Exception, RemoteException {
        Object response;
        if (_closed) {
            throw new ChannelClosedException();
        }
        if (_invoke) {
            throw new MessengerException(
                "Cannot call Channel.invoke() while ChannelListener.invoke() "
                + " in progress");
        }

        _out.setType(Packet.TC_REQUEST);
        try {
            ObjectOutputStream out = new ObjectOutputStream(_out);
            out.writeObject(request);
            out.close();
        } catch (IOException exception) {
            throw new MarshalException("Failed to perform invocation", 
                                       exception);
        }
        try {
            ObjectInputStream in = new ObjectInputStream(_in);
            response = in.readObject();
            in.close();
        } catch (Exception exception) {
            throw new UnmarshalException("Failed to deserialize response",
                                         exception);
        }

        // successfully deserialized object - make sure it is of the
        // correct type
        byte type = _in.getType();
        if (type == Packet.TC_DATA) {
            throw new MessengerException(
                "Expected reponse packet, but got an data packet");
        } else if (type == Packet.TC_REQUEST) {
            throw new MessengerException(
                "Expected reponse packet, but got an invocation request");
        } else if (type == Packet.TC_EXCEPTION) {
            throw (Exception) response;
        }
        return response;
    }

    /**
     * Register a listener to process messages asynchronously.<br/>
     * Only a single listener may be registered at any one time.<br/>
     * Setting the listener to null deregisters the listener.<p/>>
     * NOTE: the listener will be invoked in a separate thread - to maintain 
     * the single threaded nature of channels, it is an error to call 
     * {@link #send}, {@link #receive}, or {@link #invoke} from any other
     * thread other than that running the listener, while the listener is 
     * registered.
     *
     * @param listener the listener to process messages, or null,
     * to deregister the current listener
     */
    public synchronized void setChannelListener(ChannelListener listener) {
        _listener = listener;
    }
   
    /**
     * Close the channel. This notifies the opposing channel instance that this
     * has closed, and notifies the {@link ChannelListener} if one is 
     * registered.
     *
     * @throws RemoteException if an I/O error occurs closing the channel
     */
    public void close() throws RemoteException {
        synchronized (_lock) {
            if (!_closed) {
                _connection.closed(this);
            }
        }
        doClose();
    }

    /**
     * Returns <code>true</code> if this channel is closed
     */
    public synchronized boolean isClosed() {
        return _closed;
    }

    /**
     * Receive the next message that arrives within the specified timeout 
     * interval.<p>
     * This call blocks until a message arrives, the timeout expires, or 
     * the channel is closed. A timeout of zero never expires and the call 
     * blocks indefinitely.
     * 
     * @param timeout the timeout value (in milliseconds)
     * @return the received message
     * @throws RemoteException if a message can't be received
     * @throws ReceiveTimeoutException if no message is available within
     * the specified interval
     */
    protected synchronized Object doReceive(long timeout) 
        throws RemoteException {
        Object message;
        if (_closed) {
            throw new ChannelClosedException();
        }
        if (_invoke) {
            throw new MessengerException(
                "Cannot receive messages on channel while " +
                "ChannelListener.invoke() is in progress");
        }

        if (timeout != 0) {
            if (!_inQueue.waitForPacket(timeout)) {
                throw new ReceiveTimeoutException();
            }
        } else {
            if (!_inQueue.waitForPacket(0)) {
                throw new ChannelClosedException();
            }
        }

        try {
            ObjectInputStream in = new ObjectInputStream(_in);
            message = in.readObject();
            in.close();
        } catch (Exception exception) {
            throw new UnmarshalException("Failed to receive message",
                                         exception);
        }
        return message;
    }

    /**
     * Closes the channel, without notifying the opposing channel instance
     */
    protected void doClose() {
        closeListener();
        synchronized (this) {
            if (!_closed) {
                _closed = true;
                if (_listener != null) {
                    _listener.closed();
                }
                _connection = null;
                _outQueue = null;            
                _inQueue = null;
                _pool = null;
                _out = null;
                _in = null;
                _listener = null;
            }
        }
    }

    /**
     * Closes the listener
     */
    protected void closeListener() {
        synchronized (_lock) {
            if (!_closed) {
                _inQueue.close();
            }
        }
    }

    /**
     * Sets the 'invoke handler' flag.
     * This is used to prevent the channel being used during an invocation.
     *
     * @param invoke if <code>true</code> prevent the channel being used
     * for {@link #send}, {@link #receive}, and {@link #invoke}
     */
    protected synchronized void setInvoke(boolean invoke) {
        _invoke = invoke;
    }
    
    /**
     * Returns a reference to the owning connection
     */
    protected synchronized PacketConnection getConnection() {
        return _connection;
    }

    /**
     * Send a message to the specified destination channel
     *
     * @param message the message to send
     * @param type the message type
     * @param destinationId the destination channel id
     * @throws RemoteException if an I/O error occurs
     */
    protected synchronized void send(Object message, byte type, 
                                     int destinationId) 
        throws RemoteException {
        if (_closed) {
            throw new ChannelClosedException();
        }
        if (_invoke) {
            throw new MessengerException(
                "Cannot send messages on channel while " +
                "ChannelListener.invoke() is in progress");
        }

        _out.setType(type);
        _out.setDestinationId(destinationId);
        try {
            ObjectOutputStream out = new ObjectOutputStream(_out);
            out.writeObject((Serializable) message);
            out.close();
        } catch (IOException exception) {
            throw new MarshalException("Failed to send message", exception);
        }
    }

    /**
     * Enqueues a packet on the inbound queue, for de-serializing.<br>
     * If a {@link ChannelHandler} is registered, it will be queued for
     * notification using the thread pool
     *
     * @param packet the packet to queue on the inbound queue
     */
    protected void enqueue(Packet packet) {
        synchronized (_lock) {
            if (!_closed) {
                _inQueue.add(packet);
                if (_listener != null && !_scheduled) {
                    _scheduled = true;

                    ListenerInvoker invoker = new ListenerInvoker(
                        _listener, packet.getType(), packet.getChannelId());

                    _threads.queue(invoker, new CompletionHandler());
                }
            }
        }
    }

    protected synchronized void handlerCompleted() {
        if (!_closed) {
            Packet packet = _inQueue.peek();
            if (packet != null) {
                // re-queue the listener if another message is available
                ListenerInvoker invoker = new ListenerInvoker(
                    _listener, packet.getType(), packet.getChannelId());
                _threads.queue(invoker, new CompletionHandler());
            } else {
                _scheduled = false;
            }
        }
    }

    private class CompletionHandler implements CompletionListener {

        public void completed(Runnable target) {
            handlerCompleted();
        };

    } //-- CompletionHandler

    private class ListenerInvoker implements Runnable {

        private ChannelListener _listener;
        private byte _type;
        private int _destinationId;
        
        public ListenerInvoker(ChannelListener handler, 
                               byte type, int destinationId) {
            _listener = handler;
            _type = type;
            _destinationId = destinationId;
        }

        public void run() {
            try {
                synchronized (PacketChannel.this) {
                    Object request = doReceive(0);
                    synchronized (_lock) {
                        if (_type == Packet.TC_DATA) {
                            _listener.receive(request);
                        } else if (_type == Packet.TC_REQUEST) {
                            Object response;
                            try {
                                setInvoke(true);
                                response = _listener.invoke(request);
                            } finally {
                                setInvoke(false);
                            }
                            send(response, Packet.TC_RESPONSE, _destinationId);
                        } else {
                            // TODO - how to handle TC_EXCEPTION
                        }
                    }
                }
            } catch (ChannelClosedException ignore) {
            } catch (Exception exception) {
                if (!(exception instanceof RemoteException)) {
                    // wrap the exception
                    exception = new UnexpectedException(exception.getMessage(),
                                                        exception);
                }
                synchronized (_lock) {
                    if (!_closed) {
                        try {
                            _log.debug(
                                "sending exception on connection=" +
                                getConnection().getConnectionId() +
                                ", channel=" + _destinationId);
                            send(exception, Packet.TC_EXCEPTION, 
                                 _destinationId);
                        } catch (RemoteException ignore) {
                            // give up
                        }
                    }
                }
            }
        }

    } //-- ListenerInvoker

} //-- PacketChannel
