/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2002,2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: PacketChannelAcceptor.java,v 1.5 2003/06/09 06:28:40 tanderson Exp $
 */
package org.exolab.core.messenger;

import java.rmi.RemoteException;
import java.util.HashMap;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;


/**
 * A <code>ChannelAcceptor</code> that handles requests on an 
 * {@link SystemPacketChannel} to open or close user channels
 * 
 * @version     $Revision: 1.5 $ $Date: 2003/06/09 06:28:40 $
 * @author      <a href="mailto:tima@intalio.com">Tim Anderson</a>
 * @see         PacketChannel
 * @see         PacketConnection
 * @see         SystemPacketChannel
 */
class PacketChannelAcceptor implements ChannelAcceptor, ChannelListener {

    /**
     * The owning connection
     */
    private PacketConnection _connection;

    /**
     * The channel to receive requests on, and send responses
     */
     private PacketChannel _channel;

    /**
     * The set of ChannelHandler instances, keyed on name
     */
    private HashMap _handlers = new HashMap();

    /**
     * The set of open channel instances, keyed on destination id
     */
    private HashMap _channels = new HashMap();

    /**
     * The logger
     */
    private static final Log _log =
        LogFactory.getLog(PacketChannelAcceptor.class);


    /**
     * Construct a new <code>PacketChannelAcceptor</code>
     * 
     * @param connection the owning connection
     * @param channel the channel to accept requests on
     * @throws IllegalArgumentException if any argument is null
     */    
    public PacketChannelAcceptor(PacketConnection connection, 
                                 PacketChannel channel) {
        if (connection == null) {
            throw new IllegalArgumentException(
                "Argument 'connection' is null");
        }
        if (channel == null) {
            throw new IllegalArgumentException("Argument 'channel' is null");
        }
        _connection = connection;
        _channel = channel;
    }

    /**
     * Accept requests for the specified channel, and pass them on to
     * the handler.<br/>
     * This method returns immediately.
     *
     * @param name the channel name
     * @param handler the handler to pass new channels to
     * @throws RemoteException if a handler is already registered for the 
     * specified channel
     * @throws IllegalArgumentException if any argument is null
     */
    public synchronized void accept(String name, ChannelHandler handler) 
        throws RemoteException {
        if (name == null) {
            throw new IllegalArgumentException("Argument 'name' is null");
        }
        if (handler == null) {
            throw new IllegalArgumentException("Argument 'handler' is null");
        }
            
        if (_handlers.containsKey(name)) {
            throw new MessengerException(
                "A handler is already registered for " + name);
        }
        _handlers.put(name, handler);
        _log.debug("Connection id=" + _connection.getConnectionId() +
                           " accepting requests on channel=" + name);
    }

    /**
     * Accept a single channel connection for the specified name, and bind
     * the listener to it <br/>
     * This method returns immediately.
     *
     * @param name the channel name to listen on
     * @param listener the listener to handle requests on the channel
     * @throws RemoteException if an error occurs accepting channels
     * @throws IllegalArgumentException if any argument is null
     */
    public synchronized void accept(String name, ChannelListener listener) 
        throws RemoteException {
        accept(name, new ChannelListenerAcceptor(_connection, name, listener));
    }

    /**
     * Stop accepting requests on the specified channel
     *
     * @param name the channel name
     * @throws IllegalArgumentException if <code>name</code> is null
     */
    public synchronized void close(String name) {
        if (name == null) {
            throw new IllegalArgumentException("Argument 'name' is null");
        }
        _handlers.remove(name);
    }

    /**
     * Process a message sent by {@link Channel#send}
     *
     * @param message the message
     */
    public synchronized void receive(Object message) {
        if (message instanceof SystemPacketChannel.Closed) {
            closed((SystemPacketChannel.Closed) message);
        } else {
            _log.error("Cannot process messages of type=" +
                       message.getClass().getName());
        }
    }

    /**
     * Process an invocation request, and return a response
     *
     * @param request the request
     * @return the response
     * @throws RemoteException if the handler cannot process the request. This
     * is propagated back to the client.
     */
    public synchronized Object invoke(Object request) 
        throws RemoteException {
        Object response = null;
        if (request instanceof SystemPacketChannel.OpenRequest) {
            response = open((SystemPacketChannel.OpenRequest) request);
        } else {
            throw new MessengerException("Cannot process requests of type=" + 
                                         request.getClass().getName());
        }
        return response;
    }

    /**
     * Invoked when the underlying {@link SystemPacketChannel} is closed
     */
    public void closed() {
        close();
    }

    /**
     * Close the acceptor, freeing all associated resources
     */
    public synchronized void close() {
        if (_handlers != null) {
            _handlers.clear();
            _channels.clear();
            _connection = null;
            _channel = null;
            _handlers = null;
            _channels = null;
        }
    }

    /**
     * Open a new channel
     *
     * @param request the open request
     * @return the response to the request
     * @throws RemoteException if the channel is invalid
     */
    protected Object open(SystemPacketChannel.OpenRequest request) 
        throws RemoteException {
        ChannelHandler handler = 
            (ChannelHandler) _handlers.get(request.getName());
        if (handler == null) {
            throw new InvalidChannelException(
                "No ChannelHandler registered for channel=" + 
                request.getName());
        }
        int channelId = _connection.getNextChannelId();
        int destinationId = request.getChannelId();
            
        PacketChannel opened = _connection.createChannel(
            channelId, destinationId);
        
        // cache the reference to the new channel
        _channels.put(new Integer(destinationId), opened);

        // notify the handler
        handler.opened(opened);

        SystemPacketChannel.OpenResponse response = 
            new SystemPacketChannel.OpenResponse(channelId, destinationId);
        return response;
    }

    /**
     * Notify a Channel that its corresponding endpoint has been closed
     *
     * @param message the close notification message
     */
    protected void closed(SystemPacketChannel.Closed message) {
        Integer channelId = new Integer(message.getChannelId());
        PacketChannel channel = (PacketChannel) _channels.remove(channelId);
        channel.doClose();
    }

} //-- PacketChannelAcceptor
