/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2002,2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: PacketConnection.java,v 1.5 2003/06/09 06:28:40 tanderson Exp $
 */
package org.exolab.core.messenger;

import java.io.Serializable;
import java.rmi.RemoteException;
import java.util.HashMap;

import org.apache.commons.logging.LogFactory;
import org.apache.commons.logging.Log;

import org.exolab.core.threadPool.ThreadPool;


/**
 * A PacketConnection is a {@link Connection} that minimises resource usage.<p>
 * This is achieved by:
 * <ul>
 *  <li>Restricting the number of physical connections by sharing them
 *      between multiple {@link Channel} instances</li>
 *  <li>Breaking large messages into packets, allowing multiple channels
 *      to use the connection at the same time</li>
 *  <li>Optionally restricting the number of packets used by the connection,
 *      to avoid OutOfMemoryError exceptions being thrown during high activity
 *      on the connection</li>
 * </ul>
 *
 * @version     $Revision: 1.5 $ $Date: 2003/06/09 06:28:40 $
 * @author      <a href="mailto:tima@intalio.com">Tim Anderson</a>
 * @see         ConnectionEventListener
 * @see         PacketChannel
 * @see         PacketQueue
 * @see         PacketPool
 * @see         Multiplexer
 * @see         Demultiplexer
 */
public class PacketConnection implements Connection {

    /**
     * The managed connection that owns this
     */
    private ManagedPacketConnection _owner;

    /**
     * The identifier for this connection
     */
    private final int _connectionId;

    /**
     * The set of active channels, keyed on channel identifier
     */
    private HashMap _channels = new HashMap();

    /**
     * The system channel
     */
    private SystemPacketChannel _systemChannel;

    /**
     * The acceptor channel
     */
    private PacketChannel _acceptorChannel;

    /**
     * The acceptor
     */
    private ChannelAcceptor _acceptor;

    /**
     * The queue of outbound packets
     */
    private PacketQueue _queue;

    /**
     * The thread pool
     */
    private ThreadPool _threads;

    /**
     * The pool of packets
     */
    private PacketPool _pool;

    /**
     * The connection event listeners
     */
    private ConnectionEventNotifier _listeners = new ConnectionEventNotifier();

    /**
     * The seed for channel identifiers
     */
    private int _channelIdSeed = PacketChannel.LAST_SYSTEM_ID + 1;

    /**
     * Flag to mark this connection as closed
     */
    private boolean _closed = false;

    /**
     * The logging category
     */
    private static final Log _category =
        LogFactory.getLog("org.exolab.core.messenger");


    /**
     * Construct a new <code>PacketConnection</code>
     *
     * @param owner the managed connection that owns this
     * @param connectionId the connection identifier
     * @throws IllegalArgumentException if <code>owner</code> is null
     * @throws RemoteException if the connection can't be initialised
     */
    public PacketConnection(ManagedPacketConnection owner, int connectionId) 
        throws RemoteException {
        if (owner == null) {
            throw new IllegalArgumentException("Argument 'owner' is null");
        }
        _owner = owner;
        _connectionId = connectionId;

        _queue = _owner.getOutputQueue();
        _threads = _owner.getThreadPool();
        _pool = _owner.getPool();       

        // create the system channel
        _systemChannel = createSystemChannel();

        // create the acceptor channel
        _acceptorChannel = createChannel(PacketChannel.ACCEPTOR_CHANNEL_ID, 
                                         PacketChannel.SYSTEM_CHANNEL_ID);
        _acceptor = new PacketChannelAcceptor(this, _acceptorChannel);
        _acceptorChannel.setChannelListener((ChannelListener) _acceptor);
    }

    /**
     * Return the unique identifier for this connection
     *
     * @return the identifier for this connection
     */
    public int getConnectionId() {
        return _connectionId;
    }

    /**
     * Open a channel
     *
     * @param name the name of the channel
     * @return the open channel
     * @throws RemoteException if the channel can't be opened
     * @throws ConnectionClosedException if the connection is closed
     * @throws InvalidChannelException if the channel doesn't exist
     */
    public Channel open(String name) throws RemoteException {
        SystemPacketChannel channel;
        synchronized (this) {
            // the demultiplexer needs access to this, so synchronize
            // only to make sure the connection is open
            ensureOpen();
            channel = _systemChannel;
        }
        return channel.open(name);
    }

    /**
     * Register a handler to accept channel connections for the specified name.
     * <br/>
     * This method returns immediately.
     *
     * @param name the channel name to listen on
     * @param handler the handler to pass new channels to
     * @throws RemoteException if an error occurs accepting channels
     * @throws ConnectionClosedException if the connection is closed
     * @throws IllegalArgumentException if any argument is null
     */
    public synchronized void accept(String name, ChannelHandler handler) 
        throws RemoteException {
        ensureOpen();
        _acceptor.accept(name, handler);
    }

    /**
     * Register a handler to accept a single channel connection for the 
     * specified name, and bind the listener to it <br/>
     * This method returns immediately.
     *
     * @param name the channel name to listen on
     * @param listener the listener to handle requests on the channel
     * @throws RemoteException if an error occurs accepting channels
     * @throws ConnectionClosedException if the connection is closed
     * @throws IllegalArgumentException if any argument is null
     */
    public synchronized void accept(String name, ChannelListener listener) 
        throws RemoteException {
        ensureOpen();
        _acceptor.accept(name, listener);
    }

    /**
     * Stop accepting requests on the specified channel
     *
     * @param name the channel name
     * @throws IllegalArgumentException if <code>name</code> is null
     */
    public synchronized void close(String name) {
        if (name == null) {
            throw new IllegalArgumentException("Argument 'name' is null");
        }
        if (_acceptor != null) {
            _acceptor.close(name);
        }
    }

    /**
     * Add a listener for connection events
     *
     * @param listener the listener to add
     */
    public synchronized void addConnectionEventListener(
        ConnectionEventListener listener) {
        if (_listeners != null) {
            _listeners.addConnectionEventListener(listener);
        }
    }

    /**
     * Remove a listener of connection events
     *
     * @param listener the listener to remove
     */
    public synchronized void removeConnectionEventListener(
        ConnectionEventListener listener) {
        if (_listeners != null) {
            _listeners.removeConnectionEventListener(listener);
        }
    }

    /**
     * Close the connection.
     * <p>
     * This closes all channels open on the connection, and then closes the 
     * underlying physical connection, prior to notifying
     * registered {@link ConnectionEventListener} instances
     *
     * @throws RemoteException if an error occurs closing the connection
     */
    public void close() throws RemoteException {
        ConnectionEventNotifier listeners;
        synchronized (this) {
            listeners = _listeners;
        }
        if (listeners != null) {
            _category.debug("Closing connection, id=" + _connectionId);
            try {
                doClose();
            } finally {
                // notify the listeners
                listeners.closed(this);
            }
        }
    }

    /**
     * Performs the actual connection closure, without notifying registered
     * listeners
     *
     * @throws RemoteException if an error occurs closing the connection
     */
    protected synchronized void doClose() throws RemoteException {
        if (!_closed) {
            _closed = true; 
            if (_systemChannel != null) {
                _systemChannel.close();
                _systemChannel = null;
            }
            if (_acceptorChannel != null) {
                _acceptorChannel.close();
                _acceptorChannel = null;
            }
            _queue = null;
            _pool = null;
            _threads = null;
            _listeners = null;
        }
    }

    /**
     * Returns the channel for the give channel identifier
     *
     * @param channelId the channel identifier
     * @return the channel corresponding to channelId, or null, if no channel
     * exists
     */
    protected synchronized Channel getChannel(int channelId) {
        Channel result = null;
        if (_channels != null) {
            result = (Channel) _channels.get(new Integer(channelId));
        }
        return result;
    } 

    /**
     * Returns the queue of packets to be streamed down the physical connection
     */
    protected synchronized PacketQueue getOutputQueue() {
        return _queue;
    }

    /**
     * Returns the pool of free packets
     */
    protected synchronized PacketPool getPool() {
        return _pool;
    }

    /**
     * Returns the thread pool
     */
    protected synchronized ThreadPool getThreadPool() {
        return _threads;
    }

    protected synchronized PacketChannel createChannel(
        int channelId, int destinationId) throws RemoteException {
        PacketChannel channel;
        ensureOpen();
        PacketQueue input = new PacketQueue();
        channel = new PacketChannel(this, channelId, destinationId, input);
        _channels.put(new Integer(channelId), channel);
        return channel;
    }

    protected synchronized void closed(PacketChannel channel) 
        throws RemoteException {
        if (!_closed) {
            Integer id = new Integer(channel.getChannelId());
            if (_channels.remove(id) != null && 
                channel.getChannelId() <= PacketChannel.LAST_SYSTEM_ID) {
                // system channels don't need to send a message notifying
                // of closure
                _systemChannel.close(channel);
            }
        }
    }
    
    protected synchronized int getNextChannelId() {
        return ++_channelIdSeed;
    }

    protected synchronized void ensureOpen() throws RemoteException {
        if (_closed) {
            throw new RemoteException("Connection is closed");
        }
    }

    private SystemPacketChannel createSystemChannel() {
        PacketQueue input = new PacketQueue();
        SystemPacketChannel result = new SystemPacketChannel(
            this, PacketChannel.SYSTEM_CHANNEL_ID, 
            PacketChannel.ACCEPTOR_CHANNEL_ID, input);
        _channels.put(new Integer(PacketChannel.SYSTEM_CHANNEL_ID), result);
        return result;
    }

} //-- PacketConnection
