/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2002 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: PacketInputStream.java,v 1.3 2002/02/26 22:16:31 tima Exp $
 */
package org.exolab.core.messenger;

// java
import java.io.IOException;
import java.io.InputStream;


/**
 * A stream that deserializes data from a {@link PacketQueue}.
 * <p>
 * NOTE: this class is not thread-safe
 *
 * @version     $Revision: 1.3 $
 * @author      <a href="mailto:tima@intalio.com">Tim Anderson</a>
 * @see         PacketOutputStream
 * @see         Packet
 * @see         PacketPool
 * @see         PacketQueue
 */
class PacketInputStream extends InputStream {

    /**
     * The queue to add packets to
     */
    private PacketQueue _queue;

    /**
     * The current packet
     */
    private Packet _packet = null;

    /**
     * The type of the current packet
     */
    private byte _type = Packet.TC_DATA;

    /**
     * The serialized packet data
     */
    private byte[] _data = null;

    /**
     * The byte offset into the current packet
     */
    private int _position = 0;

    /**
     * The size of the current packet. This may be less than the data length
     */
    private int _size;


    /**
     * Create a new input stream, reading all packets from the specified queue
     * 
     * @param queue the queue to read packets from
     * @throws IllegalArgumentException if <code>queue</code> is null
     */
    public PacketInputStream(PacketQueue queue) {
        if (queue == null) {
            throw new IllegalArgumentException("Argument queue is null");
        }
        _queue = queue;        
    }

    /**
     * Closes this output stream and releases any system resources associated 
     * with this stream.
     */ 
    public void close() {
        if (_packet != null) {
            _packet.release();
            _packet = null;
        }
    }

    /**
     * Reads the next byte of data from the input stream. The value byte is
     * returned as an <code>int</code> in the range <code>0</code> to
     * <code>255</code>. If no byte is available because the end of the stream
     * has been reached, the value <code>-1</code> is returned. This method
     * blocks until input data is available, the end of the stream is detected,
     * or an exception is thrown.
     *
     * @return the next byte of data, or <code>-1</code> if the end of the
     * stream is reached.
     * @throws IOException if an I/O error occurs.
     */
    public int read() throws IOException {
        if (_queue == null) {
            throw new IOException("PacketInputStream has been closed");
        }
        if (_packet == null || _position >= _size) {
            load();
        }
        return (_position < _size) ? (_data[_position++] & 0xff) : -1;
    }

    /**
     * Reads up to <code>length</code> bytes of data from the input stream into
     * an array of bytes.  An attempt is made to read as many as
     * <code>length</code> bytes, but a smaller number may be read, possibly
     * zero. The number of bytes actually read is returned as an integer.
     *
     * <p> If the first byte cannot be read for any reason other than end of
     * file, then an <code>IOException</code> is thrown. In particular, an
     * <code>IOException</code> is thrown if the input stream has been closed.
     *
     * @param buffer the buffer into which the data is read
     * @param offset the start offset in array <code>buffer</code> at which
     * the data is written
     * @param length the maximum number of bytes to read
     * @return the total number of bytes read into the buffer, or
     *             <code>-1</code> if there is no more data because the end of
     *             the stream has been reached.
     * @throws IOException if an I/O error occurs.
     * @throws IndexOutOfBoundsException if <code>offset</code> is negative, 
     * or <code>length</code> is negative, or <code>offset+length</code> is 
     * greater than the length of the array
     * @throws NullPointerException if <code>buffer</code> is null
     */
    public int read(byte buffer[], int offset, int length) throws IOException {
        int result = -1;

        if (buffer == null) {
	    throw new NullPointerException("Argument 'buffer' is null");
	} else if ((offset < 0) || (offset > buffer.length) || (length < 0) ||
		   ((offset + length) > buffer.length) || 
                   ((offset + length) < 0)) {
	    throw new IndexOutOfBoundsException();
	} else if (length == 0) {
            result = 0;
	} else if (_queue == null) {
            throw new IOException("PacketInputStream has been closed");
        } else {
            int remaining = length;
            boolean done = false;
            while (remaining > 0 && !done) {
                if (_packet == null || _position == _size) {
                    // load the next packet. This may block.
                    load();
                } 
                if (_packet == null) {
                    // end-of-stream
                    done = true;
                } else {
                    int count = _size - _position;  // bytes left in packet
                    int toCopy = (remaining <= count) ? remaining : count;

                    System.arraycopy(_data, _position, buffer, offset, toCopy);
                    _position += toCopy;
                    offset += toCopy;
                    remaining -= toCopy;
                }
            }
            result = length - remaining;
        }
        return result;
    }

    /**
     * Returns the number of bytes that can be read (or skipped over) from
     * this input stream without blocking by the next caller of a method for
     * this input stream.
     *
     * @return the number of bytes that can be read from this input stream
     * without blocking.
     */
    public int available() throws IOException {
        if (_queue == null) {
            throw new IOException("PacketInputStream has been closed");
        }
        return (_packet != null) ? _data.length - _position : 0;
    }

    /**
     * Returns the type of the current packet
     *
     * @return the packet type
     */
    public byte getType() {
        return _type;
    }

    /**
     * Load this with a new packet from the input queue. This call will block 
     * until a packet is available, or the queue is closed. If the queue
     * is closed, {@link #_packet} will be null.
     */
    private void load() {
        if (_packet != null) {
            _packet.release();
        }
        _packet = _queue.remove();
        if (_packet != null) {
            _data = _packet.getData();
            _position = _packet.getHeaderSize();
            _size = _packet.getSize();
            _type = _packet.getType();
        } else {
            // queue has been closed
        }
    }

} //-- PacketInputStream
