/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2002 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: PacketOutputStream.java,v 1.2 2002/02/20 10:23:48 tima Exp $
 */
package org.exolab.core.messenger;

// java io
import java.io.IOException;
import java.io.OutputStream;


/**
 * A stream that serializes data into a {@link PacketQueue}
 * <p>
 * NOTE: this class is not thread-safe
 *
 * @version     $Revision: 1.2 $
 * @author      <a href="mailto:tima@intalio.com">Tim Anderson</a>
 * @see         PacketInputStream
 * @see         Packet
 * @see         PacketPool
 * @see         PacketQueue
 */
class PacketOutputStream extends OutputStream {

    /**
     * The packet type
     */
    private byte _type = Packet.TC_DATA;

    /**
     * The connection identifier
     */
    private int _connectionId;

    /**
     * The channel identifier of the sender
     */
    private int _channelId;

    /**
     * The channel identifier of the receiver
     */
    private int _destinationId;

    /**
     * The queue to add packets to
     */
    private PacketQueue _queue;

    /**
     * The pool to allocate packets from
     */
    private PacketPool _pool;

    /**
     * The current packet
     */
    private Packet _packet = null;

    /**
     * The serialized packet data
     */
    private byte[] _data = null;
       
    /**
     * The byte offset into the current packet
     */
    private int _position = 0;


    /**
     * Create a new output stream, adding all packets to the specified queue
     * 
     * @param connectionId the connection identifier
     * @param channelId the channel identifier of the sender
     * @param destinationId the channel identifier of the receiver
     * @param queue the queue to add packets to
     * @param pool the pool to allocate packets from
     * @throws IllegalArgumentException if any argument is null
     */
    public PacketOutputStream(int connectionId, int channelId, 
                              int destinationId, PacketQueue queue, 
                              PacketPool pool) {
        if (queue == null) {
            throw new IllegalArgumentException("Argument 'queue' is null");
        }
        if (pool == null) {
            throw new IllegalArgumentException("Argument 'pool' is null");
        }            
        _connectionId = connectionId;
        _channelId = channelId;
        _destinationId = destinationId;
        _queue = queue;
        _pool = pool;
    }

    /**
     * This implementation flushes the stream, rather than closing it, 
     * as the stream is re-used.
     *
     * @throws IOException if an I/O error occurs
     */ 
    public void close() throws IOException {
        flush();
    }

    /**
     * Flushes this output stream and forces any buffered output bytes to be 
     * written out.
     *
     * @throws IOException if an I/O error occurs
     */ 
    public void flush() throws IOException {
        if (_queue == null) {
            throw new IOException("PacketOutputStream has been closed");
        }
        if (_packet != null) {
            queue();
        }
    }

    /**
     * Writes length bytes from the specified byte array starting at offset to
     * this output stream
     *
     * @param buffer the data to write
     * @param offset the start offset in the data
     * @param length the number of bytes to write
     * @throws IOException if an I/O error occurs
     */
    public void write(byte buffer[], int offset, int length) 
        throws IOException {
        if (_queue == null) {
            throw new IOException("PacketOutputStream has been closed");
        }
        int remaining = length;
        while (remaining > 0) {
            // get the packet to populate
            if (_packet == null || _position == _data.length) {
                load();
            } 
            int space = _data.length - _position; // space left in packet
            int toCopy = (remaining <= space) ? remaining : space;

            System.arraycopy(buffer, offset, _data, _position, toCopy);
            _position += toCopy;
            offset += toCopy;
            remaining -= toCopy;

            if (_position == _data.length) {
                // queue the packet
                queue();
            }
        }
    }

    /**
     * Writes the specified byte to this output stream
     *
     * @param value the byte value
     * @throws IOException if an I/O error occurs
     */
    public void write(int value) throws IOException {
        if (_queue == null) {
            throw new IOException("PacketOutputStream has been closed");
        }
        // get the packet to populate
        if (_packet == null || _position == _data.length) {
            load();
        } 
        _data[_position++] = (byte) value;

        if (_position == _data.length) {
            // queue the packet
            queue();
        }
    }

    /**
     * Set the type to assign to packets
     *
     * @param type the packet type
     */
    public void setType(byte type) {
        _type = type;
    }

    /**
     * Set the destination channel identifier
     * 
     * @param destinationId the destination channel identifier
     */
    public void setDestinationId(int destinationId) {
        _destinationId = destinationId;
    }

    /**
     * Allocate a new packet to populate, from the pool
     *
     * @throws IOException if the pool has been closed
     */
    private void load() throws IOException {
        _packet = _pool.allocate();
        if (_packet == null) {
            throw new IOException("PacketPool has been closed");
        }
        _data = _packet.getData();
        _position = _packet.getHeaderSize();
        _packet.setType(_type);
        _packet.setConnectionId(_connectionId);
        _packet.setChannelId(_channelId);
        _packet.setDestinationId(_destinationId);
    }

    /**
     * Queue the current packet
     */
    private void queue() {
        _packet.setSize(_position);
        _queue.add(_packet);
        _packet = null;
        _data = null;
    }

} //-- PacketOutputStream
