/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2002 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: PacketPool.java,v 1.4 2002/02/26 22:17:37 tima Exp $
 */
package org.exolab.core.messenger;

// java util
import java.util.LinkedList;


/**
 * This class implements a pool of free packets, designed to minimise
 * memory use by imposing a limit on the number of packets used by 
 * {@link PacketConnection} instances.<p>
 * When a request to allocate a new packet is made, one will be returned from
 * the list of free packets. If none is available, and the size limit has
 * not been reached, a new one will be allocated. If the limit has been
 * reached, the request will block until a packet is released.
 *
 * @version     $Revision: 1.4 $
 * @author      <a href="mailto:tima@intalio.com">Tim Anderson</a>
 * @see         Packet
 * @see         PacketConnection
 * @see         PacketOutputStream
 * @see         PacketInputputStream
 */
class PacketPool {

    /**
     * The size, in bytes, to allocate to packets
     */
    private final int _packetSize;

    /**
     * The number of allocated packets
     */
    private int _allocated = 0;

    /**
     * The maximum number of elements that the pool can contain, before
     * calls to {@link #get} block. A value <= 0 indicates that the
     * pool can contain an unlimited no. of packets
     */
    private final int _limit;

    /**
     * The list of free packets
     */
    private LinkedList _packets;

    /**
     * Flag to indicate that the pool has been closed
     */
    private boolean _closed = false;

    /**
     * Creates a pool, of unlimited size
     *
     * @param packetSize the size in bytes, of packets
     */
    public PacketPool(int packetSize) {
        this(packetSize, 0);
    }

    /**
     * Creates a new pool, of limited size. Once the no. of packets
     * in the pool reaches the limit, calls to get new packets block until
     * a packet is released.
     *
     * @param limit the maximum number of packets that the pool can contain
     * at any one time.
     */
    public PacketPool(int packetSize, int limit) {
        _packetSize = packetSize;
        _limit = limit;        
        _packets = new LinkedList();
    }

    /**
     * Get a packet. If the pool has packets, then one will be returned
     * from the pool. If not, and there is room left in the pool, a new
     * one will be created and returned, otherwise the call will block
     * until a packet is available.
     *
     * @return the next available packet, or null, if this pool has been closed
     */
    public Packet allocate() {
        Packet packet = null;
        synchronized (_packets) {
            if (_packets.isEmpty()) {
                if (_limit <= 0 || _allocated < _limit) {
                    ++_allocated;
                    packet = new Packet(_packetSize, this);
                }
            }

            if (packet == null) {
                // return the first available packet
                while (!_closed && _packets.isEmpty()) {
                    try {
                        _packets.wait();
                    } catch (InterruptedException ignore) {
                    }
                }
                        
                if (!_closed) {
                    // if the pool hasn't been closed while waiting, return
                    // the first free packet
                   packet = (Packet) _packets.removeFirst();
                }
            }
        }

        return packet;
    }

    /**
     * Return a packet to the pool
     *
     * @param packet the packet to add
     */
    public void release(Packet packet) {
        synchronized(_packets) {
            if (!_closed) {
//                  NOTE: if you have inexplicable serialization problems
//                        start debugging by uncommenting the following
//                  if (_packets.contains(packet)) {
//                      throw new RuntimeException("Attempt to re-add packet");
//                  }
                _packets.add(packet);
                _packets.notify();
            }
        }
    }

    /**
     * Returns <code>true</code> if this queue is empty
     *
     * @return <code>true</code> if this queue is empty
     */
    public boolean isEmpty() {
        boolean empty;
        synchronized (_packets) {
            empty = _packets.isEmpty();
        }
        return empty;
    }

    /**
     * Closes the packet pool. If it is blocked then unblock it and release
     * it, so it can close correctly
     */
    public void close() {
        synchronized (_packets) {
            _closed = true;
            _packets.clear();
            _packets.notifyAll();
        }
    }

    /**
     * Returns <code>true</code> if the pool is closed
     */
    public boolean isClosed() {
        boolean closed;
        synchronized (_packets) {
            closed = _closed;
        }
        return closed;
    }
    
} //-- PacketPool
