/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2002 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: PacketQueue.java,v 1.4 2003/06/09 06:28:40 tanderson Exp $
 */
package org.exolab.core.messenger;

// java
import java.util.LinkedList;

// exolab
import org.apache.commons.logging.LogFactory;
import org.apache.commons.logging.Log;


/**
 * This class implements a queue of packets.<p>
 * Packets are added at the end of the queue, and removed from the front.
 * The queue can have a size limit set - when reached, all calls to add
 * a new packet will block.
 *
 * @version     $Revision: 1.4 $
 * @author      <a href="mailto:tima@intalio.com">Tim Anderson</a>
 * @see         Packet
 */
public class PacketQueue {

    /**
     * The queue of {@link Packet} instances
     */
    private LinkedList _queue;

    /**
     * This flag is used to mark the connection as closed.
     */
    private boolean _closed = false;

    /**
     * The maximum number of elements that the queue can contain, before
     * calls to {@link #add} block. A value <= 0 indicates that the
     * queue can contain an unlimited no. of packets
     */
    private int _limit = 0;


    /**
     * Creates a new empty queue, of unlimited size
     */
    public PacketQueue() {
        _queue = new LinkedList();
    }

    /**
     * Creates a new empty queue, of limited size. Once the no. of packets
     * in the queue reaches the limit, calls to add new packets block until
     * space is available.
     *
     * @param limit the maximum number of packets that the queue can contain
     * at any one time.
     */
    public PacketQueue(int limit) {
        _queue = new LinkedList();
        _limit = limit;
    }

    /**
     * Add a new packet to the end of the queue.
     * If the queue has a size limit in place, and the queue has reached
     * that limit, this call will block, until space is available.
     *
     * @param packet the packet to add
     * @throws IllegalArgumentException if packet is null
     */
    public void add(Packet packet) {
        if (packet == null) {
            throw new IllegalArgumentException("Argument 'packet' is null");
        }

        synchronized (_queue) {
            if (_limit != 0) {
                // if the queue is limited in size, only add the
                // packet when there is space available
                while (!_closed && _queue.size() >= _limit) {
                    try {
                        _queue.wait();
                    } catch (InterruptedException ignore) {
                    }
                }
            }
            if (!_closed) {
                _queue.add(packet);
                _queue.notify();
            }
        }
    }

    /**
     * Remove a packet from the top of the queue. This call blocks
     * until a packet is available.
     *
     * @return the next available packet, or <code>null</code>, if the queue 
     * has been closed
     */
    public Packet remove() {
        Packet packet = null;

        synchronized (_queue) {
            while (!_closed && _queue.isEmpty()) {
                try {
                    _queue.wait();
                } catch (InterruptedException ignore) {
                }
            }
            
            // check to see that we have not been forced to exit
            if (!_closed) {
                packet = (Packet) _queue.removeFirst();
                _queue.notify();
            }
        }
        return packet;
    }

    /**
     * Return the packet at the front of the queue, without removing it.
     * If no packet is available, this returns null.
     *
     * @return the packet at the front of the queue, or <code>null</code>
     * if no packet is currently queued
     */
    public Packet peek() {
        Packet packet = null;
        synchronized (_queue) {
            if (!_closed && !_queue.isEmpty()) {
                packet = (Packet) _queue.getFirst();
            }
        }
        return packet;
    }

    /**
     * Wait at most <code>timeout</code> milliseconds until a packet 
     * is available
     *
     * @param timeout the maximum time to wait (in milliseconds) for
     * a packet to become available. A timeout of zero never expires and the
     * call blocks indefinitely, or until the queue is closed
     * @return <code>true</code> if a packet becomes available in the 
     * specified time
     */
    public boolean waitForPacket(long timeout) {
        boolean available = false;
        synchronized (_queue) {
            if (!_closed && _queue.isEmpty()) {
                try {
                    if (timeout == 0) {
                        _queue.wait();
                    } else {
                        _queue.wait(timeout);
                    }
                } catch (InterruptedException ignore) {
                }
            }
            if (!_closed && !_queue.isEmpty()) {
                available = true;
            }
        }
        return available;
    }

    /**
     * Returns <code>true</code> if this queue is empty
     *
     * @return <code>true</code> if this queue is empty
     */
    public boolean isEmpty() {
        boolean empty = false;
        synchronized (_queue) {
            empty = _queue.isEmpty();
        }
        return empty;
    }

    /**
     * Close the queue. If it is blocked then unblock it and release
     * it, so it can close correctly
     */
    public void close() {
        synchronized (_queue) {
            _closed = true;
            _queue.notifyAll();
        }
    }

    /**
     * Returns <code>true</code> if the queue is closed
     */
    public boolean isClosed() {
        boolean closed = false;
        synchronized (_queue) {
            closed = _closed;
        }
        return closed;
    }

} //-- PacketQueue
