/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2002 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: StreamDemultiplexer.java,v 1.4 2003/06/09 06:28:41 tanderson Exp $
 */
package org.exolab.core.messenger;

// java
import java.io.IOException;
import java.io.InputStream;
import java.rmi.UnmarshalException;

// exolab
import org.apache.commons.logging.LogFactory;
import org.apache.commons.logging.Log;


/**
 * This class implements a demultiplexer that reads packets from an 
 * {@link InputStream} streams and sends them to their associated 
 * {@link Channel} instance.<p>
 * It is intended to run in a separate thread, terminating if one of the
 * following conditions occur:
 * <ul>
 *   <li>{@link #close} is invoked</li>
 *   <li>a read from the input stream fails</li>
 *   <li>the source {@link PacketQueue} closes</li>
 * </ul>
 *
 * @version     $Revision: 1.4 $ $Date: 2003/06/09 06:28:41 $
 * @author      <a href="mailto:tima@intalio.com">Tim Anderson</a>
 * @see         PacketConnection
 * @see         Demultiplexer
 */
public class StreamDemultiplexer implements Demultiplexer {

    /**
     * The connection to destream packets for
     */
    private ManagedPacketConnection _connection;

    /**
     * The stream to read packets from
     */
    private InputStream _stream;

    /**
     * The pool to allocate packets from
     */
    private PacketPool _pool;
    
    /**
     * Flag used to terminate the demultiplexer
     */
    private volatile boolean _closed = false;

    /**
     * The multiplexer thread
     */
    private Thread _thread = null;

    /**
     * The thread identifier seed value
     */
    private static int _threadIdSeed = 0;

    /**
     * The logging category
     */
    private static final Log _category =
        LogFactory.getLog("org.exolab.core.messenger");


    /**
     * Construct a new multiplexer
     *
     * @param connection the connection to read packets from
     * @param stream the stream to read packets from
     * @throws IllegalArgumentException if any argument is null
     */
    public StreamDemultiplexer(ManagedPacketConnection connection, 
                               InputStream stream) {
        if (connection == null) {
            throw new IllegalArgumentException(
                "Argument 'connection' is null");
        }
        if (stream == null) {
            throw new IllegalArgumentException("Argument 'stream' is null");
        }
        _connection = connection;
        _stream = stream;
        _pool = connection.getPool();
    }

    /**
     * Start demultiplexing. This reads packets from an {@link InputStream} 
     * instance, and dispatches them to the appropriate channel.
     *
     * @throws IllegalStateException if the demultiplexer is running, or
     * has been closed
     */
    public synchronized void start() {
        if (_closed) {
            throw new IllegalStateException(
                "StreamDemultiplexer has been closed");
        }
        if (_thread != null) {
            throw new IllegalStateException(
                "StreamDemultiplexer is already running");
        }
        
        Runnable runnable = new Runnable() {
            public void run() {
                demultiplex();
                _category.debug("StreamDemultiplexer terminating");
            }
        };

        String name = "StreamDemultiplexer-" + getNextId();
        _thread = new Thread(runnable, name);
        _thread.start();
    }

    /**
     * Close the multiplexer, freeing any allocated resources. Once a 
     * multiplexer is closed, it may not be restarted.
     */
    public synchronized void close() {
        _closed = true;
        if (_thread != null) {
            _thread.interrupt();
        }
    }

    /**
     * Returns <code>true</code> if the demultiplexer is closed
     */
    public synchronized boolean closed() {
        return _closed;
    }

    /**
     * Stream packets from the connection's input stream, and dispatch them
     * to the appropriate channel
     */
    protected void demultiplex() {
        while (!closed()) {
            Packet packet = _pool.allocate();
            if (packet == null) {
                // the pool has closed. Only notify the connection if close()
                // wasn't invoked
                if (!_closed) {
                    _closed = true;
                    _connection.streamClosed();
                }
            } else {
                try {
                    if (packet.read(_stream) != -1) {
                        dispatch(packet);
                    } else {
                        // the underlying stream has closed. Only notify
                        // the listeners if close() wasn't invoked
                        if (!_closed) {
                            _closed = true;
                            _connection.streamClosed();
                        }
                    }
                } catch (IOException exception) {
                    // Only notify the listeners if close() wasn't invoked
                    if (!_closed) {
                        _closed = true;
                        UnmarshalException wrapper = 
                            new UnmarshalException(exception.getMessage(),
                                                    exception);
                        _connection.errorOnStream(wrapper);
                    }
                } 
            }
        }

        // clean up
        synchronized (this) {
            _thread = null;
            _connection = null;
            _stream = null;
            _pool = null;
        }
    }

    /**
     * Dispatches a packet to its associated channel
     * 
     * @param packet the packet to dispatch
     */
    protected void dispatch(Packet packet) {
        int connectionId = packet.getConnectionId();
        int channelId = packet.getDestinationId();
        PacketChannel channel = 
            (PacketChannel) _connection.getChannel(connectionId, channelId);
        if (channel != null) {
            channel.enqueue(packet);
        } else {
            _category.debug("StreamDemultiplexer tossed a packet for " +
                               "non existent channel, connection id=" + 
                               connectionId + ", channel id=" + channelId);
            // TODO need to notify someone that a packet is being tossed
            // Introduce the notion of a ProtocolHandler? - this would
            // send a packet back to the source notifying it that the
            // channel is no longer valid - this would then result in
            // an IOException on the other side.
        }
    }

    private synchronized int getNextId() {
        return ++_threadIdSeed;
    }
    
} //-- StreamDemultiplexer
