/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2002 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: StreamMultiplexer.java,v 1.4 2003/06/09 06:28:41 tanderson Exp $
 */
package org.exolab.core.messenger;

// java
import java.io.IOException;
import java.io.OutputStream;
import java.rmi.MarshalException;

// exolab
import org.apache.commons.logging.LogFactory;
import org.apache.commons.logging.Log;


/**
 * This class implements a multiplexer that streams packets from a 
 * PacketConnection's outbound queue to an {@link OutputStream}.<p>
 * It is intended to run in a separate thread, terminating if one of the
 * following conditions occur:
 * <ul>
 *   <li>{@link #close} is invoked</li>
 *   <li>a write to the outbound stream fails</li>
 *   <li>the source {@link PacketQueue} closes</li>
 * </ul>
 *
 * @version     $Revision: 1.4 $ $Date: 2003/06/09 06:28:41 $
 * @author      <a href="mailto:tima@intalio.com">Tim Anderson</a>
 * @see         PacketConnection
 * @see         Multiplexer
 */
public class StreamMultiplexer implements Multiplexer {

    /**
     * The connection to stream packets for
     */
    private ManagedPacketConnection _connection;

    /**
     * The queue of packets owned by the connection
     */
    private PacketQueue _queue;

    /**
     * The stream to write packets to
     */
    private OutputStream _stream;

    /**
     * Flag used to terminate the multiplexer
     */
    private volatile boolean _closed = false;

    /**
     * The multiplexer thread
     */
    private Thread _thread = null;

    /**
     * The thread identifier seed value
     */
    private static int _threadIdSeed = 0;

    /**
     * The logging category
     */
    private static final Log _category =
        LogFactory.getLog("org.exolab.core.messenger");


    /**
     * Construct a new multiplexer
     *
     * @param connection the connection to read packets from
     * @param stream the stream to write packets to
     * @throws IllegalArgumentException if connection or stream is null
     */
    public StreamMultiplexer(ManagedPacketConnection connection, 
                             OutputStream stream) {
        if (connection == null) {
            throw new IllegalArgumentException(
                "Argument 'connection' is null");
        }
        if (stream == null) {
            throw new IllegalArgumentException("Argument 'stream' is null");
        }
        _connection = connection;
        _stream = stream;
        _queue = connection.getOutputQueue();
    }

    /**
     * Start multiplexing. This reads packets from an {@link PacketConnection} 
     * instance, and sends them down the connection's physical connection 
     * implementation.
     *
     * @throws IllegalStateException if the multiplexer is running, or has 
     * been closed
     */
    public synchronized void start() {
        if (_closed) {
            throw new IllegalStateException(
                "StreamMultiplexer has been closed");
        }

        if (_thread != null) {
            throw new IllegalStateException(
                "StreamMultiplexer is already running");
        }
        
        Runnable runnable = new Runnable() {
            public void run() {
                multiplex();
                _category.debug("StreamMultiplexer terminating");
            }
        };

        String name = "StreamMultiplexer-" + getNextId();
        _thread = new Thread(runnable, name);
        _thread.start();
    }

    /**
     * Close the multiplexer, freeing any allocated resources. Once a 
     * multiplexer is closed, it may not be restarted.
     */
    public void close() {
        _closed = true;
        if (_thread != null) {
            _thread.interrupt();
        }
    }

    /**
     * Stream packets from the connection's output queue to its output
     * stream.
     */
    protected void multiplex() {
        while (!_closed) {
            Packet packet = _queue.remove();
            if (packet == null) {
                // queue has been closed
                if (!_closed) {
                    _closed = true;
                    _connection.streamClosed();
                }
            } else {
                try {
                    packet.write(_stream);
                } catch (IOException exception) {
                    if (!_closed) {
                        _closed = true;
                        MarshalException wrapper = new MarshalException(
                            exception.getMessage(), exception);
                        _connection.errorOnStream(wrapper);
                    }
                } finally {     
                    // put the packet back in the pool
                    packet.release();
                }
            }
        }

        // clean up
        synchronized (this) {
            _thread = null;
            _connection = null;
            _stream = null;
            _queue = null;
        }
    }

    private synchronized int getNextId() {
        return ++_threadIdSeed;
    }
    
} //-- StreamMultiplexer
