/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2002 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: SystemPacketChannel.java,v 1.3 2002/02/20 10:23:48 tima Exp $
 */
package org.exolab.core.messenger;

// java
import java.io.Serializable;
import java.rmi.RemoteException;
import java.rmi.ServerException;


/**
 * SystemPacketChannel is reponsible for managing channel open and close
 * requests on a {@link PacketConnection}.
 *
 * @version     $Revision: 1.3 $ $Date: 2002/02/20 10:23:48 $
 * @author      <a href="mailto:tima@intalio.com">Tim Anderson</a>
 * @see         PacketConnection
 * @see         PacketPool
 * @see         PacketInputStream
 * @see         PacketOutputStream
 */
class SystemPacketChannel extends PacketChannel implements SystemChannel {

    /**
     * Construct a new <code>SystemPacketChannel</code>
     *
     * @param connection the owning connection
     * @param channelId the channel identifier
     * @param destinationId the destination channel identifier
     * @param inputQueue the inbound packet queue, to deserialize messages from
     * @param outputQueue the outbound packet queue, to serialize messages to
     * @param pool the pool to allocate packets from
     * @throws IllegalArgumentException if any argument is null
     */
    public SystemPacketChannel(PacketConnection connection, int channelId,
                               int destinationId, PacketQueue inputQueue) {
        super(connection, channelId, destinationId, inputQueue);
    }

    /**
     * Open a channel for the specified name. A {@link ChannelAcceptor} must
     * be available on the other end of the connection, accepting requests
     * for the named channel.
     *
     * @param name the channel name
     * @return an open channel
     * @throws RemoteException if an error occurs opening the channel
     * @throws InvalidChannelException there is no ChannelAcceptor listening 
     * for requests for the specified name
     * @throws IllegalArgumentException if <code>name</code> is null
     */
    public Channel open(String name) throws RemoteException {
        if (name == null) {
            throw new IllegalArgumentException("Argument 'name' is null");
        }
        PacketConnection owner = getConnection();
        int channelId = owner.getNextChannelId();
        OpenRequest request = new OpenRequest(name, channelId);
        OpenResponse response;
        try {
            response = (OpenResponse) invoke(request);
        } catch (RemoteException exception) {
            throw exception;
        } catch (Exception exception) {
            throw new ServerException(exception.getMessage(), exception);
        }
        int destinationId = response.getChannelId();
        return owner.createChannel(channelId, destinationId);
    }

    /**
     * Close a channel
     *
     * @param channel the channel to close
     * @throws RemoteException if an error occurs closing the channel
     * @throws IllegalArgumentException if <code>channel</code> is null
     */
    public void close(Channel channel) throws RemoteException {
        if (channel == null) {
            throw new IllegalArgumentException("Argument 'channel' is null");
        }
        int channelId = ((PacketChannel) channel).getChannelId();
        Closed request = new Closed(channelId);
        send(request);
    }

    public static final class OpenRequest implements Serializable {
        static final long serialVersionUID = 1;

        private String _name = null;
        private int _channelId = -1;

        public OpenRequest() {
        }

        public OpenRequest(String name, int channelId) {
            _name = name;
            _channelId = channelId;
        }

        public String getName() {
            return _name;
        }

        public int getChannelId() {
            return _channelId;
        }

    } //-- OpenRequest

    public static final class OpenResponse implements Serializable {
        static final long serialVersionUID = 1;

        private int _channelId = -1;
        private int _destinationId = -1;

        public OpenResponse() {
        }

        public OpenResponse(int channelId, int destinationId) {
            _channelId = channelId;
            _destinationId = destinationId;
        }

        public int getChannelId() {
            return _channelId;
        }

        public int getDestinationId() {
            return _destinationId;
        }

    } //-- OpenResponse

    public static final class Closed implements Serializable {
        static final long serialVersionUID = 1;

        private int _channelId = -1;

        public Closed() {
        }
        
        public Closed(int channelId) {
            _channelId = channelId;
        }

        public int getChannelId() {
            return _channelId;
        }

    } // -- Closed

} //-- SystemPacketChannel
