/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2002 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: TcpConnector.java,v 1.6 2003/06/09 06:28:41 tanderson Exp $
 */
package org.exolab.core.messenger.tcp;

// java
import java.io.IOException;
import java.net.InetAddress;
import java.net.MalformedURLException;
import java.net.ServerSocket;
import java.net.Socket;
import java.rmi.ConnectIOException;
import java.rmi.RemoteException;
import java.rmi.UnknownHostException;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

// exolab
import org.apache.commons.logging.LogFactory;
import org.apache.commons.logging.Log;
import org.exolab.core.messenger.AcceptorEventListener;
import org.exolab.core.messenger.Connection;
import org.exolab.core.messenger.ConnectionAcceptor;
import org.exolab.core.messenger.ConnectionProperties;
import org.exolab.core.messenger.Connector;
import org.exolab.core.messenger.Context;
import org.exolab.core.messenger.Limits;
import org.exolab.core.messenger.ManagedConnection;
import org.exolab.core.messenger.ManagedConnectionAcceptor;
import org.exolab.core.messenger.ManagedConnectionEventListener;
import org.exolab.core.messenger.MessengerException;
import org.exolab.core.messenger.ResourceAllocator;
import org.exolab.core.util.URI;


/**
 * A factory for {@link TcpConnection} and {@link TcpManagedConnectionAcceptor}
 * objects
 *
 * @version     $Revision: 1.6 $ $Date: 2003/06/09 06:28:41 $
 * @author      <a href="mailto:tima@intalio.com">Tim Anderson</a>
 * @see         Connector
 * @see         TcpConnection
 * @see         TcpConnectionAcceptor
 */
public class TcpConnector implements Connector, ManagedConnectionEventListener,
                                     AcceptorEventListener {

    /**
     * The resource allocator
     */
    private ResourceAllocator _allocator;

    /**
     * The connection limits
     */
    private Limits _limits = new Limits();

    /**
     * The set of opened TcpManagedConnection instances, keyed on URI
     */
    private HashMap _opened = new HashMap();

    /**
     * The set of all connections, and their corresponding URIs
     */
    private HashMap _connections = new HashMap();

    /**
     * The set of accepted TcpManagedConnection instances
     */
    private List _accepted = new LinkedList();

    /**
     * The connection acceptors, keyed on URI
     */
    private HashMap _acceptors = new HashMap();

    /**
     * The logging category
     */
    private static final Log _category =
        LogFactory.getLog("org.exolab.core.messenger");


    /**
     * Create a new TcpConnector
     */
    public TcpConnector() {
        _allocator = new ResourceAllocator(3, _limits);
    }

    /**
     * Returns a connection for the specified URI and connection properties
     *
     * @param uri the connection address
     * @param properties connection properties (may be null)
     * @return a Connection if the connector finds an acceptable match,
     * otherwise <code>null</code>
     * @throws MalformedURLException if <code>uri</code> is an invalid URI
     * @throws RemoteException if a connection cannot be established
     * @throws IllegalArgumentException if <code>uri</code> is null
     */
    public Connection getConnection(String uri, 
                                    ConnectionProperties properties) 
        throws MalformedURLException, RemoteException {
        ManagedConnection managed = null;
        if (uri == null) {
            throw new IllegalArgumentException("Argument 'uri' is null");
        }
        
        URI parser = URIHelper.getURI(uri);
        String address = URIHelper.getAddress(parser);
        String name = parser.getPath();
        synchronized (this) {
            managed = (ManagedConnection) _opened.get(address);
        }
        if (managed == null) {
            managed = open(address, properties);
        }
        return managed.open(name, properties);
    }

    /**
     * Returns a connection acceptor for the specified URI and connection 
     * properties
     *
     * @param uri the address to accept connections on
     * @param properties connection properties (may be null)
     * @return a connection acceptor, for accepting connection requests on the
     * specified URI
     * @throws MalformedURLException if <code>uri</code> is an invalid URI
     * @throws RemoteException if an I/O error occurs
     * @throws IllegalArgumentException if <code>uri</code> is null
     */
    public synchronized ManagedConnectionAcceptor getManagedConnectionAcceptor(
        String uri, ConnectionProperties properties) 
        throws MalformedURLException, RemoteException {
        ManagedConnectionAcceptor acceptor;
        if (uri == null) {
            throw new IllegalArgumentException("Argument 'uri' is null");
        }

        URI parser = URIHelper.getURI(uri);
        String address = URIHelper.getAddress(parser);
        acceptor = (ManagedConnectionAcceptor) _acceptors.get(address);

//          if (parser.getHost() == null) {
//              throw new MalformedURLException("Host not specified in uri");
//          }

        if (acceptor == null) {
            ServerSocket socket;
            try {
                socket = new ServerSocket(parser.getPort());
            } catch (IOException exception) {
                throw new MessengerException(
                    "Failed to create server socket on port=" + 
                    parser.getPort(),
                    exception);
            }
            acceptor = new TcpManagedConnectionAcceptor(this, address, socket);
            acceptor.addAcceptorEventListener(this);
            _acceptors.put(address, acceptor);
        }
        return acceptor;
    }

    /**
     * Returns the protocol that the connector supports
     *
     * @return the protocol that the connector supports
     */
    public String getScheme() {
        return "tcp";
    }

    /**
     * Invoked when a managed connection is closed
     *
     * @param connection the closed connection
     */
    public synchronized void closed(ManagedConnection connection) {
        String uri = (String) _connections.get(connection);
        if (uri != null) {
            _opened.remove(uri);
        }
    }

    /**
     * Invoked when a fatal connection error occurs, just before an Exception 
     * is thrown to the application
     *
     * @param connection the connection the error occurred on
     * @param exception the error
    */
    public synchronized void error(ManagedConnection connection, 
                                   Exception exception) {
        _connections.remove(connection);
    }

    /**
     * Invoked when a connection acceptor is closed
     *
     * @param acceptor the closed connection acceptor
     */
    public synchronized void closed(ManagedConnectionAcceptor acceptor) {
        _acceptors.remove(acceptor.getURI());
    }

    /**
     * Invoked when a fatal connection acceptor error occurs, just before an 
     * Exception is thrown to the application
     *
     * @param acceptor the connection the error occurred on
     * @param exception the error
    */
    public synchronized void error(ManagedConnectionAcceptor acceptor, 
                                   Exception exception) {
        _acceptors.remove(acceptor.getURI());
    }

    /**
     * Opens a connection for the specified URI and connection properties
     *
     * @param uri the connection address
     * @param properties connection properties (may be null)
     * @return a Connection if the connector finds an acceptable match,
     * otherwise <code>null</code>
     * @throws MalformedURLException if <code>uri</code> is an invalid URI
     * @throws RemoteException if a connection cannot be established
     */
    protected ManagedConnection open(String uri, 
                                     ConnectionProperties properties) 
        throws MalformedURLException, RemoteException {
        URI parser = URIHelper.getURI(uri);
        if (parser.getHost() == null) {
            throw new MalformedURLException("Host not specified in uri");
        }

        Socket socket;
        try {
            InetAddress address = InetAddress.getByName(parser.getHost());
            socket = new Socket(address, parser.getPort());
        } catch (java.net.UnknownHostException exception) {
            throw new UnknownHostException(exception.getMessage(), exception);
        } catch (IOException exception) {
            throw new ConnectIOException(exception.getMessage(), exception);
        }
        Context context = _allocator.allocateContext();

        // Cannot synchronize here to avoid deadlocks if a connection is 
        // being opened to a port whose server socket resides in the same JVM -
        // this will invoke accepted() below
        // TODO - fix race condition establishing 2 connections to same port
        TcpManagedConnection connection = new TcpManagedConnection(
            context, socket);

        synchronized (this) {
            _opened.put(uri, connection);
            _connections.put(connection, uri);
            connection.setManagedConnectionEventListener(this);
            _category.debug("Opened connection, uri=" + uri);
        }
        return connection;
    }

    /**
     * Construct a new connection to process messages on a socket, accepted 
     * by a {@link TcpConnectionAcceptor}
     *
     * @param socket the tcp socket
     * @param handlers the set of connection handlers, keyed on name
     * @return a new connection to process messages on the socket
     * @throws RemoteException for any error
     * @throws IllegalArgumentException if any argument is null
     */
    protected TcpManagedConnection accepted(Socket socket, HashMap handlers)
        throws RemoteException {
        if (socket == null) {
            throw new IllegalArgumentException("Argument 'socket' is null");
        }
        if (handlers == null) {
            throw new IllegalArgumentException("Argument 'handlers' is null");
        }
        Context context = _allocator.allocateContext();
        TcpManagedConnection connection = 
            new TcpManagedConnection(context, socket, handlers);
        synchronized (this) {
            _accepted.add(connection);
        }
        return connection;
    }
    
} //-- TcpConnector
