/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2002 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: TcpManagedConnection.java,v 1.2 2002/02/21 09:49:41 tima Exp $
 */
package org.exolab.core.messenger.tcp;

// java
import java.io.IOException;
import java.net.Socket;
import java.rmi.RemoteException;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;

// exolab
import org.exolab.core.messenger.Context;
import org.exolab.core.messenger.ConnectionHandler;
import org.exolab.core.messenger.ManagedPacketConnection;


/**
 * A <code>ManagedConnection</code> that manages {@link Connection} instances
 * over a tcp socket.<p/>
 *
 * @version     $Revision: 1.2 $ $Date: 2002/02/21 09:49:41 $
 * @author      <a href="mailto:tima@intalio.com">Tim Anderson</a>
 * @see         ManagedPacketConnection
 * @see         TcpConnector
 */
class TcpManagedConnection extends ManagedPacketConnection {

    /**
     * The tcp socket
     */
    private Socket _socket;


    /**
     * Construct a new <code>TcpManagedConnection</code>
     *
     * @param context the connection context
     * @param socket the tcp socket
     * @throws RemoteException if a connection cannot be established
     * @throws IllegalArgumentException if any argument is null
     */
    public TcpManagedConnection(Context context, Socket socket)
        throws RemoteException {
        this (context, socket, null);
    }
    
    /**
     * Construct a new <code>TcpManagedConnection</code> and register
     * a set of {@link ConnectionHandler} instances
     *
     * @param context the connection context
     * @param socket the tcp socket
     * @param handlers the set of connection handlers, keyed on name
     * @throws RemoteException if a connection cannot be established
     * @throws IllegalArgumentException if <code>context</code> or
     * <code>socket</code> is null
     */
    public TcpManagedConnection(Context context, Socket socket,
                                HashMap handlers) throws RemoteException {
        super(context, new TcpMultiplexerFactory(socket));
        _socket = socket;

        if (handlers != null) {
            // register all ConnectionHandler instances, prior to initialising
            // the connection - this will avoid potential 'Connection refused'
            // problems
            Iterator iter = handlers.entrySet().iterator();
            while (iter.hasNext()) {
                Map.Entry entry = (Map.Entry) iter.next();
                String name = (String) entry.getKey();
                ConnectionHandler handler = 
                    (ConnectionHandler) entry.getValue();
                accept(name, handler);
            }
        }

        // initialise the connection
        try {
            initialise();
        } catch (RemoteException exception) {
            try {
                _socket.close();
                _socket = null;
            } catch (IOException ignore) {
            }
            throw exception;
        }
    }

    /**
     * Performs the actual connection closure, without notifying registered
     * listeners
     *
     * @throws RemoteException if an error occurs closing the connection
     */
    protected synchronized void doClose() throws RemoteException {
        super.doClose();
        if (_socket != null) {
            try {
                _socket.close();
            } catch (IOException ignore) {
            } finally {
                _socket = null;
            }
        }
    }
    
} //-- TcpManagedConnection
