/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2002 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: TcpManagedConnectionAcceptor.java,v 1.1 2002/02/21 09:46:42 tima Exp $
 */
package org.exolab.core.messenger.tcp;

// java
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.ServerSocket;
import java.net.Socket;
import java.rmi.RemoteException;
import java.util.HashMap;

// exolab
import org.exolab.core.messenger.AcceptorEventListener;
import org.exolab.core.messenger.AcceptorEventNotifier;
import org.exolab.core.messenger.ConnectionHandler;
import org.exolab.core.messenger.Context;
import org.exolab.core.messenger.ManagedConnection;
import org.exolab.core.messenger.ManagedConnectionAcceptor;
import org.exolab.core.messenger.MessengerException;
import org.exolab.core.messenger.ResourceAllocator;
import org.exolab.core.util.URI;


/**
 * A <code>ManagedConnectionAcceptor</code> that waits for a connection 
 * request over a tcp socket, and constructs a {@link TcpManagedConnection} 
 * object to handle it.
 *
 * @version     $Revision: 1.1 $ $Date: 2002/02/21 09:46:42 $
 * @author      <a href="mailto:tima@intalio.com">Tim Anderson</a>
 * @see         TcpManagedConnection
 * @see         TcpConnector
 * @see         Messenger
 */
class TcpManagedConnectionAcceptor implements ManagedConnectionAcceptor {

    /**
     * The connector that manages this
     */
    private TcpConnector _connector;

    /**
     * The URI that connections are being accepted on
     */
    private final String _uri;

    /**
     * The listener socket
     */
    private ServerSocket _socket;

    /**
     * The set of ConnectionHandler instances, keyed on name
     */
    private HashMap _handlers = new HashMap();

    /**
     * The connection acceptor event listeners
     */
    private AcceptorEventNotifier _listeners = new AcceptorEventNotifier();

    /**
     * Create a new acceptor for the specified URI
     *
     * @param connector the connector that manages this
     * @param uri the URI that this accepts connections on
     * @throws IllegalArgumentException if any argument is null
     */
    public TcpManagedConnectionAcceptor(TcpConnector connector, String uri, 
                                        ServerSocket socket) {
        if (connector == null) {
            throw new IllegalArgumentException("Argument 'connector' is null");
        }
        if (uri == null) {
            throw new IllegalArgumentException("Argument 'uri' is null");
        }
        if (socket == null) {
            throw new IllegalArgumentException("Argument 'socket' is null");
        }
        _connector = connector;
        _uri = uri;
        _socket = socket;
    }

    /**
     * Returns the URI that connections are being accepted on
     *
     * @return the URI that connections are being accepted on
     */
    public String getURI() {
        return _uri;
    }

    /**
     * Register a connection handler to accept new connections
     *
     * @param uri the URI that the handler accepts connections for
     * @param handler the connection handler
     * @throws MalformedURLException if <code>uri</code> is an invalid URI
     * @throws RemoteException if the handler cannot be registered
     * @throws IllegalArgumentException if any argument is null
     */
    public synchronized void accept(String uri, ConnectionHandler handler) 
        throws MalformedURLException, RemoteException {
        if (uri == null) {
            throw new IllegalArgumentException("Argument 'uri' is null");
        }
        if (handler == null) {
            throw new IllegalArgumentException("Argument 'handler' is null");
        }
        URI parser = URIHelper.getURI(uri);
        String address = URIHelper.getAddress(parser);
        if (!address.equals(_uri)) {
            throw new MessengerException(
                "Mismatched URI. Acceptor is accepting physical connections " +
                "on URI=" + uri);
        }

        String name = parser.getPath();
        if (name == null || name.trim().length() == 0) {
            throw new MalformedURLException(
                "No connection name specified in URI=" + uri);
        }
        if (_handlers.containsKey(name)) {
            throw new MessengerException(
                "A connection handler is already registered for URI=" + uri);
        }

        _handlers.put(name, handler);
    }

    /**
     * Stop accepting connections for the specified URI.<br/>
     * If the URI represents the last connection handler listening for
     * connections, then the acceptor will be closed.
     *
     * @param uri the URI to close
     * @throws MalformedURLException if <code>uri</code> is an invalid URI
     * @throws RemoteException if an error occurs closing the connection
     * @throws IllegalArgumentException if <code>uri</code> is null
     */
    public synchronized void close(String uri) 
        throws MalformedURLException, RemoteException {
        if (uri == null) {
            throw new IllegalArgumentException("Argument 'uri' is null");
        }
        URI parser = URIHelper.getURI(uri);
        String address = URIHelper.getAddress(parser);
        if (!address.equals(_uri)) {
            throw new MessengerException(
                "Mismatched URI. Acceptor is accepting physical connections " +
                "on URI=" + uri);
        }
        String name = parser.getPath();
        ConnectionHandler handler = (ConnectionHandler) _handlers.remove(name);
        if (_handlers.isEmpty()) {
            // no more handlers registered, so close the acceptor
            close();
        }
    }

    /**
     * Listens for a connection to be made to this acceptor and accepts it.
     *
     * @return the new managed connection
     * @throws RemoteException if an error occurs when waiting for a 
     * connection
     */
    public ManagedConnection accept() throws RemoteException {
        Socket socket;
        try {
            socket = _socket.accept();
        } catch (IOException exception) {
            AcceptorEventNotifier listeners;
            synchronized (this) {
                listeners = _listeners;
                doClose();
            }
            if (listeners != null) {
                // notify any registered listeners
                MessengerException wrapper = new MessengerException(
                    exception.getMessage(), exception);
                listeners.error(this, wrapper);
            }
            throw new MessengerException("Failed to accept TCP connection",
                                         exception);
        }
        HashMap handlers;
        synchronized (this) {
            handlers = (HashMap) _handlers.clone();
        }
        return _connector.accepted(socket, handlers);
    }

    /**
     * Close the acceptor, freeing all associated resources
     *
     * @throws RemoteException if an error occurs closing the connection
     */
    public void close() throws RemoteException {
        AcceptorEventNotifier listeners;
        synchronized (this) {
            listeners = _listeners;
            doClose();
        }
        if (listeners != null) {
            // notify any registered listeners
            listeners.closed(this);
        }
    }

    /**
     * Add a listener for connection acceptor events
     *
     * @param listener the listener to add
     */
    public void addAcceptorEventListener(AcceptorEventListener listener) {
        _listeners.addAcceptorEventListener(listener);
    }

    /**
     * Remove a listener of connection acceptor events
     *
     * @param listener the listener to remove
     */
    public void removeAcceptorEventListener(AcceptorEventListener listener) {
        _listeners.removeAcceptorEventListener(listener);
    }

    /**
     * Performs the actual connection closure, without notifying registered
     * listeners
     *
     * @throws RemoteException if an error occurs closing the acceptor
     */
    protected void doClose() throws RemoteException {
        if (_socket != null) {
            try {
                _socket.close();
            } catch (IOException exception) {
                throw new MessengerException(exception.getMessage(), 
                                             exception);
            } finally {
                _connector = null;
                _socket = null;
                _listeners = null;
            }
        }
    }

} //-- TcpManagedConnectionAcceptor
