/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2002 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: TcpMultiplexerFactory.java,v 1.3 2002/02/20 10:23:48 tima Exp $
 */
package org.exolab.core.messenger.tcp;

// java
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Serializable;
import java.net.Socket;
import java.rmi.ConnectIOException;
import java.rmi.RemoteException;

// exolab
import org.exolab.core.messenger.ManagedConnection;
import org.exolab.core.messenger.Demultiplexer;
import org.exolab.core.messenger.Multiplexer;
import org.exolab.core.messenger.MultiplexerFactory;
import org.exolab.core.messenger.StreamDemultiplexer;
import org.exolab.core.messenger.StreamMultiplexer;


/**
 * A factory for {@link StreamMultiplexer} and {@link StreamDemultiplexer} 
 * objects, for use with a {@link TcpConnection}
 *
 * @version     $Revision: 1.3 $ $Date: 2002/02/20 10:23:48 $
 * @author      <a href="mailto:tima@intalio.com">Tim Anderson</a>
 * @see         StreamMultiplexer
 * @see         StreamDemultiplexer
 * @see         TcpConnection
 */
class TcpMultiplexerFactory implements MultiplexerFactory {
    
    /**
     * The open tcp socket
     */
    private Socket _socket;

    
    /**
     * Create a new tcp multiplexer factory
     *
     * @param socket the tcp socket
     * @throws IllegalArgumentException if <code>socket</code> is null
     */
    public TcpMultiplexerFactory(Socket socket) {
        if (socket == null) {
            throw new IllegalArgumentException("Argument 'socket' is null");
        }
        _socket = socket;
    }

    /**
     * Creates a new multiplexer for the specified connection
     *
     * @param connection the connection to create the multiplexer for
     * @return a new multiplexer for the connection
     * @throws RemoteException for any error
     * @throws IllegalArgumentException if connection is of the wrong type
     */
    public Multiplexer createMultiplexer(ManagedConnection connection) 
        throws RemoteException {
        Multiplexer result;
        if (!(connection instanceof TcpManagedConnection)) {
            throw new IllegalArgumentException(
                "Argument 'connection' must be of type " + 
                TcpManagedConnection.class.getName());
        }
        TcpManagedConnection tcp = (TcpManagedConnection) connection;
        try {
            result = new StreamMultiplexer(tcp, _socket.getOutputStream());
        } catch (IOException exception) {
            throw new ConnectIOException(exception.getMessage(), exception);
        }
        return result;
    }

    /**
     * Creates a new demultiplexer for the specified connection
     *
     * @param connection the connection to create the demultiplexer for
     * @return a new demultiplexer for the connection
     * @throws RemoteException for any error
     * @throws IllegalArgumentException if connection is of the wrong type
     */
    public Demultiplexer createDemultiplexer(ManagedConnection connection) 
        throws RemoteException {
        Demultiplexer result;
        if (!(connection instanceof TcpManagedConnection)) {
            throw new IllegalArgumentException(
                "Argument 'connection' must be of type " + 
                TcpManagedConnection.class.getName());
        }
        TcpManagedConnection tcp = (TcpManagedConnection) connection;
        try {
            result = new StreamDemultiplexer(tcp, _socket.getInputStream());
        } catch (IOException exception) {
            throw new ConnectIOException(exception.getMessage(), exception);
        }
        return result;
    }
    
} //-- TcpMultiplexerFactory
