/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2001-2004 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: Demultiplexer.java,v 1.8 2004/06/02 03:23:17 jalateras Exp $
 *
 * Date         Author  Changes
 * 02/14/2001   artw    Created
 */

package org.exolab.core.mipc;

import java.io.IOException;
import java.util.Hashtable;
import java.util.Iterator;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;


/**
 * This class is responsible for demultiplexing messages coming from a source
 * message input stream. Users of this class register output message streams
 * associated with each channel.  
 * 
 * Message coming in on the input stream are expected to have been placed
 * there by a {@link MultiplexOutputStream}. This class decodes the stream by 
 * using a {@link MulitplexInputStream}.  
 *
 * @version     $Revision: 1.8 $
 * @author      <a href="mailto:artw@wfactor.com">Art Whitten</a>
 * @see         MultiplexInputStream
 * @see         MultiplexOutputStream
 * @see         MultiplexConnection		
 */
class Demultiplexer {

    /**
     * A reference to the owning MultiplexConnection
     */
    private MultiplexConnectionIfc _owner;

    /**
     * The source input stream 
     */
    private MessageInput _in;
	
    /** 
     * The channel consumers. This maps channel names to MessageOutputStreams 
     */
    private Hashtable _targets = new Hashtable();

    /**
     * Used for stopping the Demultiplexer thread
     */
     private boolean _finished;

    /**
     * The logger
     */
    private static final Log _log = LogFactory.getLog(Demultiplexer.class);

    /**
     * Constructs a new Demultiplexer on the given input stream.
     *
     * @param in stream to read messages off
     * @param owner the owning multiplex connection that created this
     */
    public Demultiplexer(MessageInput in, MultiplexConnectionIfc owner) {
        _in = in;
        _owner = owner;
    }

    /**
     * Register a new channel consumer.  There can only be one consumer
     * per channel.  If there is an existing consumer, it is overwritten.
     *
     * @param channel the name of the channel.
     * @param out messages received on the channel are sent here
     */
    public synchronized void register(String channel, MessageOutput out) {
        _targets.put(channel, out);
    }

    /**
     * Remove a channel consumer.
     *
     * @param channel the name of the channel
     */
    public synchronized void deregister(String channel) {
        _targets.remove(channel);
    }

    /**
     * Gracefully stop the Demultiplexer thread.
     */
    public void finish() {
        _finished = true;
        try {
            _in.close();
        } catch (IOException ignore) {
        }
    }

    /**
     * This is called when the Demultiplexer thread is started. 
     * The Demultiplexer sits in a loop reading messages from the input stream
     * and dispatches them to the appropriate channel consumer.  Reads on the
     * input stream are blocking.  Users must call finish() to stop the thread.
     *
     */
    public void run() {
        try {
            while (!_finished) {
                Packet message = _in.receive();
                String channel = message.getChannel();
                MessageOutput out = (MessageOutput) _targets.get(channel);
                if (out != null) {
                    out.send(message);
                } else {
                    _log.debug("Demultiplexer - tossed a message " +
                               "because there is no listener for " +
                               "channel=" + channel);
                }
            }
        } catch (Exception exception) {
            if (!_finished) {
                _log.debug("Demultiplexer terminating on exception",
                           exception);
            }
        } finally {
            synchronized (this) {
                Iterator iterator = _targets.values().iterator();
                while (iterator.hasNext()) {
                    MessageOutput output = (MessageOutput) iterator.next();
                    try {
                        output.close();
                    } catch (Exception exception) {
                        _log.debug("Error closing channel", exception);
                    }
                }
            }
            _owner.disconnected();
            _owner = null;
            _in = null;
        }
    }

} //-- Demultiplexer
