/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2001-2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: MessageCopier.java,v 1.3 2003/06/09 06:28:41 tanderson Exp $
 *
 * Date         Author  Changes
 * 02/14/2001   artw    Created
 */

package org.exolab.core.mipc;

import java.io.IOException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;


/**
 * This class is responsible for copying messages from a given message input
 * stream to a given output stream.  It does its work on a separate thread.
 * The idea here is to isolate the senders of messages from underlying 
 * network issues. Typically the source input stream is a QueueInputStream.
 * 
 * See MultiplexConnection for usage. 
 *
 * @version     $Revision: 1.3 $
 * @author      <a href="mailto:artw@wfactor.com">Art Whitten</a>
 * @see         MessageInput
 * @see         MessageOutput
 * @see         MultiplexConnection		
 */
class MessageCopier extends Thread {

    /**
     * The source input stream 
     */
    private MessageInput _in;
	
    /**
     * The destination output stream 
     */
    private MessageOutput _out;
	
    /** 
     * We interrupt this thread when something is amiss 
     */
    private Thread _peer;

    /**
     * Used for stopping the MessageCopier thread
     */
    protected boolean _finished;

    
    /**
     * Seed for creating descriptive Thread names, which are useful for 
     * debugging.
     */
    private static int _copierNumber;

    /**
     *  The logger
     */
    private static final Log _log = LogFactory.getLog(MessageCopier.class);


    /**
     * Constructs a new MessageCopier
     *
     * @param peer the thread to interrupt when something goes wrong. 
     * Can be null.
     * @param in source stream
     * @param out destination stream
     */
    public MessageCopier(Thread peer, MessageInput in, MessageOutput out) {
        super("MessageCopier-" + nextCopierNum());
        setDaemon(true);
        _in = in;
        _out = out;
        _peer = peer;
    }

    /**
     * Gracefully stop the MessageCopier thread.
     */
    public void finish() {
        _finished = true;
        interrupt();
        	
        // close the input and output channles
        try {
            _in.close();
            _out.close();
        } catch (Exception exception) {
            // just swallow :-)
        }
    }

    /**
     * This is called when the MessageCopier thread is started. It is the 
     * implementation of the Runnable interface.
     * 
     * The MessageCopier sits in a loop pulling messages from the
     * source InputStream and places them on the destination OutputStream.
     * If an IOException occurs, the peer thread is interrupted (if there is 
     * one).
     *
     */
    public void run () {
        try {
            copy();
        }  catch (Exception exception) {
            if (!_finished) {
                _log.debug(
                    "MessageCopier terminating on exception", exception);
            }
            if(_peer != null) {
                _peer.interrupt();
            }
        }
    }

    /**
     * The message copying loop.
     *
     * The MessageCopier sits in a loop pulling messages from the
     * source InputStream and places them on the destination OutputStream.
     * The loop is terminated when an IOException occurs or if the thread
     * is interrupted.
     */
    protected void copy () throws ClassNotFoundException, IOException {
        while (!interrupted()) {
            Packet message = _in.receive();
            _out.send(message);
	}
    }

    private static synchronized int nextCopierNum () { 
        return _copierNumber ++; 
    }

} //-- MessageCopier
