/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2001-2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: MultiplexConnection.java,v 1.6 2003/08/26 04:59:19 tanderson Exp $
 *
 * Date         Author  Changes
 * 02/14/2001   artw    Created
 */

package org.exolab.core.mipc;

import java.io.InputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.net.Socket;
import java.net.UnknownHostException;

import org.apache.commons.logging.LogFactory;
import org.apache.commons.logging.Log;


/** 
 * The MultiplexConnection encapsulates a single network connection.  It allows
 * for multiple channels or multiplexed message streams.  The object is thread
 * safe and uses threads within it to do its job.
 * 
 * The design for the Multiplexed streaming borrows concepts from the book
 * "Java Network Programming : A Complete Guide to Networking, Streams, 
 * and Distributed Computing", by Merlin Hughes, Michael Shoffner, Derek Hamner
 *  
 * The MultiplexConnection is designed as follows:
 *
 *  - users of the connection register themselves as channels.  each channel
 *  is given a stream to place outbound messages on, and gives the MC a
 *  stream to write inbound messages on.  Typically this "inbound" stream
 *  is a Queue Stream and the user sucks messages off of the queue
 *  on a separate thread.  That way the multiplex delivery mechanism
 *  does not have to wait for a channel handler to process a message.  
 *
 *  - As channel producers send messages, they are multiplexed and placed
 *  on a queue for outbound delivery.  A separate MessageCopier thread takes
 *  messages from this queue and sends them onto to the "raw" outbound stream.
 *
 *  - Inbound messages are handled by Demultiplixer.  The Demux manages a list
 *  of channels.  It unwraps a packet and uses the name encoded in the packet
 *  to pass the message onto an output stream.  This outputstream is the one
 *  passed into "register" by the channel consumer/producer.   The Demux runs 
 *  within the same thread as MultiplexConnection.
 *
 *  It is important to note:  All activity on the "raw" inbound and outbound
 *  streams are synchronized, as well as the objects used to build the
 *  MultiplexConnection.  So everything is thread-safe. 
 *
 * @version     $Revision: 1.6 $
 * @author      <a href="mailto:artw@wfactor.com">Art Whitten</a>
 * @see         MultiplexConnectionServer
 * @see         MessageCopier
 * @see         Demultiplexer
 */

public class MultiplexConnection extends Thread 
    implements MultiplexConnectionIfc {
    
    /**
     * The raw input stream 
     */
    private InputStream _in;

    /** 
     * The raw output stream 
     */
    private OutputStream _out;

    /** 
     * Outbound messages are placed on this queue 
     */
    private Queue _queue;

    /** 
     * Queued outbound message stream.  It wraps _queue.  As channels register
     * themselves, a MutliplexOutputStream is created that wraps this stream
     */
    private QueueOutputStream _queueOut;

    /** 
     * Listens to inbound messages on _in and routes them to the appropriate
     * channels
     */
    private Demultiplexer _demux;

    /**
     * Copies messages from _queue to _out
     */
    private MessageCopier _copier;

    /** 
     * This is a listener that will be notified if a mutliplex connection is
     * abnormally terminated
     */
    private DisconnectionEventListener _listener = null;

    /**
     * Used for stopping the thread
     */
    protected boolean _finished = false;     

    /**
     * Cache the hostname of this connection
     */
    private String _host = null;

    /**
     * Cache the port number of this connection
     */
    private int _port = 0;

    /**
     *  The logger
     */
    private static final Log _log =
        LogFactory.getLog(MultiplexConnection.class);


    /**
     * Creates new MultiplexConnection on the specified host and port.
     *
     * @param hostname the hostname to create a socket connection to
     * @param port the port to connect to
     */
    public MultiplexConnection(String hostname, int port) 
        throws UnknownHostException, IOException {
        this(new Socket(hostname, port));
    }

    /**
     * Creates a MultiplexConnection on top of an existing Socket
     *
     * @param socket The existing socket to wrap.
     */
    public MultiplexConnection(Socket socket) throws IOException {
        super(getName(socket));
        init(socket);
    }

    /**
     * Creates a MultiplexConnection on top of an existing Socket
     *
     * @param group the thread group that the connection belongs to
     * @param socket the existing socket to wrap
     */
    public MultiplexConnection(ThreadGroup group, Socket socket) 
        throws IOException {
        super(group, getName(socket));
        init(socket);
    }

    /**
     * Gracefully stop the MultiplexConnection thread.
     */
    public void finish () {
        _finished = true;
        if (_copier != null) {
            _copier.finish();
        }
        if (_demux != null) {
            _demux.finish();
        }
    }

    /** 
     * Set the listener that will be called back when the multiplex connection
     * is dropped
     *
     * @param listener - the disconnection event listener
     */
    public void setDisconnectionEventListener(
        DisconnectionEventListener listener) {
        _listener = listener;
    }

    /**
     * The workhorse for the MultiplexConnection.  Typically this method is
     * invoked as a result of calling 'start' on the MultiplexConnection.
     * MultiplexConnection derives from Thread.
     */
    public void run () {
        // Outbound messages are placed on '_queue'.  Setup an 
        // input stream on the queue and a message copier that
        // will take the messages from the queue and deliver them onto
        // the outbound "raw" network connection.
        QueueInputStream queueIn = new QueueInputStream(_queue);
        MessageOutputStream messageOut = new MessageOutputStream(_out);
        _copier = new MessageCopier(this, queueIn, messageOut);

        try {
            // Start up outbound message delivery thread
            _copier.start();

            // The demultiplexer runs within the current thread. It receives 
            // inbound messages and dispatches them to the registered channels.
            _demux.run();
        } catch (Exception exception) {
            if (!_finished) {
                _log.debug(
                    "MultiplexConnection terminating on exception", exception);
            }
        } finally {
            _copier.finish();
        }
    }

    /**
     * Register a new multiplex channel.  'label' is the identifier for the 
     * channel and 'messageOutput' is where inbound messages are placed.  
     * If there is an existing channel, it is overwritten.
     * 
     * The return value is a MessageOutput that is used by the channel 
     * producer to send messages.  The interface encapsulates a 
     * MutliplexOutputStream.
     *
     * @param label The channel name to register for
     * @param messageOutput Inbound messages are placed here
     * @return MessageOuput Channel producers place outbound message on this 
     * stream.
     */
    public MessageOutput register(String label, MessageOutput out) {
        // The Demultiplexer will send all messages it receives for "label"
        // and place them on "out"
        _demux.register(label, out);
        // Return a stream in which the channel producer can place outbound
        // messages.
        return _queueOut;
    }

    /**
     * Deregister a multiplex channel.  
     *
     * @param label the name of the channel to unregister
     */
    public void deregister(String label) {
        _demux.deregister(label);
    }

    /** 
     * This is called by the demultiplexer when the channel has been abnormally
     * terminated. It will call finish on the connection and then notify any
     * registered listener
     */
    public void disconnected() {
        finish();
        if (_listener != null) {
            _listener.disconnected(this);
        }
    }

    /**
     * Return the hostname that this connection bound to
     *
     * @return the host name
     */
    public String getHost() {
        return _host;
    }

    /**
     * Return the port that this connection is bound to
     *
     * @return the port number
     */
    public int getPort() {
        return _port;
    }

    protected void init(Socket socket) throws IOException {
        _in = socket.getInputStream();
        _out = socket.getOutputStream();

        // cache some attributes of this connection
        _host = socket.getLocalAddress().getHostName();
        _port = socket.getLocalPort();
        _queue = new Queue();
        _queueOut = new QueueOutputStream(_queue);

        _demux = new Demultiplexer(new MessageInputStream(_in), this);
        setDaemon(true);
    }

    private static String getName(Socket socket) throws IOException {
        String host = socket.getLocalAddress().getHostName();
        int port = socket.getLocalPort();
        return "MultiplexConnection-" + host + ":" + port;
    }

} //-- MultiplexConnection
