/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2001-2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: MultiplexConnectionServer.java,v 1.5 2003/06/19 05:58:52 tanderson Exp $
 *
 * Date         Author  Changes
 * 02/14/2001   artw    Created
 */

package org.exolab.core.mipc;

import java.io.IOException;
import java.io.Serializable;
import java.net.InetAddress;
import java.net.Socket;
import java.net.ServerSocket;
import java.net.UnknownHostException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;


/**
 * Create a listener socket and wait for requests to come in. When a 
 * MultiplexConnection is made, a client is notified so that it can initialize
 * channels on the freshly minted MultiplexConnection.
 *
 * This class derives from Thread and is meant to be spawned off as a separate
 * Thread.
 *
 * The design for the Multiplexed streaming borrows concepts from the book
 * "Java Network Programming : A Complete Guide to Networking, Streams, 
 * and Distributed Computing", by Merlin Hughes, Michael Shoffner, Derek Hamner
 * 
 * @version     $Revision: 1.5 $
 * @author      <a href="mailto:artw@wfactor.com">Art Whitten</a>
 * @see         org.exolab.core.mipc.MultiplexConnection
 * @see         org.exolab.core.mipc.ConnectionNotifierIfc
 */
public class MultiplexConnectionServer extends Thread
    implements MultiplexConnectionServerIfc {

    /**
     * The main server socket to listen for incoming calls
     */
    private ServerSocket _server;

    /**
     * The list of all current connections to the server
     */
     private ThreadGroup _connections;

    /**
     * While the active flag is set, the server continues to run
     */
    private boolean _active = true;

    /**
     * Callback for new connections
     */
    private ConnectionNotifierIfc _notifier = null;

    /**
     * The logger
     */
     private static final Log _log =
         LogFactory.getLog(MultiplexConnectionServer.class);

    /**
     * Creates a new MultiplexConnectionServer that sets up a listener port
     * and adds a notifier to call with all new connections.
     *
     * @param port the port to connect on
     * @param notifier the client to handle the call when a new connection is 
     * made
     */
    public MultiplexConnectionServer(int port, ConnectionNotifierIfc notifier)
        throws IOException {
        this(null, port, notifier);
    }

    /**
     * Creates a new MultiplexConnectionServer that sets up a listener port
     * and adds a notifier to call with all new connections.
     *
     * @param address the local IP address to bind to. May be null.
     * @param port the port to connect on
     * @param notifier the client to handle the call when a new connection is 
     * made
     */
    public MultiplexConnectionServer(InetAddress address, int port, 
                                     ConnectionNotifierIfc notifier)
        throws IOException {
        _connections = new ThreadGroup("MultiplexConnections");
        _notifier = notifier;
        _server = new ServerSocket(port, -1, address);
    }

    /**
     * Start the service running. This call never returns until the service
     * is shutdown.
     *
     * When a connection is made, a new MultiplexConnection is created and 
     * spawned off as a separate thread.  The new threads are created in the
     * ThreadGroup _connections so that we can keep track of the active 
     * threads.
     */
    public void run() {
        _log.debug("MultiplexConnectionServer is ready, on port " +
                   _server.getLocalPort());
        while (_active) {
            try {
                Socket s = _server.accept();
                _log.debug(
                    "MultiplexConnectionServer accept()'ed connection");
                synchronized(_connections) {
                    if(_active == true) {
                        MultiplexConnection mc = new MultiplexConnection(
                            _connections, s);
                        _notifier.connected(mc);
                        mc.start();
                    } else {
                        s.close();
                    }
                }
            } catch(IOException exception) {
                _log.error("Failed to create new connection", exception);
            }
        }
    }

    /**
     * Shutdown all active connections, including the server connection.
     *
     * @throws IOException If the service reports an error during shutdown.
     */
    public void shutdownAll() throws IOException {
        _log.debug("MultiplexConnectionServer shutdownAll called.");
        synchronized(_connections) {
            int size = _connections.activeCount();
            Thread[] connections = new Thread[size];
            
            _connections.enumerate(connections, false);
 
            for (int i = 0, j = size; i < j; ++i) {
                ((MultiplexConnectionIfc) connections[i]).finish();
            }
		
            _active = false;
            _server.close();
            interrupt();
        }
    }
    
    /**
     * Return the port number this server is listening on.
     *
     * @return the port number the server is using.
     */
    public int getPort() {
        return _server.getLocalPort();
    }

    /**
     * Return the machines address.
     * 
     * @return the machines IP address.
     */
    public String getHost() {
        String	host = TcpConstants.LOCAL_HOST;
		
        try {
            host = InetAddress.getLocalHost().getHostAddress();
        } catch(UnknownHostException exception) {
            _log.error("Failed to get IP address for this host",
                       exception);
        }
        return host;
    }

} //-- MultiplexConnectionServer
