/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2001-2002 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: MultiplexSSLConnection.java,v 1.5 2003/06/09 06:28:41 tanderson Exp $
 *
 * Date         Author      Changes
 * 07/06/2001   mourikis    Created
 */
package org.exolab.core.mipc;

// java io
import java.io.IOException;

// java net
import javax.net.ssl.SSLSocketFactory;
import javax.net.ssl.SSLSocket;

// exolab
import org.apache.commons.logging.LogFactory;
import org.apache.commons.logging.Log;


/** 
 * The MultiplexSSLConnection encapsulates a single network connection.  
 * It allows for multiple channels or multiplexed message streams.  
 * The object is thread-safe and uses threads within it to do its job.
 * 
 *	The design for the Multiplexed streaming borrows concepts from the book
 *	"Java Network Programming : A Complete Guide to Networking, Streams, 
 *	and Distributed Computing", by Merlin Hughes, Michael Shoffner, 
 *      Derek Hamner
 *
 *  The MultiplexConnection is designed as follows:
 *
 *  - users of the connection register themselves as channels.  each channel
 *  is given a stream to place outbound messages on, and gives the MC a
 *  stream to write inbound messages on.  Typically this "inbound" stream
 *  is a Queue Stream and the user sucks messages off of the queue
 *  on a separate thread.  That way the multiplex delivery mechanism
 *  does not have to wait for a channel handler to process a message.  
 *
 *  - As channel producers send messages, they are multiplexed and placed
 *  on a queue for outbound delivery.  A separate MessageCopier thread takes
 *  messages from this queue and sends them onto to the "raw" outbound stream.
 *
 *  - Inbound messages are handled by Demultiplixer.  The Demux manages a list
 *  of channels.  It unwraps a packet and uses the name encoded in the packet
 *  to pass the message onto an output stream.  This outputstream is the one
 *  passed into "register" by the channel consumer/producer.   The Demux runs 
 *  within the same thread as MultiplexConnection.
 *
 *  It is important to note:  All activity on the "raw" inbound and outbound
 *  streams are synchronized, as well as the objects used to build the
 *  MultiplexConnection.  So everything is thread-safe. 
 *
 * @version     $Revision: 1.5 $
 * @author      <a href="mailto:mourikis@intalio.com">Jim Mourikis</a>
 * @see         MultiplexConnectionServer
 * @see         MessageCopier
 * @see         Demultiplexer
 */
public class MultiplexSSLConnection extends MultiplexConnection {

    /**
     * Creates new MultiplexSSLConnection on the specified host and port.
     *
     * @param host the host name to create a socket connection to
     * @param port the port to connect to
     */
    public MultiplexSSLConnection(String host, int port) 
        throws IOException {
        super(getSocket(host, port));
    }

    /**
     * Creates a MultiplexSSLConnection on top of an existing Socket
     *
     * @param socket The existing socket to wrap.
     */
    public MultiplexSSLConnection(SSLSocket socket) throws IOException {
        super(socket);
    }

    /**
     * Creates a MultiplexSSLConnection on top of an existing Socket
     *
     * @param group the thread group that the connection belongs to
     * @param socket the existing socket to wrap
     */
    public MultiplexSSLConnection(ThreadGroup group, SSLSocket socket) 
        throws IOException {
        super(group, socket);
    }
    
    private static SSLSocket getSocket(String host, int port) 
        throws IOException {
        SSLSocketFactory sslFact = (SSLSocketFactory) 
            SSLSocketFactory.getDefault();
        return (SSLSocket) sslFact.createSocket(host, port);
    }

} //-- MultiplexSSLConnection
