/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2001-2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: ObjectChannel.java,v 1.4 2003/02/16 01:57:53 tanderson Exp $
 *
 * Date         Author  Changes
 * 02/14/2001   artw    Created
 */

package org.exolab.core.mipc;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.io.IOException;

import org.exolab.core.ipc.IpcIfc;


/**
 * The ObjectChannel is associated with a MultiplexConnection.  It makes it 
 * easy to read and write objects from/to a multiplex channel.  
 *
 * @version     $Revision: 1.4 $
 * @author      <a href="mailto:artw@wfactor.com">Art Whitten</a>
 * @see         MultiplexConnection
 */
public class ObjectChannel implements IpcIfc {

    /**
     * The underlying connection to read/write
     */
    private MultiplexConnectionIfc _connection = null;

    /**
     * The input stream on top of _connection
     */
     private MessageInput _input = null;
    
    /**
     * The output stream on top of _connection. 
     * This is managed by the connection.
     */
    private MessageOutput _output = null;
        
    /**
     * The multiplex channel name
     */
    private String _label = null;
    
    /**
     * Creates a new ObjectChannel with the given name on the specified
     * connection.
     *
     * @param label the name of the multiplex channel to read/write from/to.
     * @param connection the underlying connection 
     */
    public ObjectChannel(String label, MultiplexConnectionIfc connection) {
        _connection = connection;
        _label = label;
        
        // This is where we want our messages placed
        Queue queue = new Queue();
        QueueOutputStream queueOut = new QueueOutputStream(queue);
		
        _output = _connection.register(_label, queueOut);
        _input = new QueueInputStream(queue);
    }

    /**
     * Close the channel
     *
     * @throws IOException
     */
    public void close()  throws IOException {
        _connection.deregister(_label);
        _input.close();
    }

    /**
     * Send an object on the underlying connection
     *
     * @param object the object to send
     * @throws IOException if the object cannot be written to the output 
     * stream.
     */
    public void send(Serializable object) throws IOException {
        _output.send(new Packet(_label, object));
    }
	
    /**
     * Receive an object from the underlying connection
     *
     * @return the received object
     * @throws IOException if the object cannot be received
     */
    public Object receive() throws ClassNotFoundException, IOException {
        Packet packet = _input.receive();
        return packet.getObject();
    }
    
    /**
     * Retrieve a unique id for this connection.
     *
     * @return the unique string id
     */
    public String toString() {
        return ("ObjectChannel - " + _label);
    }
	
    /**
     * Get the connection that this encapsulates 
     *
     * @return MultiplexConnection the underlying connection object.
     */
    public MultiplexConnectionIfc getConnection() {
        return _connection;
    }

} //-- ObjectChannel
