/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2001 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: Queue.java,v 1.7 2003/02/13 11:13:24 tanderson Exp $
 *
 * Date         Author  Changes
 * 02/14/2001   artw    Created
 */

package org.exolab.core.mipc;

import java.util.LinkedList;

/**
 * A simple FIFO Queue implementation that allows for concurrent access
 * by consumers and producers.  Remove()'s are blocking.
 *
 *
 * @version     $Revision: 1.7 $
 * @author      <a href="mailto:artw@wfactor.com">Art Whitten</a>
 **/

public class Queue {

    /**
     * Implements the actual queueing
     */
    private LinkedList queue_;

    /**
     * This flag is used to mark the connection as closed.
     */
    private boolean closed_ = true;

    /**
     * Creates a new empty FIFO Queue.
     */
    public Queue () {
        queue_ = new LinkedList ();
        closed_ = false;
    }

    /**
     * Remove an Object from the top of the queue. This call blocks
     * until an object is available.  
     *
     * @return Object The next object available.
     */
    public Object remove () throws InterruptedException
    {
        Object item = null;
        synchronized (queue_) {
            while ((!closed_) && (queue_.isEmpty ())) {
                queue_.wait();
            }
            
            // check to see that we have not been forced to exit
            if (!closed_) {
                item = queue_.removeFirst();
            }
            
            if (queue_.size() < 100) {
                queue_.notifyAll();
            }
            return item;
        }
    }

    /**
     * Push a new Object onto the bottom of the queue.
     * Any threads blocking for an object will be notified.
     *
     * @param item Add this object.
     */
    public void add (Object item) {
        synchronized(queue_) {
            if (queue_.size() > 1000) {
                try {
                    queue_.wait();
                } catch (InterruptedException err) {}
            }
            queue_.add (item);
            queue_.notifyAll();
        }
    }

    /**
     * Is the queue empty?
     *
     * @return boolean True if empty, otherwise False.
     */
    public boolean isEmpty() {
        synchronized(queue_) {
            return queue_.isEmpty ();
        }
    }

    /**
     * Close the queue. If it is blocked then unblock it and release
     * it so it can close correctly
     */
    public void close() {
        synchronized(queue_) {
            closed_ = true;
            queue_.notifyAll();
        }
    }
}
