/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 1999-2001,2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: Service.java,v 1.2 2003/06/09 06:28:41 tanderson Exp $
 *
 * Date         Author  Changes
 * 12/23/1999   jima    Created
 * 03/01/2000   jima    Changed the constructor to only take a name. There
 *                      is no need for the Runnable arg, since this object
 *                      is Runnable
 */
package org.exolab.core.service;


/**
 * Service is an implementation of the {@link Servicable} interface that 
 * provides default implementations for the {@link #start} and {@link #stop}
 * methods.
 *
 * @version     $Revision: 1.2 $ $Date: 2003/06/09 06:28:41 $
 * @author      <a href="mailto:jima@intalio.com">Jim Alateras</a>
 * @see         Serviceable
 */
public abstract class Service implements Serviceable {

    /**
     * The name of the service
     */
    private String _name = null;

    /**
     * The state of the service
     */
    private volatile ServiceState _state = ServiceState.STOPPED;

    /**
     * The default constructor for the Service does nothing.
     */
    protected Service() {
    }

    /**
     * Instantiates a service with the specified name
     * 
     * @param name the unique name of the service       
     */
    protected Service(String name) {
        _name = name;
    }

    /**
     * Start the service.
     *
     * @throws ServiceException if the service fails to start, or is already
     * running
     */
    public synchronized void start() throws ServiceException {
        if (_state.isStopped()) {
            _state = ServiceState.RUNNING;
        } else {
            throw new ServiceException("Failed to start service " + this); 
        }
    }

    /**
     * Stop the service.
     *
     * @throws ServiceException if the service fails to stop, or is already
     * stopped
     */
    public synchronized void stop() throws ServiceException {
        if (_state.isRunning()) {
            _state = ServiceState.STOPPED;
        } else {
            throw new ServiceException("Failed to stop service " + this); 
        }
    }

    /**
     * Convenience method for restarting the service. This operation can
     * be called regardless the current state of the service.
     *
     * @throws ServiceException if the service fails to restart
     */
    public synchronized void restart() throws ServiceException {
        if (_state.isRunning()) {
            stop();
        }
        start();
    }

    /**
     * Return the current state of the service.
     *
     * @return the current state of the service
     */
    public ServiceState getState() {
        return _state;
    }

    /**
     * Return the name of the service.
     *
     * @return the service name
     */
    public String getName() {
        return _name;
    }

    /**
     * Return the state of the object as a string
     *
     * @return a string form of the object state
     */
    public String toString() {
        StringBuffer buf = new StringBuffer("Service:[");
        buf.append("name=");
        buf.append(_name);
        buf.append("state=");
        buf.append(_state);
        buf.append("]");
        return buf.toString();
    }

    /**
     * Set the state of the service
     *
     * @param state the service state
     */
    protected void setState(ServiceState state) {
        _state = state;
    }
    
} //-- Service
