/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 1999-2001 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * Date         Author  Changes
 * 12/23/1999   jima    Created
 */

package org.exolab.core.service;


/**
 * This class defines an enumeration of service states. 
 *
 * stopped      the service is not running
 * running      the service is running
 * 
 * @version     $Revision: 1.3 $ $Date: 2001/10/07 07:28:05 $
 * @author      <a href="mailto:jima@intalio.com">Jim Alateras</a>
 */
public final class ServiceState {

    /**
     * Instatiate instances related to each of the enumerations
     */
    public static final ServiceState STOPPED = new ServiceState("stopped");
    public static final ServiceState RUNNING = new ServiceState("running");

    private static int _upperSentinel = 0;

    /**
     * The state name
     */ 
    private String _state;

    /**
     * The ordinal value of the state
     */
    private int _ord;

    /**
     * Limit the creation scope of this object. All instances of this 
     * class are created at class initialisation time.
     *
     * @param state the state name
     */
    private ServiceState(String state) {
        _state = state;
        _ord = _upperSentinel++;
    }

    /**
     * Returns the ordinal value for this state
     *
     * @return the ordinal value of the state
     */
    public int getOrd() {
        return _ord; 
    }

    /**
     * Returns the name of this state
     *
     * @return the state name
     */
    public String toString() {
        return _state;
    }

    /**
     * Returns the number of states in this enumeration
     *
     * @return the number of states in the enumeration
     */
    public static int size() {
        return _upperSentinel; 
    }

    /**
     * Return true if the two specified object are of the same type 
     * and their ordinal values are the same.
     *
     * @param object the object to compare against
     * @return true if the two objects are equivalent
     */
    public boolean equals(Object object) {
        boolean equal = (object == this);
        if (!equal && object instanceof ServiceState) {
            equal = (_ord == ((ServiceState) object)._ord);
        }

        return equal;
    }

    /**
     * Return true if this object represents the 'running' state and false
     * otherwise
     *
     * @return true if this state is the 'running' state
     */
    public boolean isRunning() {
        return getOrd() == RUNNING.getOrd();
    }

    /**
     * Return true if this object represents the 'stopped' state and false
     * otherwise
     *
     * @return true if this state is the 'stopped' state
     */
    public boolean isStopped() {
        return getOrd() == STOPPED.getOrd();
    }

} //-- ServiceState
