/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 1999 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: LruObjectCacheCT.java,v 1.1 2000/06/07 09:17:28 jima Exp $
 *
 * Date         Author  Changes
 * 06/07/2000   jima    Created
 */
package org.exolab.testharness.core.foundation;

import java.io.IOException;
import java.util.Hashtable;
import java.util.Enumeration;
import org.exolab.jtf.CWTestCategory;
import org.exolab.jtf.CWTestCase;
import org.exolab.jtf.CWVerboseStream;
import org.exolab.exceptions.CWClassConstructorException;
import org.exolab.jtf.CWVerboseStream;
import org.exolab.core.foundation.LruObjectCache;


/**
 * This file includes test cases for te LruObjectCache. This will check the
 * following aspects of the LruObjectCache
 * <p>
 * TC01 Test the creation and polution of the test case
 * TC02 Test removal of entries from the cache
 * TC03 Test the LRU algorithm.
 *
 * @version     $Revision: 1.1 $ $Date: 2000/06/07 09:17:28 $
 * @author      <a href="mailto:jima@exoffice.com">Jim Alateras</a>
 * @see         org.exolab.jtf.CWTestCateogry
 * @see         org.exolab.jtf.CWTestCase
 */
public class LruObjectCacheCT
    extends CWTestCategory 
{
    /**
     * The constructor is responsible for constructing a test category and
     * adding the suite of test cases. It throws CWClassConstructorException
     * if it cannot construct the category
     *
     * @exception   CWClassConstructorException   
     */
    public LruObjectCacheCT() 
        throws CWClassConstructorException
	{
		super("lrucache", "LRU Object Cache Component");
		CWTestCase tc;

		tc = new LruObjectCacheComponentTC01();
		add(tc.name(), tc, true);
		tc = new LruObjectCacheComponentTC02();
		add(tc.name(), tc, true);
		tc = new LruObjectCacheComponentTC03();
		add(tc.name(), tc, true);
    }
}


/**
 * This is a base class that has some common helper functions. All other test
 * cases extend this class.
 */
abstract class LruObjectCacheBase extends CWTestCase
{
    LruObjectCacheBase(String id, String info)
        throws CWClassConstructorException
    {
        super(id, info);
    }

    /**
     * Iterate and print the entries in the cache using the 
     * writeVerbose flag, The entries are all expected to be 
     * Strings
     */
    protected void printCacheEntries(CWVerboseStream stream,
        LruObjectCache cache)
        throws IOException
    {
        Enumeration iterator = cache.elements();
        String last = null;
        while (iterator.hasMoreElements())
        {
            last = (String)iterator.nextElement();
            stream.writeVerbose(last);
        }
        stream.writeVerbose("\n");
    }
}


/**
 * 1. Create an LruObjectCache instance with max size 10.
 * 1. Add five entries to the cache
 * 3. Iterate through the entries and ensure that they are all
 *    present. The last entry is the least-recently used
 * 4. Retrieve the 3rd entry and ensure that the is it the most
 *    recently used
 */
class LruObjectCacheComponentTC01 extends LruObjectCacheBase
{
    /**
     * Calls the super class and throws CWClassConstructorException if it fails
     *
     * @exception   CWClassConstructorException
     */
    public LruObjectCacheComponentTC01() 
        throws CWClassConstructorException
	{
		super("TC01", "Test the creation and polution of the test case");
	}

    /**
     * Execute the test case described in the class header. The CWVerboseStream
     * is used to display detailed information on the test case execution
     *
     * @param           stream          this is where verbose info is streamed
     * @return          boolean         true if successful; false otherwise
     */
	public boolean run(CWVerboseStream stream)
	{
		boolean result = true;

		try
		{
            LruObjectCache cache = new LruObjectCache(10);

            cache.insert(new Long(1), "this is 1");
            cache.insert(new Long(2), "this is 2");
            cache.insert(new Long(3), "this is 3");
            cache.insert(new Long(4), "this is 4");
            cache.insert(new Long(5), "this is 5");

            stream.writeVerbose("Added all the entries to the cache");
            Enumeration iterator = cache.elements();
            String last = null;
            int count = 0;
            while (iterator.hasMoreElements())
            {
                last = (String)iterator.nextElement();
                stream.writeVerbose(last);
                count++;
            }

            if ((count != 5) ||
                (!last.equals("this is 1")))
            {
                result = false;
            }

            stream.writeVerbose((String)cache.get(new Long(3)));
            iterator = cache.elements();
            last = (String)iterator.nextElement();
            stream.writeVerbose("Most recently touched entry " + last);
            if (!last.equals("this is 3"))
            {
                result = false;
            }
 		}
		catch(Exception exception)
		{
			System.out.println("[ERR : LruObjectCacheComponentTC01] " + 
                exception);
            exception.printStackTrace();
		}

		return result;
	}
}

/**
 * 1. Create an LruObjectCache instance with max size 10.
 * 1. Add five entries to the cache
 * 3. Remove a couple of entries
 * 4. Ensure that the correct number of entries are present
 */
class LruObjectCacheComponentTC02 extends LruObjectCacheBase
{
    /**
     * Calls the super class and throws CWClassConstructorException if it fails
     *
     * @exception   CWClassConstructorException
     */
    public LruObjectCacheComponentTC02() 
        throws CWClassConstructorException
	{
		super("TC02", "Test removal of entries from the cache");
	}

    /**
     * Execute the test case described in the class header. The CWVerboseStream
     * is used to display detailed information on the test case execution
     *
     * @param           stream          this is where verbose info is streamed
     * @return          boolean         true if successful; false otherwise
     */
	public boolean run(CWVerboseStream stream)
	{
		boolean result = true;

		try
		{
            LruObjectCache cache = new LruObjectCache(10);

            cache.insert(new Long(1), "this is 1");
            cache.insert(new Long(2), "this is 2");
            cache.insert(new Long(3), "this is 3");
            cache.insert(new Long(4), "this is 4");
            cache.insert(new Long(5), "this is 5");

            printCacheEntries(stream, cache);
            if (cache.getUsedSlots() != 5)
            {
                result = false;
            }

            cache.remove(new Long(2));
            printCacheEntries(stream, cache);
            if (cache.getUsedSlots() != 4)
            {
                result = false;
            }

            cache.remove(new Long(4));
            printCacheEntries(stream, cache);
            if (cache.getUsedSlots() != 3)
            {
                result = false;
            }

            cache.remove(new Long(1));
            printCacheEntries(stream, cache);
            if (cache.getUsedSlots() != 2)
            {
                result = false;
            }

            cache.remove(new Long(5));
            printCacheEntries(stream, cache);
            if (cache.getUsedSlots() != 1)
            {
                result = false;
            }

            cache.remove(new Long(3));
            printCacheEntries(stream, cache);
            if (cache.getUsedSlots() != 0)
            {
                result = false;
            }

 		}
		catch(Exception exception)
		{
			System.out.println("[ERR : LruObjectCacheComponentTC02] " + 
                exception);
            exception.printStackTrace();
		}

		return result;
	}
}

/**
 * 1. Create an LruObjectCache instance with max size 10.
 * 1. Add five entries to the cache
 * 3. Iterate through the entries and ensure that they are all
 *    present. The last entry goes is the most-recently used
 */
class LruObjectCacheComponentTC03 extends LruObjectCacheBase
{
    /**
     * Calls the super class and throws CWClassConstructorException if it fails
     *
     * @exception   CWClassConstructorException
     */
    public LruObjectCacheComponentTC03() 
        throws CWClassConstructorException
	{
		super("TC03", "Test the LRU algorithm");
	}

    /**
     * Execute the test case described in the class header. The CWVerboseStream
     * is used to display detailed information on the test case execution
     *
     * @param           stream          this is where verbose info is streamed
     * @return          boolean         true if successful; false otherwise
     */
	public boolean run(CWVerboseStream stream)
	{
		boolean result = true;

		try
		{
            LruObjectCache cache = new LruObjectCache(10);

            cache.insert(new Long(1), "this is 1");
            cache.insert(new Long(2), "this is 2");
            cache.insert(new Long(3), "this is 3");
            cache.insert(new Long(4), "this is 4");
            cache.insert(new Long(5), "this is 5");
            cache.insert(new Long(6), "this is 6");
            cache.insert(new Long(7), "this is 7");
            cache.insert(new Long(8), "this is 8");
            cache.insert(new Long(9), "this is 9");
            cache.insert(new Long(10), "this is 10");
            printCacheEntries(stream, cache);

            cache.get(new Long(7));
            printCacheEntries(stream, cache);
            Enumeration iterator = cache.elements();
            String last = (String)iterator.nextElement();
            if (!last.equals("this is 7"))
            {
                System.err.println("False at 1");
                result = false;
            }

            cache.insert(new Long(11), "this is 11");
            cache.insert(new Long(12), "this is 12");
            cache.insert(new Long(13), "this is 13");
            printCacheEntries(stream, cache);

            if ((cache.get(new Long(1)) != null) ||
                (cache.get(new Long(2)) != null) ||
                (cache.get(new Long(3)) != null))
            {
                System.err.println("Failed in retrieval " + 
                    cache.get(new Long(1)) + "," +
                    cache.get(new Long(2)) + ","  +
                    cache.get(new Long(3)));
                result = false;
            }

            iterator = cache.elements();
            last = (String)iterator.nextElement();
            if (!last.equals("this is 13"))
            {
                System.err.println("Last is " + last);
                result = false;
            }
 		}
		catch(Exception exception)
		{
			System.out.println("[ERR : LruObjectCacheComponentTC02] " + 
                exception);
            exception.printStackTrace();
		}

		return result;
	}
}


