/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 1999 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: RecordManagerCT.java,v 1.4 2000/06/01 09:39:08 jima Exp $
 *
 * Date         Author      Changes
 * 12/23/1999   jima        Created
 */

package org.exolab.testharness.core.recman;

import java.util.Properties;
import java.util.Enumeration;
import java.util.Hashtable;
import java.io.File;
import org.exolab.jtf.CWTestCategory;
import org.exolab.jtf.CWTestCase;
import org.exolab.exceptions.CWClassConstructorException;
import org.exolab.jtf.CWVerboseStream;
import org.exolab.core.database.recman.PageManagedDatabase;
import org.exolab.core.foundation.FailedToCreateDatabaseException;
import org.exolab.core.foundation.DatabaseIOException;

/**
 * This file includes test cases for the PageManagedDatabase, which is a 
 * wrapper around the RecordManager, a paged level database by Cees de
 * Groot.
 * <p>
 * TC01 Test the opening closing and creation of the database
 * TC02 Testing the insert and retrieve methods for all primitive and object types
 * TC03 Testing the update methods for all primitive and object types
 * TC04 Testing the remove method for all the primitive and object types
 * TC05 Testing the addition, retrieval and deletion of root objects
 * TC06 Test the database locking. No two apps can open the same db
 *
 * @version     $Revision: 1.4 $ $Date: 2000/06/01 09:39:08 $
 * @author      <a href="mailto:jima@exoffice.com">Jim Alateras</a>
 * @see         org.exolab.jtf.CWTestCateogry
 * @see         org.exolab.jtf.CWTestCase
 */
public class RecordManagerCT
    extends CWTestCategory 
{
    /**
     * The constructor is responsible for constructing a test category and
     * adding the suite of test cases. It throws CWClassConstructorException
     * if it cannot construct the category
     *
     * @exception   CWClassConstructorException   
     */
    public RecordManagerCT() 
        throws CWClassConstructorException
	{
		super("recman", "PageManagedDatabase Component");
		CWTestCase tc;

		tc = new RecManComponentTC01();
		add(tc.name(), tc, true);
		tc = new RecManComponentTC02();
		add(tc.name(), tc, true);
		tc = new RecManComponentTC03();
		add(tc.name(), tc, true);
		tc = new RecManComponentTC04();
		add(tc.name(), tc, true);
		tc = new RecManComponentTC05();
		add(tc.name(), tc, true);
		tc = new RecManComponentTC06();
		add(tc.name(), tc, true);
    }
}

/**
 * This test case executes the following
 * 1. Ensure that the post create removes the database from 
 * 1. Attempt to create the database pgman.db
 * 3. Close the database
 * 4. Attempt to create it again
 * 5. Clean up by rmeoving the original database.
 */
class RecManComponentTC01 extends CWTestCase
{
    static private final String dbname__ = "pgman";

    /**
     * Calls the super class and throws CWClassConstructorException if it fails
     *
     * @exception   CWClassConstructorException
     */
    public RecManComponentTC01() 
        throws CWClassConstructorException
	{
		super("TC01", "Test database creation, opening and closing");
	}

	/**
	 * Remove the database from the current directory.
	 */
	public void preExecute()
	{
        (new File(dbname__ + ".db")).delete();
        (new File(dbname__ + ".lg")).delete();
        (new File(dbname__ + ".lock")).delete();
	}
	
    /**
     * Execute the test case described in the class header. The CWVerboseStream
     * is used to display detailed information on the test case execution
     *
     * @param           stream          this is where verbose info is streamed
     * @return          boolean         true if successful; false otherwise
     */
	public boolean run(CWVerboseStream stream)
	{
		boolean result = true;

		try
		{
            long id;

            PageManagedDatabase db = new PageManagedDatabase(dbname__);
            id = db.insertInt(10);
            stream.writeVerbose("Wrote int with the record id set to " + id);
            db.close();

            // open it and retrieve the integer
            db = new PageManagedDatabase(dbname__);
            stream.writeVerbose("About to retrieve the record " + id);
            int value = db.readInt(id);
            db.close();

            // ensure that they are equal
            if (value != 10)
            {
                result = false;
                System.err.println("Failed since the int read back is " + value
                    + " instead of  10.");
            }
 		}
		catch(Exception exception)
		{
			System.out.println("[ERR : RecManComponentTC01] " + 
                exception.getMessage());
			result = false;
		}

		return result;
	}

	/**
	 * Delete the file name after the method has completed.
	 */
	public void postExecute()
    {
        (new File(dbname__ + ".db")).delete();
        (new File(dbname__ + ".lg")).delete();
        (new File(dbname__ + ".lock")).delete();
	}
}


/**
 * This test case does the following
 * 1 Creates the database
 * 2 For every primitive type insert and instance of it and then read it back
 * 3 For object types insert and instance of it and then read it back
 * 4 Close the database
 * 5 Clean up by rmeoving the original database.
 */
class RecManComponentTC02 extends CWTestCase
{
    static private final String dbname__ = "pgman";

    /**
     * Calls the super class and throws CWClassConstructorException if it fails
     *
     * @exception   CWClassConstructorException
     */
    public RecManComponentTC02() 
        throws CWClassConstructorException
	{
		super("TC02", "Test insert and read methods on primitives and object types");
	}

	/**
	 * Remove the database from the current directory.
	 */
	public void preExecute()
	{
        (new File(dbname__ + ".db")).delete();
        (new File(dbname__ + ".lg")).delete();
        (new File(dbname__ + ".lock")).delete();
	}
	
    /**
     * Execute the test case described in the class header. The CWVerboseStream
     * is used to display detailed information on the test case execution
     *
     * @param           stream          this is where verbose info is streamed
     * @return          boolean         true if successful; false otherwise
     */
	public boolean run(CWVerboseStream stream)
	{
		boolean result = true;

		try
		{
            long id;

            PageManagedDatabase db = new PageManagedDatabase(dbname__);

            id = db.insertBoolean(true);
            stream.writeVerbose("Wrote the bool_value true");
            boolean bool_value = db.readBoolean(id);
            if (bool_value != true)
            {
                System.err.println("Read back bool_value " + bool_value);
                result = false;
            }
            
            id = db.insertShort((short)7);
            stream.writeVerbose("Wrote the short_value 7");
            short short_value = db.readShort(id);
            if (short_value != 7)
            {
                System.err.println("Read back bool_value " + short_value);
                result = false;
            }

            id = db.insertChar('A');
            stream.writeVerbose("Wrote the char_value A");
            char char_value = db.readChar(id);
            if (char_value != 'A')
            {
                System.err.println("Read back char_value " + char_value);
                result = false;
            }

            id = db.insertByte((byte)-1);
            stream.writeVerbose("Wrote the byte_value 0xFF");
            byte byte_value = (byte)db.readByte(id);
            if (byte_value != -1)
            {
                System.err.println("Read back byte_value " + byte_value);
                result = false;
            }

            id = db.insertInt(7777);
            stream.writeVerbose("Wrote the int_value 7777");
            int int_value = db.readInt(id);
            if (int_value != 7777)
            {
                System.err.println("Read back int_value " + int_value);
                result = false;
            }

            id = db.insertLong(77777777);
            stream.writeVerbose("Wrote the long_value 77777777");
            long long_value = db.readLong(id);
            if (long_value != 77777777)
            {
                System.err.println("Read back long_value " + long_value);
                result = false;
            }

            id = db.insertFloat((float)7777.77);
            stream.writeVerbose("Wrote the float_value 7777.77");
            String float_value = (new Float(db.readFloat(id))).toString();
            if (!float_value.equals("7777.77"))
            {
                System.err.println("Read back float_value " + float_value);
                result = false;
            }

            id = db.insertDouble(7777.7777);
            stream.writeVerbose("Wrote the double_value 7777.7777");
            String double_value = (new Double(db.readDouble(id))).toString();
            if (!double_value.equals("7777.7777"))
            {
                System.err.println("Read back double_value " + double_value);
                result = false;
            }

            id = db.insertString("record manager rules!!");
            stream.writeVerbose("Wrote the obj_value -record manager rules!!-");
            String obj_value = db.readString(id);
            if (!obj_value.equals("record manager rules!!"))
            {
                System.err.println("Read back obj_value " + obj_value);
                result = false;
            }
            db.close();
 		}
		catch(Exception exception)
		{
			System.out.println("[ERR : RecManComponentTC02] " + 
                exception.getMessage());
            exception.printStackTrace();
			result = false;
		}

		return result;
	}

	/**
	 * Delete the file name after the method has completed.
	 */
	public void postExecute()
    {
        (new File(dbname__ + ".db")).delete();
        (new File(dbname__ + ".lg")).delete();
        (new File(dbname__ + ".lock")).delete();
	}
}


/**
 * This test case does the following
 * 1 Creates the database
 * 2 For every primitive and object type do the following
 * 3 Insert an instance of t hat type
 * 4 Retrieve the type, modify it and update it
 * 5 Clean up by rmeoving the original database.
 */
class RecManComponentTC03 extends CWTestCase
{
    static private final String dbname__ = "pgman";

    /**
     * Calls the super class and throws CWClassConstructorException if it fails
     *
     * @exception   CWClassConstructorException
     */
    public RecManComponentTC03() 
        throws CWClassConstructorException
	{
		super("TC03", "Test the update methods on primitive and object types");
	}

	/**
	 * Remove the database from the current directory.
	 */
	public void preExecute()
	{
        (new File(dbname__ + ".db")).delete();
        (new File(dbname__ + ".lg")).delete();
        (new File(dbname__ + ".lock")).delete();
	}
	
    /**
     * Execute the test case described in the class header. The CWVerboseStream
     * is used to display detailed information on the test case execution
     *
     * @param           stream          this is where verbose info is streamed
     * @return          boolean         true if successful; false otherwise
     */
	public boolean run(CWVerboseStream stream)
	{
		boolean result = true;

		try
		{
            long id;

            PageManagedDatabase db = new PageManagedDatabase(dbname__);

            id = db.insertBoolean(true);
            stream.writeVerbose("Wrote the bool_value true");
            boolean bool_value = db.readBoolean(id);
            if (bool_value != true)
            {
                System.err.println("Read back bool_value " + bool_value);
                result = false;
            }
            else
            {
                db.updateBoolean(id, false);
                stream.writeVerbose("Updated the bool_value to false");
                boolean boolvalue = db.readBoolean(id);
                if (boolvalue != false)
                {
                    System.err.println("Read back bool_value " + boolvalue);
                    result = false;
                }
            }
            
            id = db.insertShort((short)7);
            stream.writeVerbose("Wrote the short_value 7");
            short short_value = db.readShort(id);
            if (short_value != 7)
            {
                System.err.println("Read back bool_value " + short_value);
                result = false;
            }
            else
            {
                id = db.updateShort(id, (short)8);
                stream.writeVerbose("Wrote the short_value 8");
                short shortvalue = db.readShort(id);
                if (shortvalue != 8)
                {
                    System.err.println("Read back bool_value " + shortvalue);
                    result = false;
                }
            }

            id = db.insertChar('A');
            stream.writeVerbose("Wrote the char_value A");
            char char_value = db.readChar(id);
            if (char_value != 'A')
            {
                System.err.println("Read back char_value " + char_value);
                result = false;
            }
            else
            {
                id = db.updateChar(id, 'B');
                stream.writeVerbose("Wrote the char_value B");
                char charvalue = db.readChar(id);
                if (charvalue != 'B')
                {
                    System.err.println("Read back char_value " + charvalue);
                    result = false;
                }
            }

            id = db.insertByte((byte)-1);
            stream.writeVerbose("Wrote the byte_value 0xFF");
            byte byte_value = (byte)db.readByte(id);
            if (byte_value != -1)
            {
                System.err.println("Read back byte_value " + byte_value);
                result = false;
            }
            else
            {
                id = db.updateByte(id, (byte)1);
                stream.writeVerbose("Wrote the byte_value 1");
                byte bytevalue = (byte)db.readByte(id);
                if (bytevalue != 1)
                {
                    System.err.println("Read back byte_value " + bytevalue);
                    result = false;
                }
            }

            id = db.insertInt(7777);
            stream.writeVerbose("Wrote the int_value 7777");
            int int_value = db.readInt(id);
            if (int_value != 7777)
            {
                System.err.println("Read back int_value " + int_value);
                result = false;
            }
            else
            {
                id = db.updateInt(id, 8888);
                stream.writeVerbose("Wrote the int_value 8888");
                int intvalue = db.readInt(id);
                if (intvalue != 8888)
                {
                    System.err.println("Read back int_value " + intvalue);
                    result = false;
                }
            }

            id = db.insertLong(77777777);
            stream.writeVerbose("Wrote the long_value 77777777");
            long long_value = db.readLong(id);
            if (long_value != 77777777)
            {
                System.err.println("Read back long_value " + long_value);
                result = false;
            }
            else
            {
                id = db.updateLong(id, 77777778);
                stream.writeVerbose("Wrote the long_value 77777778");
                long longvalue = db.readLong(id);
                if (longvalue != 77777778)
                {
                    System.err.println("Read back long_value " + longvalue);
                    result = false;
                }
            }

            id = db.insertFloat((float)7777.77);
            stream.writeVerbose("Wrote the float_value 7777.77");
            String float_value = (new Float(db.readFloat(id))).toString();
            if (!float_value.equals("7777.77"))
            {
                System.err.println("Read back float_value " + float_value);
                result = false;
            }
            else
            {
                id = db.updateFloat(id, (float)7777.88);
                stream.writeVerbose("Wrote the float_value 7777.88");
                String floatvalue = (new Float(db.readFloat(id))).toString();
                if (!floatvalue.equals("7777.88"))
                {
                    System.err.println("Read back float_value " + floatvalue);
                    result = false;
                }
            }

            id = db.insertDouble(7777.7777);
            stream.writeVerbose("Wrote the double_value 7777.7777");
            String double_value = (new Double(db.readDouble(id))).toString();
            if (!double_value.equals("7777.7777"))
            {
                System.err.println("Read back double_value " + double_value);
                result = false;
            }
            else
            {
                id = db.updateDouble(id, 7777.8888);
                stream.writeVerbose("Wrote the double_value 7777.8888");
                String doublevalue = (new Double(db.readDouble(id))).toString();
                if (!doublevalue.equals("7777.8888"))
                {
                    System.err.println("Read back double_value " + doublevalue);
                    result = false;
                }
            }

            id = db.insertString("record manager rules!!");
            stream.writeVerbose("Wrote the obj_value -record manager rules!!-");
            String obj_value = db.readString(id);
            if (!obj_value.equals("record manager rules!!"))
            {
                System.err.println("Read back obj_value " + obj_value);
                result = false;
            }
            else
            {
                id = db.insertString("record manager does updates!!");
                stream.writeVerbose("Wrote the obj_value -record manager does updates!!-");
                String objvalue = db.readString(id);
                if (!objvalue.equals("record manager does updates!!"))
                {
                    System.err.println("Read back obj_value " + objvalue);
                    result = false;
                }
            }
            db.close();
 		}
		catch(Exception exception)
		{
			System.out.println("[ERR : RecManComponentTC03] " + 
                exception.getMessage());
            exception.printStackTrace();
			result = false;
		}

		return result;
	}

	/**
	 * Delete the file name after the method has completed.
	 */
	public void postExecute()
    {
        (new File(dbname__ + ".db")).delete();
        (new File(dbname__ + ".lg")).delete();
        (new File(dbname__ + ".lock")).delete();
	}
}


/**
 * This test case does the following
 * 1 Creates the database
 * 2 For every primitive type insert an instance of it and then delete it 
 *   using the id
 * 3 Close the database
 * 4 Clean up by rmeoving the original database.
 */
class RecManComponentTC04 extends CWTestCase
{
    static private final String dbname__ = "pgman";

    /**
     * Calls the super class and throws CWClassConstructorException if it fails
     *
     * @exception   CWClassConstructorException
     */
    public RecManComponentTC04() 
        throws CWClassConstructorException
	{
		super("TC04", "Test delete method by identifier");
	}

	/**
	 * Remove the database from the current directory.
	 */
	public void preExecute()
	{
        (new File(dbname__ + ".db")).delete();
        (new File(dbname__ + ".lg")).delete();
        (new File(dbname__ + ".lock")).delete();
	}
	
    /**
     * Execute the test case described in the class header. The CWVerboseStream
     * is used to display detailed information on the test case execution
     *
     * @param           stream          this is where verbose info is streamed
     * @return          boolean         true if successful; false otherwise
     */
	public boolean run(CWVerboseStream stream)
	{
		boolean result = true;

		try
		{
            long id;

            PageManagedDatabase db = new PageManagedDatabase(dbname__);

            id = db.insertBoolean(true);
            stream.writeVerbose("Wrote the bool_value true");
            boolean bool_value = db.readBoolean(id);
            if (bool_value != true)
            {
                System.err.println("Read back bool_value " + bool_value);
                result = false;
            }
            else
            {
                db.deleteObject(id);
                // try and retrieve it. If we can't then assume that it
                // has been released
                try
                {
                    boolean boolvalue = db.readBoolean(id);
                }
                catch(DatabaseIOException exception)
                {
                    // should get here so it is successful
                    result = true;
                }
            }
            db.close();
 		}
		catch(Exception exception)
		{
			System.out.println("[ERR : RecManComponentTC04] " + 
                exception.getMessage());
            exception.printStackTrace();
			result = false;
		}

		return result;
	}

	/**
	 * Delete the file name after the method has completed.
	 */
	public void postExecute()
    {
        (new File(dbname__ + ".db")).delete();
        (new File(dbname__ + ".lg")).delete();
        (new File(dbname__ + ".lock")).delete();
	}
}


/**
 * This test case does the following
 * 1 Creates the database
 * 2 Creates 10 roots, named root1->root10 and associates a Hashtable with
 *   5 random integer values to each table.
 * 3 Restores each of the roots and ensures that they are correct
 * 4 Delete the roots
 * 5 Close the database
 * 6 Clean up by rmeoving the original database.
 */
class RecManComponentTC05 extends CWTestCase
{
    static private final String dbname__ = "pgman";

    /**
     * Calls the super class and throws CWClassConstructorException if it fails
     *
     * @exception   CWClassConstructorException
     */
    public RecManComponentTC05() 
        throws CWClassConstructorException
	{
		super("TC05", "Test create, retrieval and removal of root objects");
	}

	/**
	 * Remove the database from the current directory.
	 */
	public void preExecute()
	{
        (new File(dbname__ + ".db")).delete();
        (new File(dbname__ + ".lg")).delete();
        (new File(dbname__ + ".lock")).delete();
	}
	
    /**
     * Execute the test case described in the class header. The CWVerboseStream
     * is used to display detailed information on the test case execution
     *
     * @param           stream          this is where verbose info is streamed
     * @return          boolean         true if successful; false otherwise
     */
	public boolean run(CWVerboseStream stream)
	{
		boolean result = true;

		try
		{
            long id;

            PageManagedDatabase db = new PageManagedDatabase(dbname__);

            //  Create the 10 root nodes 
            for (int index = 0; index < 10; index++)
            {
                String root_name = "root_index-" + index;
                Hashtable table = new Hashtable();
                for (int count = 0; count < 5; count++)
                {
                    Integer int_val = new Integer(count);
                    table.put(int_val, int_val);
                }
                db.setRootObject(root_name, table);
            }

            // retrieve the nodes and check them
            Enumeration nodes = db.getRootNames();
            while (nodes.hasMoreElements())
            {
                String name = (String)nodes.nextElement();
                Hashtable temp_table = (Hashtable)db.getRootObject(name);
                if (temp_table != null)
                {
                    for (int count = 0; count < 5; count++)
                    {
                        Integer int_val = new Integer(count);
                        if (!(temp_table.get(int_val) instanceof Integer))
                        {
                            result = false;
                        }
                    }
                }
                else
                {
                    System.err.println("Failed to retrieve the root " + name);
                    result = false;
                }
            }

            // unbind the nodes
            for (int index = 0; index < 10; index++)
            {
                String root_name = "root_index-" + index;
                db.unbind(root_name);
            }

            // retrieve the nodes again
            for (int index = 0; index < 10; index++)
            {
                String root_name = "root_index-" + index;
                if (db.getRootObject(root_name) != null)
                {
                    result = false;
                }
            }
            db.close();
 		}
		catch(Exception exception)
		{
			System.out.println("[ERR : RecManComponentTC05] " + 
                exception.getMessage());
            exception.printStackTrace();
			result = false;
		}

		return result;
	}

	/**
	 * Delete the file name after the method has completed.
	 */
	public void postExecute()
    {
        (new File(dbname__ + ".db")).delete();
        (new File(dbname__ + ".lg")).delete();
        (new File(dbname__ + ".lock")).delete();
	}
}


/**
 * This test case does the following
 * 1 Open or create the database
 * 2 Open the same database another time. Which should fail because the db
 *   is already opened
 * 3. Close the database
 * 4. Open the database which should now suceed
 * 5 Clean up by rmeoving the original database.
 */
class RecManComponentTC06 extends CWTestCase
{
    static private final String dbname__ = "pgman";

    /**
     * Calls the super class and throws CWClassConstructorException if it fails
     *
     * @exception   CWClassConstructorException
     */
    public RecManComponentTC06() 
        throws CWClassConstructorException
	{
		super("TC06", "Test database locking at the file system level");
	}

	/**
	 * Remove the database from the current directory.
	 */
	public void preExecute()
	{
        (new File(dbname__ + ".db")).delete();
        (new File(dbname__ + ".lg")).delete();
        (new File(dbname__ + ".lock")).delete();
	}
	
    /**
     * Execute the test case described in the class header. The CWVerboseStream
     * is used to display detailed information on the test case execution
     *
     * @param           stream          this is where verbose info is streamed
     * @return          boolean         true if successful; false otherwise
     */
	public boolean run(CWVerboseStream stream)
	{
		boolean result = true;

		try
		{
            long id;

            PageManagedDatabase db = new PageManagedDatabase(dbname__);

            try
            {
                // attempt to open it again
                new PageManagedDatabase(dbname__);

                // should not normally get here unless the locking has not
                // worked
                stream.writeVerbose("Was allow to open the database a second time");
                result = false;
            }
            catch (FailedToCreateDatabaseException exception)
            {
                // this is cool since the database should be locked
                stream.writeVerbose("Can't open the db again b/c it is locked");
            }
            catch (Exception exception)
            {
                // this is some other error set the result to false
                stream.writeVerbose("Error created the db second time " + 
                    exception);
                result = false;
            }

            // close the data
            db.close();
            db = null;

            // now try to create the database again
            try
            {
                // attempt to open it again. It should be able to acquire the
                // lock and continue
                db = new PageManagedDatabase(dbname__);
                db.close();
            }
            catch (Exception exception)
            {
                stream.writeVerbose("Error created the db  " + 
                    exception);
                result = false;
            }
 		}
		catch(Exception exception)
		{
			System.out.println("[ERR : RecManComponentTC06] " + 
                exception.getMessage());
            exception.printStackTrace();
			result = false;
		}

		return result;
	}

	/**
	 * Delete the file name after the method has completed.
	 */
	public void postExecute()
    {
        (new File(dbname__ + ".db")).delete();
        (new File(dbname__ + ".lg")).delete();
        (new File(dbname__ + ".lock")).delete();
	}
}



