/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 1999 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: SessionManagerCT.java,v 1.2 2000/03/13 06:27:36 jima Exp $
 *
 * Date         Author      Changes
 * 03/08/1999   jima        Created
 */

package org.exolab.testharness.core.recman;

import java.util.Properties;
import java.util.Enumeration;
import java.util.Hashtable;
import java.io.File;
import org.exolab.jtf.CWTestCategory;
import org.exolab.jtf.CWTestCase;
import org.exolab.jtf.CWVerboseStream;
import org.exolab.exceptions.CWClassConstructorException;
import org.exolab.core.foundation.SessionIfc;
import org.exolab.core.database.recman.PMDSessionManager;
import org.exolab.core.database.recman.PageManagedDatabase;
import org.exolab.core.foundation.FailedToDestroySessionException;
import org.exolab.core.foundation.SessionIfc;


/**
 * This file includes test cases for the PMDSessionManager which provides
 * session management capabilties to the PageManagedDatabase. The PMDSession-
 * Manager is responsible for co-ordinating access to the database through 
 * multiple threads. 
 * <p>
 * TC01 Test the initialisation and access to the SessionManager
 * TC02 Test creation and destruction of multiple sessions
 
 *
 * @version     $Revision: 1.2 $ $Date: 2000/03/13 06:27:36 $
 * @author      <a href="mailto:jima@exoffice.com">Jim Alateras</a>
 * @see         org.exolab.jtf.CWTestCateogry
 * @see         org.exolab.jtf.CWTestCase
 */
public class SessionManagerCT
    extends CWTestCategory 
{
    /**
     * The constructor is responsible for constructing a test category and
     * adding the suite of test cases. It throws CWClassConstructorException
     * if it cannot construct the category
     *
     * @exception   CWClassConstructorException   
     */
    public SessionManagerCT() 
        throws CWClassConstructorException
	{
		super("sessionmgr", "PMD Session Manager Component");
		CWTestCase tc;

		tc = new SessionMgrComponentTC01();
		add(tc.name(), tc, true);
		tc = new SessionMgrComponentTC02();
		add(tc.name(), tc, true);
    }
}

/**
 * This test case executes the following
 * 1. Create an new database
 * 1. Attach it to the session manager
 * 3. get a session for this thread multiple times
 * 4. destroy the session multiple times
 * 5. Clean up by rmeoving the original database.
 */
class SessionMgrComponentTC01 extends CWTestCase
{
    static private final String dbname__ = "pgman";

    /**
     * Calls the super class and throws CWClassConstructorException if it fails
     *
     * @exception   CWClassConstructorException
     */
    public SessionMgrComponentTC01() 
        throws CWClassConstructorException
	{
		super("TC01", "Test the init and access to the SessionManager");
	}

	/**
	 * Remove the database from the current directory.
	 */
	public void preExecute()
	{
        (new File(dbname__ + ".db")).delete();
        (new File(dbname__ + ".lg")).delete();
	}
	
    /**
     * Execute the test case described in the class header. The CWVerboseStream
     * is used to display detailed information on the test case execution
     *
     * @param           stream          this is where verbose info is streamed
     * @return          boolean         true if successful; false otherwise
     */
	public boolean run(CWVerboseStream stream)
	{
		boolean result = false;

		try
		{
            PageManagedDatabase db = new PageManagedDatabase(dbname__);
            PMDSessionManager.init(db);

            SessionIfc session = PMDSessionManager.instance().getSession();
            if (session == null)
            {
                stream.writeVerbose("Couldn't create session for thread");
            }
            
            SessionIfc session1= PMDSessionManager.instance().getSession();
            if (session1 != session)
            {
                stream.writeVerbose("Different sessions for same thread");
            }

            PMDSessionManager.instance().destroySession();
            stream.writeVerbose("Destroy called for the first object");

            // if we try and delete it again we should get an exception
            try
            {
                PMDSessionManager.instance().destroySession();
                stream.writeVerbose("Called destroy twice for the same obj");
            }
            catch (FailedToDestroySessionException exception)
            {
                result = true;
            }

            db.close();
 		}
		catch(Exception exception)
		{
			System.out.println("[ERR : SessionMgrComponentTC01] " + 
                exception.getMessage());
			result = false;
		}

		return result;
	}

	/**
	 * Delete the file name after the method has completed.
	 */
	public void postExecute()
    {
        (new File(dbname__ + ".db")).delete();
        (new File(dbname__ + ".lg")).delete();
	}
}

/**
 * This test case executes the following
 * 1. Create an new database
 * 1. Attach it to the session manager
 * 3. get a session for the two different threads
 * 4  compare and ensure that they are different
 */
class SessionMgrComponentTC02 extends CWTestCase
{
    static private final String dbname__ = "pgman";

    /**
     * Calls the super class and throws CWClassConstructorException if it fails
     *
     * @exception   CWClassConstructorException
     */
    public SessionMgrComponentTC02() 
        throws CWClassConstructorException
	{
		super("TC02", "Test the creation and destruction of sessions " +
              " in different threads");
	}

	/**
	 * Remove the database from the current directory.
	 */
	public void preExecute()
	{
        (new File(dbname__ + ".db")).delete();
        (new File(dbname__ + ".lg")).delete();
	}
	
    /**
     * Execute the test case described in the class header. The CWVerboseStream
     * is used to display detailed information on the test case execution
     *
     * @param           stream          this is where verbose info is streamed
     * @return          boolean         true if successful; false otherwise
     */
	public boolean run(CWVerboseStream stream)
	{
		boolean result = true;

		try
		{
            PageManagedDatabase db = new PageManagedDatabase(dbname__);
            PMDSessionManager.init(db);

            Thread threads[] = new Thread[2];
            SessionClientA client1 = new SessionClientA(stream);
            threads[0] = new Thread(client1);

            SessionClientA client2 = new SessionClientA(stream);
            threads[1] = new Thread(client2);

            // start both the threads
            threads[0].start();
            threads[1].start();

            // wait for both threads to complete
            for (int index = 0; index < 2; index++)
            {
                threads[index].join();
            }

            // check that the two sessions are not identical
            stream.writeVerbose("About to compare the two sessions");
            if (client1.getSession() == client2.getSession())
            {
                stream.writeVerbose("Sessions from different threads are identical");
                result = false;
            }

            // destroy both the sessions
            db.close();
 		}
		catch(Exception exception)
		{
			System.out.println("[ERR : SessionMgrComponentTC02] " + 
                exception.getMessage());
			result = false;
		}

		return result;
	}

	/**
	 * Delete the file name after the method has completed.
	 */
	public void postExecute()
    {
        (new File(dbname__ + ".db")).delete();
        (new File(dbname__ + ".lg")).delete();
	}
}


/** 
 * This class, when run will connect to the session manager, and 
 * get a session
 */
class SessionClientA implements Runnable
{
    /**
     * Instantiate an instance of this class with the specified stream
     *
     * @param       CWVerboseStream
     */
    public SessionClientA(CWVerboseStream stream)
    {
        stream_ = stream;
    }

    /**
     * Get access to the session manager and create a session for this
     * thread
     */
    public void run()
    {
        try
        {
            session_ = PMDSessionManager.instance().getSession();
            stream_.writeVerbose("Created the session " + session_.toString());
            Thread.currentThread().sleep(5000);
            PMDSessionManager.instance().destroySession();
            stream_.writeVerbose("Destroyed the session " + session_.toString());
        }
        catch (Exception exception)
        {
            // error trying to create the sessions
        }
    }

    /**
     * Return the session
     *
     * @return      SessionIfc
     */
    public SessionIfc getSession()
    {
        return session_;
    }


    private SessionIfc session_ = null;
    private CWVerboseStream stream_ = null;
}



