/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2000 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: ThreadPoolCT.java,v 1.5 2002/01/29 04:51:05 tima Exp $
 *
 * Date         Author  Changes
 * $Date	    jimm    Created
 */


package org.exolab.testharness.core.threadPool;

import org.exolab.jtf.CWTestCategory;
import org.exolab.jtf.CWTestCase;
import org.exolab.exceptions.CWClassConstructorException;
import org.exolab.jtf.CWVerboseStream;
import org.exolab.core.threadPool.ThreadPool;
import org.exolab.core.logger.LoggerIfc;
import org.exolab.core.logger.LoggerFactory;

/**
 * This is the testharness for the ThreadPool and FifoQueue.
 * The following test cases are executed
 * <p>
 * TC01 Creates 5 runnable objects and a ThreadPool of three threads.
 * the 5 objects are run one after the other, but 4 and 5 will block 
 * until a thread becomes available for them to run on.
 *
 * @version     $Revision: 1.5 $ $Date: 2002/01/29 04:51:05 $
 * @author      <a href="mailto:mourikis@exolab.org">Jim Mourikis</a>
 * @see         org.exolab.core.ThreadPool
 * @see         org.exolab.jtf.CWTestCateogry
 * @see         org.exolab.jtf.CWTestCase
 */
public class ThreadPoolCT extends CWTestCategory {

    /**
     * Construct all the test cases
     *
     * @throws   CWClassConstructorException If category cannot be created.
     */
    public ThreadPoolCT() throws CWClassConstructorException {
        super("ThreadPool", "ThreadPool tester");
        CWTestCase tc = new ThreadTest();
        add(tc.name(), tc, true);
        tc = new QueueTest();
        add(tc.name(), tc, true);
    }

} //-- ThreadPoolCT


/**
 * Test the ThreadPool, by creating some runnable objects and getting a thread
 * from the pool to run them.
 *
 */
class ThreadTest extends CWTestCase {

    // The class level logger for output.
    static LoggerIfc _logger;


    /**
     * Calls the super class and throws CWClassConstructorException if it fails
     * Creates the logger.
     *
     * @throws CWClassConstructorException
     */
    public ThreadTest() throws CWClassConstructorException {
        super("TC01", "Test Thread pool");
        _logger = LoggerFactory.getLogger();
    }

    /**
     * Execute the test case described in the class header. The CWVerboseStream
     * is used to display detailed information on the test case execution
     *
     * @param           stream          this is where verbose info is streamed
     * @return          boolean         true if successful; false otherwise
     */
    public boolean run(CWVerboseStream stream) {
        boolean result = true;

        try {
            ThreadPool pool = new ThreadPool(3);

            Runnable ra = RunnableHelper.makeRunnable("Worker A", 3000);
            pool.execute(ra);

            Runnable rb = RunnableHelper.makeRunnable("Worker B", 1000);
            pool.execute(rb);

            Runnable rc = RunnableHelper.makeRunnable("Worker C", 2000);
            pool.execute(rc);

            Runnable rd = RunnableHelper.makeRunnable("Worker D", 60000);
            pool.execute(rd);

            Runnable re = RunnableHelper.makeRunnable("Worker E", 1000);
            pool.execute(re);

            pool.stopRequestIdleWorkers();
            Thread.sleep(2000);
            pool.stopRequestIdleWorkers();
            Thread.sleep(5000);
            pool.stopRequestAllWorkers();
        } catch(Exception exception) {
            _logger.logError(exception.getMessage(), exception);
            result = false;
        }
		
        return result;
    }    
} //-- ThreadTest

/**
 * Test the ThreadPool, by creating some runnable objects and queueing them
 * for execution
 */
class QueueTest extends CWTestCase {

    // The class level logger for output.
    static LoggerIfc _logger;


    /**
     * Calls the super class and throws CWClassConstructorException if it fails
     * Creates the logger.
     *
     * @throws CWClassConstructorException
     */
    public QueueTest() throws CWClassConstructorException {
        super("TC02", "Test Thread pool queueing");
        _logger = LoggerFactory.getLogger();
    }

    /**
     * Execute the test case described in the class header. The CWVerboseStream
     * is used to display detailed information on the test case execution
     *
     * @param           stream          this is where verbose info is streamed
     * @return          boolean         true if successful; false otherwise
     */
    public boolean run(CWVerboseStream stream) {
        boolean result = true;

        try {
            ThreadPool pool = new ThreadPool(3);

            Runnable ra = RunnableHelper.makeRunnable("Worker A", 3000);
            pool.queue(ra);

            Runnable rb = RunnableHelper.makeRunnable("Worker B", 1000);
            pool.queue(rb);

            Runnable rc = RunnableHelper.makeRunnable("Worker C", 2000);
            pool.queue(rc);

            Runnable rd = RunnableHelper.makeRunnable("Worker D", 60000);
            pool.queue(rd);

            Runnable re = RunnableHelper.makeRunnable("Worker E", 1000);
            pool.queue(re);

            pool.stopRequestIdleWorkers();
            Thread.sleep(2000);
            pool.stopRequestIdleWorkers();
            Thread.sleep(5000);
            pool.stopRequestAllWorkers();
        } catch(Exception exception) {
            _logger.logError(exception.getMessage(), exception);
            result = false;
        }
		
        return result;
    }    
} //-- ThreadTest

class RunnableHelper {

    /**
     * A simple method to return a runnable object that can be used
     * to run on a thread from the pool
     *
     * @param name		Give the object a name for tracking purposes.
     * @param firstDelay When started, sleep for this long
     * @return Runnable A runnable object.
     *
     */
    public static Runnable makeRunnable(final String name, 
                                        final long firstDelay) {
        return new Runnable() {
            public void run() {
                LoggerIfc logger = LoggerFactory.getLogger();
                try {
                    logger.logDebug(name +": starting up");
                    Thread.sleep(firstDelay);
                    logger.logDebug(name + ": doing some stuff");
                    Thread.sleep(2000);
                    logger.logDebug(name + ": leaving");
                } catch (InterruptedException exception) {
                    logger.logDebug(name + ": got interrupted!");
                } catch (Exception exception) {
                    logger.logError(exception.getMessage(), exception);
                }
            }

            public String toString() {
                return name;
            }
        };
    }

} //-- RunnableHelper
