/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2002-2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: CoreTestSuite.java,v 1.2 2003/06/09 06:28:42 tanderson Exp $
 */

package org.exolab.core.test;
                                             
import java.util.Enumeration;
import java.io.PrintStream;

import junit.framework.AssertionFailedError;
import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestListener;
import junit.framework.TestResult;
import junit.framework.TestSuite;
import junit.textui.TestRunner;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.GnuParser;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.Parser;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.log4j.xml.DOMConfigurator;
import org.apache.oro.text.regex.Pattern;
import org.apache.oro.text.regex.MalformedPatternException;
import org.apache.oro.text.regex.Perl5Compiler;
import org.apache.oro.text.regex.Perl5Matcher;

import org.exolab.core.test.messenger.MessengerTestSuite;
import org.exolab.core.test.threadPool.ThreadPoolTest;


/**
 * The test harness extends the JUnit framework to support extended 
 * facilities, such as running a portion of the test harness or displaying
 * all the test cases in the test harness. It also uses an extension to 
 * TestResult, which captures the results as an XML document, which can 
 * then be translated to a different format through XSLT.
 *
 * @author <a href="mailto:tima@intalio.com">Tim Anderson</a>
 * @author <a href="mailto:jima@intalio.com">Jim Alateras</a>
 * @version $Revision: 1.2 $ $Date: 2003/06/09 06:28:42 $
 */
final public class CoreTestSuite implements TestListener {

    /**
     * Used to count the number of test cases
     */
    private static int _index = 0;

    /** 
     * Caches an instance of a test runner that is used by this application
     */
    private TestRunner _runner = null;

    /**
     * Holds the test suite, which maintains a list of test cases to execute
     */
    private TestSuite _suite = null;

    /**
     * The default name name of the log4j properties file
     */
    private static String _logPropertyFile = "log4j.xml";


    /**
     * The following options are supported by this application
     * <p>
     * -display     
     *      display all the test cases
     * -execute <expression>
     *      if expression is missing it will execute all the tests in the
     *      test suite, otherwise it will only execute matching test cases.
     */
    public static void main(String[] args) throws Exception {
        // process the command line
        Options options = new Options();
        options.addOption("display", false, "display all the test cases ");
        options.addOption("execute", false, "execute test cases");
        options.addOption("help", false, "displays usage");
        Parser parser = new GnuParser();

        CommandLine commands = parser.parse(options, args);

        boolean help = commands.hasOption("help");
        boolean display = commands.hasOption("display");
        boolean execute = commands.hasOption("execute");
        if (help) {
            usage();
        } else if (display || execute) {
            // the correct combination of command lines were specified so
            // we can go ahead and process it.
            CoreTestSuite harness = new CoreTestSuite();
            if (display) {
                _index = 0;
                harness.displayTests(harness.getTestSuite());
            }
            if (execute) {
                try {          
                    // initialise the logger
                    DOMConfigurator.configure(_logPropertyFile);
                } catch (Exception exception) {
                    exception.printStackTrace();
                }

                // determine whether execute was specified as a switch
                // or a command line option. If it was a switch then we 
                // need to execute the complete test suite. If it was a 
                // parameter then we treat the value as a regular 
                // expression, find all matching test cases and execute 
                // them
                TestSuite suite = null;
                String pattern = commands.getOptionValue("execute");
                if (pattern == null) {
                    suite = harness.getTestSuite();
                } else {
                    try {
                        Perl5Compiler compiler = new Perl5Compiler();
                        Pattern expression = compiler.compile(pattern);
                        TestSuite matching = new TestSuite();
                        harness.getMatchingTests(harness.getTestSuite(), 
                                                 matching, expression);
                        
                        // assign them
                        suite = matching;
                    } catch (MalformedPatternException exception) {
                        throw new RuntimeException(
                            "Bad regular expression: " + 
                            exception.toString());
                    }       
                }
    
                // execute the test suite and print the results
                TestResult results = harness.execute(suite);
                harness.getTestRunner().print(results);
            }
        } else {
            usage();
        }
    }
     
    // implementation of TestListener.startTest
    public void startTest(Test test) {
        if (test instanceof TestCase) {
            System.err.print("\nexecuting " + test.getClass().getName() + "." 
                             + ((TestCase)test).getName());
        } else if (test instanceof TestSuite) {
            System.err.print("\nin test suite " + ((TestSuite)test).getName());
        }
    }
    
    // implementation of TestListenr.addError
    public void addError(Test test, Throwable throwable) {
        System.err.print("...[error]");
    }

    // implementation of TestListener.addFailure
    public void addFailure(Test test, AssertionFailedError throwable) {
        System.err.print("...[failure]");
    }

    // implementation of TestListener.endTest
    public void endTest(Test test) {
        System.err.print("...[passed]");
    }

    /**
     * Return the compiled test suite. 
     *
     * @return the test suite
     */
    TestSuite getTestSuite() {
        if (_suite == null) {
            _suite = new TestSuite("exolabcore");
            _suite.addTest(ThreadPoolTest.suite());
            _suite.addTest(MessengerTestSuite.suite());
        }

        return _suite;
    }

    /**
     * Execute all the tests in the specified test suite. This method will
     * create and return a TestResult object, which maintains the state of 
     * all tests executed.
     *
     * @param suite - the test suite to execute
     * @return TestResult - the result set
     */
    TestResult execute(TestSuite suite) {
    
        TestResult result = new TestResult();  
        result.addListener(this);

        // execute the test cases and return the result set to
        // the calling application
        suite.run(result);

        return result;
    }

    /**
     * This is a recursive function that selects all tests matching 
     * the specified expression and places them in the supplied results.
     * <p>
     * The results parameter cannot be non-null otherwise an 
     * {@link IllegalArgumentException} exception is thrown. Similar
     * for expression and suite.
     *
     * @param suite - the test suite to inspect
     * @param results - matching test cases are stored here.
     * @param expression - a regular expresion
     */
    void getMatchingTests(TestSuite suite, TestSuite results, 
                          Pattern expression) {
        if ((suite == null) || (results == null) || expression == null) {
            throw new IllegalArgumentException(
                "getMatchingTests : cannot have null input param");
        }
        Perl5Matcher matcher = new Perl5Matcher();

        Enumeration tests = suite.tests();
        while (tests.hasMoreElements()) {
            // for every test case. If the expression matches the
            // name of the test case it is added to the result
            // set.
            Test test = (Test)tests.nextElement();
            if (test instanceof TestCase) {
                String name = test.getClass().getName() + "." +
                    ((TestCase)test).getName();
                if (matcher.matches(name, expression)) {
                    results.addTest(test);
                }
            } else if (test instanceof TestSuite) {
                // make a recursice call
                getMatchingTests((TestSuite)test, results, expression);
            }
        }
    }

    /**
     * Display the list of test cases registered with the specified 
     * test suite on to the console
     *
     * @param suite - the test suite
     */
    void displayTests(TestSuite suite) {
        Enumeration tests = suite.tests();
        while (tests.hasMoreElements()) {
            Test test = (Test)tests.nextElement();
            if (test instanceof TestCase) {
                System.err.println(++_index + ". " + 
                                   test.getClass().getName() + 
                                   "." + ((TestCase)test).getName());
            } else if (test instanceof TestSuite) {
                displayTests((TestSuite)test);
            } else { 
                System.err.println(++_index + ". " + 
                                   test.getClass().getName());
            }
        }
    }

    /**
     * Retrieve an instance of the {@link TestRunner}. If one does not exist 
     * then create it
     *
     * @return TestRunner
     */
    TestRunner getTestRunner() {
        if (_runner == null) {
            _runner = new TestRunner();
        }

        return _runner;
    }

    /**
     * Displays usage information on the console
     */
    public static void usage() {
        PrintStream out = System.out;
                                                                   
        out.println("Usage: java [jvm options] " + CoreTestSuite.class + 
                    " [args...]");
        out.println("\nwhere args are:");
        out.println("  -execute <regex>   execute the test cases. If a " +
                    "regular expression");
        out.println("                     is not specified all test cases "+
                    "are executed.");
        out.println("                     If a regular expression is " +
                    "specified, only matching");
        out.println("                     test cases are executed.");
        out.println("    -display         display a list of registered " +
                    "test cases.");
        out.println("    -help            display this help screen");
        out.println("");
    }
    
} //-- CoreTestSuite
