/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2002-2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: ChannelTest.java,v 1.4 2003/06/09 06:28:43 tanderson Exp $
 */
package org.exolab.core.test.messenger;

import java.io.IOException;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.exolab.core.messenger.Channel;
import org.exolab.core.messenger.ChannelHandler;
import org.exolab.core.messenger.ChannelListener;
import org.exolab.core.messenger.Connection;
import org.exolab.core.messenger.ConnectionEventListener;
import org.exolab.core.messenger.ConnectionHandler;
import org.exolab.core.messenger.InvalidChannelException;
import org.exolab.core.messenger.Messenger;


/**
 * This class tests the behaviour of {@link Messenger}
 *
 * @author <a href="mailto:tma@netspace.net.au">Tim Anderson</a>
 * @version $Revision: 1.4 $
 * @see AbstractSessionTestCase
 * @see com.intalio.jms.test.common.SessionTestRunner
 */
public class ChannelTest extends TestCase {

    private static final Log _log = LogFactory.getLog(ChannelTest.class);


    /**
     * Construct an instance of this class for a specific test case.
     *
     * @param name the name of test case
     */
    public ChannelTest(String name) {
        super(name);
    }

    /**
     * Sets up the test suite
     *
     * @return an instance of this class that may be run by 
     * {@link JMSTestRunner}
     */
    public static Test suite() {
        return new TestSuite(ChannelTest.class);
    }

    public void testChannel() throws Exception {
        String uri = "tcp://localhost:3030/test";
        Messenger messenger = Messenger.getInstance();

        final ChannelHandler channelHandler = new ChannelHandler() {
            public void opened(Channel channel) {
                _log.debug("got a new channel");
            }
        };

        ConnectionHandler handler = new ConnectionHandler() {
            public void connected(Connection connection) {
                try {
                    connection.accept("test", channelHandler);
                } catch (IOException exception) {
                    _log.error(exception.getMessage(), exception);
                }
            }
        };

        messenger.accept(uri, handler, null);
        Connection connection = messenger.open(uri, null);
        Channel channel = connection.open("test");
        
        connection.close();
        messenger.close(uri);
    }

    public void testInvalidChannel() throws Exception {
        String uri = "tcp://localhost:3030/test";
        Messenger messenger = Messenger.getInstance();

        ConnectionHandler handler = new ConnectionHandler() {
            public void connected(Connection connection) {
                _log.debug("got a connection");
            }

            public void error(IOException exception) {
                _log.error(exception.getMessage(), exception);
            }
        };

        messenger.accept(uri, handler);
        Connection connection = messenger.open(uri);
        for (int i = 0; i < 10; ++i) {
            try {
                Channel channel = connection.open("test" + i);
                fail("Expected open for channel='test' to fail");
            } catch (InvalidChannelException ignore) {
            }
        }
        
        connection.close();
        messenger.close(uri);
    }

    public void testSend() throws Exception {
        String uri = "tcp://localhost:3030/test";
        Messenger messenger = Messenger.getInstance();

        final ChannelListener listener = new ChannelListener() {
            public void receive(Object message) {
                _log.debug("Received message");
            }

            public Object invoke(Object request) throws Exception {
                _log.debug("Received invocation");
                return request;
            }
            
            public void closed() {
                _log.debug("Channel closed");
            }
        };

        ConnectionHandler handler = new ConnectionHandler() {
            public void connected(Connection connection) {
                try {
                    connection.accept("test", listener);
                } catch (IOException exception) {
                    _log.error(exception.getMessage(), exception);
                }
            }
        };

        messenger.accept(uri, handler);
        StringBuffer message = new StringBuffer();
        for (int i = 0; i < 1000; ++i) {
            message.append(i);
        }

        int[] packet = new int[400];
        for (int i = 0; i < packet.length; ++i) {
            packet[i] = i;
        }

        Connection connection = messenger.open(uri);
        try {
            // give the server time to set up
            Thread.currentThread().sleep(1000);
        } catch (InterruptedException ignore) {
        }

        Channel channel = connection.open("test");
        _log.debug("opened channel=test");
        for (int i = 0; i < 1; ++i) {
            int[] response = (int[]) channel.invoke(packet);
            for (int j = 0; j < response.length; ++j) {
                if (packet[j] != response[j]) {
                    fail("Request different to response");
                }
            }
        }
        
        channel.close();
        connection.close();
        messenger.close(uri);
    }

    // TODO - test RuntimeException, Error with ChannelListener
    
} //-- ChannelTest
