/**
 * Redistribution and use of this software and associated documentation
 * ("Software"), with or without modification, are permitted provided
 * that the following conditions are met:
 *
 * 1. Redistributions of source code must retain copyright
 *    statements and notices.  Redistributions must also contain a
 *    copy of this document.
 *
 * 2. Redistributions in binary form must reproduce the
 *    above copyright notice, this list of conditions and the
 *    following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. The name "Exolab" must not be used to endorse or promote
 *    products derived from this Software without prior written
 *    permission of Exoffice Technologies.  For written permission,
 *    please contact info@exolab.org.
 *
 * 4. Products derived from this Software may not be called "Exolab"
 *    nor may "Exolab" appear in their names without prior written
 *    permission of Exoffice Technologies. Exolab is a registered
 *    trademark of Exoffice Technologies.
 *
 * 5. Due credit should be given to the Exolab Project
 *    (http://www.exolab.org/).
 *
 * THIS SOFTWARE IS PROVIDED BY EXOFFICE TECHNOLOGIES AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT
 * NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
 * EXOFFICE TECHNOLOGIES OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Copyright 2002-2003 (C) Exoffice Technologies Inc. All Rights Reserved.
 *
 * $Id: MessengerTest.java,v 1.4 2003/06/09 06:28:44 tanderson Exp $
 */
package org.exolab.core.test.messenger;

import java.io.IOException;
import java.net.ServerSocket;
import java.rmi.RemoteException;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import org.exolab.core.messenger.Channel;
import org.exolab.core.messenger.Connection;
import org.exolab.core.messenger.ConnectionEventListener;
import org.exolab.core.messenger.ConnectionHandler;
import org.exolab.core.messenger.AcceptorAlreadyRegisteredException;
import org.exolab.core.messenger.Messenger;


/**
 * This class tests the behaviour of {@link Messenger}
 *
 * @author <a href="mailto:tma@netspace.net.au">Tim Anderson</a>
 * @version $Revision: 1.4 $
 * @see AbstractSessionTestCase
 * @see com.intalio.jms.test.common.SessionTestRunner
 */
public class MessengerTest extends TestCase {

    /**
     * The logger
     */
    private static final Log _log = LogFactory.getLog(MessengerTest.class);


    /**
     * Construct an instance of this class for a specific test case.
     *
     * @param name the name of test case
     */
    public MessengerTest(String name) {
        super(name);
    }

    /**
     * Sets up the test suite
     *
     * @return an instance of this class that may be run by 
     * {@link JMSTestRunner}
     */
    public static Test suite() {
        return new TestSuite(MessengerTest.class);
    }

    /**
     * Test that the Messenger can accept new connections
     *
     * @throws Exception for any error
     */
    public void testAccept() throws Exception {
        String uri = "tcp://localhost:3030/test";
        Messenger messenger = Messenger.getInstance();

        TestConnectionListener handler = new TestConnectionListener();
        
        messenger.accept(uri, handler);
        try {
            Connection connection = messenger.open(uri);
            connection.addConnectionEventListener(handler);
            connection.close();
        } catch (Exception exception) {
            _log.error(exception.getMessage(), exception);
        } finally {
            messenger.close(uri);
        }
    }

    /**
     * Test that a Remotexception is thrown if an attempt to open a connection 
     * for a URI where there is no listener
     * 
     * @throws Exception for any error
     */
    public void testNoListener() throws Exception {
        String uri = "tcp://localhost:3030/test";
        Messenger messenger = Messenger.getInstance();

        try {
            Connection connection = messenger.open(uri);
            fail("Managed to open a connection, when no acceptor is " +
                 "listening");
        } catch (RemoteException ignore) {
        } catch (Exception exception) {
            fail("Expected a RemoteException but got exception=" + exception);
        }
    }

    /**
     * Test that a RemoteException is thrown if an attempt is made to accept
     * connections on a port that already in use
     *
     * @throws Exception for any error
     */
    public void testAcceptForPortInUse() throws Exception {
        String uri = "tcp://localhost:3030/test";
        Messenger messenger = Messenger.getInstance();
        ServerSocket socket = new ServerSocket(3030);
        try {
            messenger.accept(uri, new TestConnectionListener());
            fail("Expected a RemoteException to be thrown");
        } catch (RemoteException ignore) {
        } catch (Exception exception) {
            fail("Expected a RemoteException to be thrown, but got " +
                 "exception=" + exception);
        } finally {
            socket.close();
        }
    }

    /**
     * Test that a RemoteException is thrown if an attempt is made to
     * connection to a port owned by a non-Messenger application
     * 
     * @throws Exception for any error
     */
    public void testConnectForPortInUse() throws Exception {
        String uri = "tcp://localhost:3030/test";
        Messenger messenger = Messenger.getInstance();
        final ServerSocket socket = new ServerSocket(3030);
        Runnable runnable = new Runnable() {
            public void run() {
                try {
                    socket.accept();
                } catch (IOException ignore) {
                }
            }
        };

        Thread thread = new Thread(runnable);
        thread.start();

        try {
            Thread.currentThread().sleep(1000); 
            // give the thread time to start

            Connection connection = messenger.open(uri);
            fail("Expected a RemoteException to be thrown");
        } catch (RemoteException ignore) {
        } catch (Exception exception) {
            fail("Expected a RemoteException to be thrown, but got " +
                 "exception=" + exception);
        } finally {
            thread.interrupt();
            socket.close();
        }
    }
    
} //-- MessengerTest
